# -*- coding: utf-8 -*-
require 'sidfm_controller_support'
require 'sidfm_histories_controller_support'

class SidfmHistoriesController < ApplicationController
  unloadable
  layout 'sidfm'
  self.main_menu = false
  
  helper :sidfm
  protect_from_forgery with: :exception

  include SidfmControllerSupport, SidfmHostsControllerSupport

  before_action :update_session_updated_at
  before_action :create_common_data
  before_action :get_sidfm_roles
  before_action :get_vm_version
  before_action :clear_import_session
  before_action :clear_export_session, only: [ :index ]
  before_action :check_roles 
  before_action :prevent_access, only: [ :edit_comment, :get_data_for_bulk_update, :bulk_update_comment ]
  before_action -> { get_summary_table_lines('histories') }
 
  private
  
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = if @role_admin then
                           nil
                         else
                           User.current.categories.order(:id).pluck(:id)
                         end
    @user_host_ids = if @role_admin then
                       nil
                     else
                       SidfmdCategoriesHost.joins(:sidfmd_host)
                                           .where(:sidfmd_category_id => @user_category_ids)
                                           .where(:sidfmd_hosts => { :flg_delete => false })
                                           .select(:sidfmd_host_id)
                                           .order(:sidfmd_host_id)
                                           .distinct
                                           .pluck(:sidfmd_host_id)
                     end
  end

  def prevent_access
    if !@sidfm_roles.include?(:use) or @sidfm_roles.include?(:host_limited_admin) then #管理者と報告者のみアクセス可能
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
  end

  def check_param_journal_ids(journal_ids)
    #管理者権限以外の時に、journal_idsに自分のカテゴリ範囲外のものが含まれていないかチェック
    !SidfmdIssue.joins(:issue)
                .joins(%| INNER JOIN journals ON journals.journalized_id = issues.id |) 
                .where.not(:sidfmd_issues => { :sidfmd_host_id => @user_host_ids })
                .where(:journals => { :id => journal_ids })
                .exists?
  end

  # パラメータ解析
  def parse_params
    @histories = {}
    pages_params = params.fetch(:pages, {}).permit(:order, :offset, :per_page, :filters, :checked, :all_checked)

    @histories[:order] = (pages_params[:order] || 'created_on_desc')
    @histories[:filters] = JSON.parse((pages_params[:filters] || '{}'), symbolize_names: true)
    @histories[:offset] = (pages_params[:offset] || 0).to_i
    @histories[:per_page] = (@summary_table_lines || pages_params[:per_page] || 10).to_i
    @histories[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @histories[:all_checked] = pages_params[:all_checked] || ''
  end

  def create_content
    @histories[:all_content] = SidfmdIssue.create_journals_content(
      :role_admin => @role_admin,
      :user_host_ids => @user_host_ids
    )
  end

  FILTER_KEYS = {}

  #フィルターの初期化
  def init_filter_keys
    if FILTER_KEYS.size == 0 then
      default_histories_filters.each do |key, value|
        FILTER_KEYS[key] = value
      end
    end  
  end

  def check_filters
    init_filter_keys
    FILTER_KEYS.each do |key, val|
      unless @histories[:filters].has_key?(key) then
        @histories[:filters][key] = val
      end
    end
    @histories[:filters][:roles][:ids] = @histories[:filters][:roles][:ids].map(&:to_i)
    @histories[:filters][:users][:ids] = @histories[:filters][:users][:ids].map(&:to_i)
    @histories[:filters][:hosts][:ids] = @histories[:filters][:hosts][:ids].map(&:to_i)
    @histories[:filters][:categories][:ids] = @histories[:filters][:categories][:ids].map(&:to_i)
    @histories[:filters][:ctags][:ids] = @histories[:filters][:ctags][:ids].map(&:to_i) 
    @default_filters = FILTER_KEYS
  end

  def filter_content(is_limit_data = true)
    # filter
    check_filters
    @histories[:content_filtered] = filter_histories_content(
      :content => @histories[:all_content],
      :filters => @histories[:filters],
      :role_admin => @role_admin,
      :user_category_ids => @user_category_ids
    )
    # sort
    @histories[:content_filtered] = sort_histories_content(
      :content => @histories[:content_filtered],
      :order => @histories[:order]
    )

    @histories[:content] = @histories[:content_filtered]
    if is_limit_data then
      @histories[:content] = @histories[:content].offset(@histories[:offset])
                                                 .limit(@histories[:per_page])
    end
  end

  def get_roles
    role_names = [ 'SIDfmVM管理者','SIDfmVM報告者', 'SIDfmVM利用者', 'SIDfmVM閲覧者' ]
    Role.where(:name => role_names)
        .map{|r| {:id => r.id, :name => r.name } }
        .sort_by{|role| role_names.index(role[:name])}
  end
 
  def get_user_categories
    sort_categories = ->(categories){
      categories.order(%|
        CASE WHEN is_system = true THEN 10 ELSE 100 END
      |)
      .order(%|
        CASE WHEN is_system = true THEN CAST(id AS text) ELSE LOWER(name COLLATE "C") END
      |)
    }
    user_categories = if @role_admin then
                        sort_categories.(SidfmdCategory.all)
                      else
                        sort_categories.(SidfmdCategory.where("is_system = 'true' OR id IN (?)", @user_category_ids))
                      end
  end

  def create_summary_history_html(history, **opts)
    data = {
      :chkbox => view_context.create_summary_history_chkbox(history.journal_id, opts[:checked]),
      :created_on => view_context.create_summary_history_created_on(history.journal_created_on),
      :login_name => view_context.create_summary_history_user(history.login_name, history.user_fullname, history.email_address, history.is_anonymous),
      :operation => view_context.create_summary_history_operation(history.operation, history.old_value, history.current_value),
      :comment => view_context.create_summary_history_comment(history.comment, history.comment == HISTORY_COMMENT_DELETED_NOTE, history.journal_id, @sidfm_roles),
      :vul_title => view_context.create_summary_history_vul_title(
        :vul_title => history.vul_title,
        :vul_short_description => history.vul_short_description,
        :host_id => history.host_id,
        :html_id => history.html_id,
        :create_datetime => @htmlid_dates[history.html_id] ? @htmlid_dates[history.html_id][:create_datetime] : nil,
        :update_datetime => @htmlid_dates[history.html_id] ? @htmlid_dates[history.html_id][:update_datetime] : nil,
      ),
      :hostname => view_context.create_summary_history_hostname(
        history.hostname,
        history.ext_name,
        history.display_name,
        history.host_id,
        history.html_id
      ),
      :status => view_context.create_summary_history_status(history.status_id),
      :due_date => view_context.create_summary_history_due_date(history.due_date),
      :ctags => view_context.create_summary_custom_tag(
        (@issue_ctags[history.id] || []).first,
        history.html_id,
        history.vul_title,
        @sidfm_roles,
        { :read_only => true, :sidfmd_host_id => history.host_id, :elm_id => history.journal_id }
      ) 
    }
  end

  def create_issue_ctags(issue_ids)
    @issue_ctags = {} 
    issue_ctags = SidfmdIssueCtag.joins(:sidfmd_ctag)
      .where(:sidfmd_issue_ctags => { :sidfmd_issue_id => issue_ids })
    unless @role_admin then
      issue_ctags = issue_ctags.where("sidfmd_ctags.sidfmd_category_id IS NULL OR sidfmd_ctags.sidfmd_category_id IN (?)", @user_category_ids)
    end
    issue_ctags
      .order("sidfmd_ctags.priority DESC")
      .order(%| LOWER(sidfmd_ctags.keyword COLLATE "C") |)
      .pluck( #pluckが高速
        "sidfmd_issue_id",
        "sidfmd_ctags.id",
        "sidfmd_ctags.keyword",
        "sidfmd_ctags.sidfmd_category_id",
        "sidfmd_ctags.priority",
        "sidfmd_ctags.color",
        "sidfmd_ctags.bgcolor",
      )
      .each do |issue_ctag|
        @issue_ctags[issue_ctag[0]] = [] unless @issue_ctags[issue_ctag[0]]
        @issue_ctags[issue_ctag[0]] << {
          :id => issue_ctag[1],
          :keyword => issue_ctag[2],
          :category_id => issue_ctag[3],
          :priority => issue_ctag[4],
          :text_color => issue_ctag[5],
          :bg_color => issue_ctag[6],
        }
      end
  end

  def create_htmlid_create_update_time(html_ids)
    @htmlid_dates = SidfmdHtmlidDate.create_htmlid_create_update_time(html_ids)
  end

  public

  def index
    parse_params
    create_content
    filter_content
    @roles = get_roles
    @user_categories = get_user_categories
    @deleted_comment = HISTORY_COMMENT_DELETED_NOTE
  end

  def bulk_update_comment
    f_params = params.permit(:journal_ids, :comment, :comment_delete_flg)#, :page_info)
    begin
      journal_ids = JSON.parse(f_params[:journal_ids] || '[]')
      #page_info = JSON.parse(f_params[:page_info])
    rescue JSON::ParserError => e
      log_exception(e)
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end
    if journal_ids.empty? then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end
    unless @role_admin then
      unless check_param_journal_ids(journal_ids) then
        render :layout => false, :json => { :err_message => '不正な操作です。※ユーザのカテゴリが変更されていないかご確認下さい。' } and return
      end
    end
    begin
      target_journals = Journal.where(:id => journal_ids)
      if f_params[:comment_delete_flg] then
        target_journals.update_all(:notes => HISTORY_COMMENT_DELETED_NOTE)
      elsif f_params[:comment] and f_params[:comment].size > 0 then
        target_journals.update_all(:notes => f_params[:comment])
      else
        render :layout => false, :json => { :err_message =>  '変更に失敗しました。' } and return
      end
    rescue Exception => e
      log_exception(e)
      render :layout => false, :json => { :err_message =>  '変更に失敗しました。' } and return
    end
    #dump = {
    #  :action => :index,
    #  :'pages[offset]' => page_info["offset"],
    #  :'pages[per_page]' => page_info["per_page"],
    #  :'pages[filters]' => JSON.generate(page_info["filters"]),
    #  :'pages[order]' => page_info["order"]
    #}
    #redirect_to dump
    render :layout => false, :json => { :err_message => '' }
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze
  MAX_CELL_LENGTH = 30000
  MAX_LINE_NUM = 254 # =最大改行数は253

  def read_download_data
    GC.start
    f_params = params.permit(:target, :ids, :executed_size, :process, :order, :filters)
    if !f_params[:ids] or !f_params[:target] or !f_params[:order] or !f_params[:filters] or !f_params[:executed_size] or !f_params[:process] then
      logger.info('params error')
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    executed_size = f_params[:executed_size].to_i

    begin 
      journal_ids = JSON.parse(f_params[:ids])
      filters = JSON.parse(f_params[:filters], symbolize_names: true)
    rescue JSON::ParserError => e
      logger.info('json parser error')
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    if ![ 'json', 'tsv' ].include?(f_params[:target]) then
      logger.info('file target error')
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end
  
    
    @histories = {}
    @histories[:order] = f_params[:order]
    @histories[:filters] = filters
    create_content
    filter_content(false)
  
    status_names = {
      SidfmdIssue::STATUS_NOTCONFIRMED => '未対策',
      SidfmdIssue::STATUS_ONGOING => '対策中',
      SidfmdIssue::STATUS_FIXED => '対策済',
      SidfmdIssue::STATUS_NOTAFFECT => '影響無'
    }
   
    case f_params[:target]
    when 'json' then
      
      if f_params[:process] == 'init' then
        @histories[:content] = @histories[:content].where(:journals => {:id => journal_ids }) if journal_ids.any?
        journal_ids = @histories[:content].map{|j| j.journal_id }
        separate_char_index = @histories[:order].rindex('_')
        col_name = @histories[:order].slice(0..separate_char_index - 1)
        order_type = @histories[:order].slice(separate_char_index + 1..@histories[:order].size - 1) 
        init_json_summary_table_progress_download(
          :ids =>journal_ids, 
          :order_col_name => col_name, 
          :order_type => order_type, 
          :filters => filters, 
          :max_download_size => Sidfm.get_history_max_download_size
        )
      else
        create_issue_ctags(@histories[:content].where(:journals => { :id => journal_ids }).pluck("sidfmd_issues.id"))
        create_htmlid_create_update_time(@histories[:content].where(:journals => { :id => journal_ids }).pluck("sidfmd_issues.html_id"))
        create_json_data = Proc.new{|id|
          history = @histories[:content].find_by(:journals => {:id => id })
          next unless history
          item = {}
          item[:journal_id] = history.journal_id
          item[:created_on] = history.journal_created_on.strftime('%Y/%m/%d %H:%M:%S')
          item[:user] = {}
          item[:user][:id] = history.user_id
          item[:user][:login] = history.is_anonymous ? '削除済みユーザ' : history.login_name
          item[:user][:fullname] = history.is_anonymous ? '' : history.user_fullname
          item[:user][:email_address] = history.is_anonymous ? '' : history.email_address
          item[:operation] = {
            :title => history.operation ? history.operation : '',
            :content => if history.operation == 'ステータス変更' and history.old_value and history.current_value then
                          status_names[history.old_value.to_i] + ' から ' + status_names[history.current_value.to_i] + 'へ'
                        elsif history.operation == '対策予定日の変更' then
                          (history.old_value ? history.old_value : '未設定') + ' から ' + (history.current_value ? history.current_value : '未設定') + 'へ'
                        else
                          ''
                        end                         
          }
          item[:comment] = history.comment
          item[:vul] = {
            :title => "[#{history.html_id > 0 ? history.html_id : "KB#{-history.html_id}"}] #{history.vul_title}",
            :short_description => history.vul_short_description,
            :url => url_for(
                      :controller => :sidfm_vulnerability,
                      :action => :index,
                      :id => history.html_id,
                      :'form[history]' => [ history.host_id, history.html_id ].to_json
                    ),
            :create_date => (@htmlid_dates[history.html_id] && @htmlid_dates[history.html_id][:create_datetime]) ? @htmlid_dates[history.html_id][:create_datetime].strftime('%Y/%m/%d') : nil,
            :update_date => (@htmlid_dates[history.html_id] && @htmlid_dates[history.html_id][:update_datetime]) ? @htmlid_dates[history.html_id][:update_datetime].strftime('%Y/%m/%d') : nil, 
          }
          item[:host] = {
            :hostname => history.hostname,
            :ext_name => history.ext_name,
            :display_name => history.display_name,
            :url => url_for(
                      :controller => :sidfm_host,
                      :action => :index,
                      :id => history.host_id,
                      :'form[history]' => [ history.html_id, history.host_id ].to_json
                    )
          }
          item[:status] = status_names[history.status_id]
          item[:due_date] = history.due_date ? history.due_date.strftime('%Y/%m/%d') : ''
          item[:ctags] = (@issue_ctags[history.id] || []).map{|ctag|
            {
              :keyword => ctag[:keyword],
              :category_id => ctag[:category_id],
              :color => ctag[:text_color],
              :background_color => ctag[:bg_color],
              :priority => ctag[:priority],
            }
          }
          item
        }
        json_tsv_summary_table_progress_download(journal_ids, create_json_data, executed_size)
      end 
    when 'tsv' then
      header = [ '日時', 'ログイン名', 'メールアドレス', '操作', '操作内容', 'コメント', '脆弱性', '脆弱性登録日', '脆弱性最終更新日', 'ホスト名', '対応状況', '対策予定日', 'カスタムタグ' ]
      if f_params[:process] == 'init' then
        @histories[:content] = @histories[:content].where(:journals => {:id => journal_ids }) if journal_ids.any?
        journal_ids = @histories[:content].map{|j| j.journal_id }
        init_tsv_summary_table_progress_download(journal_ids, header, Sidfm.get_history_max_download_size)
      else
        create_issue_ctags(@histories[:content].where(:journals => { :id => journal_ids }).pluck("sidfmd_issues.id"))
        create_htmlid_create_update_time(@histories[:content].where(:journals => { :id => journal_ids }).pluck("sidfmd_issues.html_id"))
        create_tsv_data = Proc.new{|id|
          history = @histories[:content].find_by(:journals => {:id => id })
          item = []
          item << history.journal_created_on.strftime('%Y/%m/%d %H:%M:%S')
          item << (history.is_anonymous ? '削除済みユーザ' : history.login_name)
          item << (history.is_anonymous ? '' : history.email_address)
          item << (history.operation ? history.operation : '')
          ope_content = if history.operation == 'ステータス変更' and history.old_value and history.current_value then
                          status_names[history.old_value.to_i] + ' から ' + status_names[history.current_value.to_i] + 'へ'
                        elsif history.operation == '対策予定日の変更' then
                          (history.old_value ? history.old_value : '未設定') + ' から ' + (history.current_value ? history.current_value : '未設定') + 'へ'
                        else
                          ''
                        end
          item << ope_content
          comment = history.comment
          split_comment_by_line = comment.split("\n")
          if split_comment_by_line.size > MAX_LINE_NUM then
            comment = split_comment_by_line.slice(0..MAX_LINE_NUM - 1).join("\n")
          end
          if comment.size > MAX_CELL_LENGTH then
            comment = comment.slice(0..MAX_CELL_LENGTH - 1)
          end
          item << comment
          item << "[#{history.html_id > 0 ? history.html_id : "KB#{-history.html_id}"}] #{history.vul_title}"
          item << (@htmlid_dates[history.html_id] && @htmlid_dates[history.html_id][:create_datetime] ? @htmlid_dates[history.html_id][:create_datetime].strftime('%Y/%m/%d') : '')
          item << (@htmlid_dates[history.html_id] && @htmlid_dates[history.html_id][:update_datetime] ? @htmlid_dates[history.html_id][:update_datetime].strftime('%Y/%m/%d') : '')
          item << history.display_name
          item << status_names[history.status_id]
          item << (history.due_date ? history.due_date.strftime('%Y/%m/%d') : '')
          item << (@issue_ctags[history.id] || []).map{|ctag| escape_category_tsv(ctag[:keyword]) }.join(',')
          item
        }
        json_tsv_summary_table_progress_download(journal_ids, create_tsv_data, executed_size)
      end
    end
  end

  def download
    f_params = params.fetch(:download, {}).permit(:format)
    case f_params[:format]
    when 'json' then
      send_summary_table_json_file("vmlist-histories-json-#{Time.current.strftime('%Y%m%d%H%M%S')}.json")
    when 'tsv' then
      send_summary_table_tsv_file("vmlist-histories-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv")
    end
  end

  #api
  def edit_comment
    f_params = params.permit(:journal_id, :comment, :comment_delete_flg)
    if !f_params[:journal_id] then
      render :layout => false, :json => {:err_message => '不正な操作です。'} and return 
    end

    comment = f_params[:comment]
    begin
      unless @role_admin then
        unless check_param_journal_ids([ f_params[:journal_id].to_i ]) then
          render :layout => false, :json => { :err_message => '不正な操作です。※ユーザのカテゴリが変更されていないかご確認下さい。' } and return
        end
      end 
      journal = Journal.find_by(:id => f_params[:journal_id].to_i)
      if journal then
        if f_params[:comment_delete_flg] == 'true' then
          journal.update(:notes => HISTORY_COMMENT_DELETED_NOTE)
          comment = ''
        elsif comment and comment.size > 0 then
          journal.update(:notes => comment)
        else
          render :layout => false, :json => {:err_message => '変更に失敗しました。'} and return
        end
      end
    rescue Exception => e
      log_exception(e)
      render :layout => false, :json => {:err_message => '変更に失敗しました。'} and return
    end

    render :layout => false, :json => {:err_message => '', :comment => comment}

  end

  def get_data_for_bulk_update
    f_params= params.permit(:journal_ids)
    @histories = {}
    journal_ids = JSON.parse(f_params[:journal_ids] || '[]')
    create_content
    r = @histories[:all_content].where(:journals => { :id => journal_ids })
                                .limit(1000)
                                .select(
                                  'journals.notes AS comment',
                                  %| CASE WHEN users.type = 'AnonymousUser' THEN '削除済みユーザ' ELSE users.login END AS login |,
                                  'email_addresses.address AS address',
                                  " CASE 
                                      WHEN users.type = 'AnonymousUser' THEN
                                        ''
                                      ELSE
                                        (users.lastname || ' ' || users.firstname)
                                      END
                                      AS user_fullname
                                   ",
                                    %| CASE 
                                       WHEN users.type = 'AnonymousUser' THEN
                                         true
                                       ELSE
                                         false
                                       END
                                       AS is_anonymous
                                    |
                                )
    render :layput => false, :json => r
  end

  #def api_get_display_summary_rows
  #  GC.start
  #  f_params = params.permit(:ids, :checked)
  #  journal_ids = JSON.parse(f_params[:ids])
  #  checked = JSON.parse(f_params[:checked]).map(&:to_i)
  #  @histories = {}
  #  create_content
  #  ret = {}
  #  @histories[:all_content].where(:journals => { :id => journal_ids }).each do |history|
  #    ret[history.journal_id] = create_summary_history_html(history, { :checked =>  checked })
  #  end
  #
  #  render :layout => false, :json => ret
  #end

  def api_get_display_summary_content
    GC.start
    f_params = params.fetch(:pages, {}).permit(:per_page)
    parse_params
    @histories[:per_page] = f_params[:per_page].to_i
    create_content
    filter_content
    create_issue_ctags(@histories[:content].pluck("sidfmd_issues.id"))
    create_htmlid_create_update_time(@histories[:content].pluck("sidfmd_issues.html_id"))
    ids = @histories[:content_filtered].pluck('journals.id')
    ret = {
      :rows => @histories[:content].map{|history|
        create_summary_history_html(history, { :checked =>  @histories[:checked] })
      },
      :total => ids.size,
      :all_ids => ids,
      :sidfm_roles => @sidfm_roles,
      :user_cats => @user_category_ids
    }
    render :layout => false, :json => ret
  end

end
