# -*- coding: utf-8 -*-

require 'sidfm_controller_support'
require 'sidfm_users_controller_support'
require 'sidfm_hosts_controller_support'
require 'sidfm_host_controller_support'
require 'sidfm_vulnerabilities_controller_support'
require 'sidfm_vulnerability_controller_support'
require 'sidfm_histories_controller_support'

# SIDfm VM API
class SidfmVmApiController < ApplicationController
  unloadable

  layout false

  skip_before_action :check_if_login_required
  skip_before_action :verify_authenticity_token

  private

  include SidfmControllerSupport, SidfmUsersControllerSupport, SidfmHostsControllerSupport, SidfmHostControllerSupport, SidfmVulnerabilitiesControllerSupport, SidfmVulnerabilityControllerSupport, SidfmHostsControllerSupport
  before_action :authorize, :except => [ :index ]
  before_action :get_vm_version, :except => [ :index ]
  before_action :get_sidfm_user, :except => [ :index ]
  before_action :get_sidfm_roles, :except => [ :index ]
  before_action :check_license, :except => [ :index ]
  before_action :check_roles, :except => [ :index ]

  # 応答の雛形を返す
  def create_response
    { "success" => false, "response" => {} }
  end

  # User.currentを設定する
  def set_user(user)
    User.current = user
    logger.info("  API Current user: #{user.login} (id=#{user.id})")
  end

  # ※ Redmineの認証処理を上書きする.
  # 認証処理
  # 認証成功時は、@auth=true, @apikeyにSidfmdVmApikeyのインスタンス、User.currentにapikeyに対応したユーザが設定される。
  def authorize
    params_fetched = params.fetch(:q, {})
    @api_params = params_fetched.permit(:api_key, :order, :offset, :per_page, :host_id, :html_id).tap do |whitelist|
      # q[filters][任意]をpermit
      whitelist[:filters] = params_fetched[:filters] if params_fetched[:filters].is_a?(Hash)
    end
    @filters = (@api_params[:filters] || {})
    header_apikey = request.headers['X-API-Key']
    params_apikey = @api_params[:api_key]
    @auth = false
    @apikey = nil
    if header_apikey then
      # X-API-Keyヘッダの認証
      @apikey = SidfmdVmApikey.auth(header_apikey)
      if @apikey then
        @auth = true
        set_user(@apikey.user)
      end
    end
    if !@auth and params_apikey then
      # q[api_key]パラメータの認証
      @apikey = SidfmdVmApikey.auth(params_apikey)
      if @apikey then
        @auth = true
        set_user(@apikey.user)
      end
    end
    unless @auth then
      res = create_response
      res['response']['message'] = 'invalid apikey'
      render :json => res and return
    end
    # offsetの正規化
    @offset = @api_params[:offset]
    if @offset then
      @offset = @offset.to_i
      if @offset < 0 then
        @offset = 0
      end
    else
      @offset = 0
    end
    # per_pageの正規化
    @per_page = @api_params[:per_page]
    if @per_page then
      @per_page = @per_page.to_i
      if @per_page <= 0 then
        @per_page = 10
      end
    else
      @per_page = 10
    end
    if @per_page > 200 then
      @per_page = 200
    end
  end

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      res = create_response
      res['response']['message'] = 'invalid user.'
      render :json => res and return
    end
    unless @sidfm_roles.include?(:view) then
      res = create_response
      res['response']['message'] = 'unauthorized access.'
      render :json => res and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = if @role_admin then
                           nil
                         else
                           User.current.categories.pluck(:id)
                         end
    @user_host_ids = if @role_admin then
                       nil
                     else
                       SidfmdCategoriesHost.joins(:sidfmd_host)
                         .where(:sidfmd_category_id => @user_category_ids)
                         .where(:sidfmd_hosts => { :flg_delete => false })
                         .select(:sidfmd_host_id)
                         .order(:sidfmd_host_id).distinct.pluck(:sidfmd_host_id)
                     end
  end

  def get_category_data(category)
    {
      :id => category.id,
      :name => category.name,
      :desc => category.desc,
      :hostlim => category.hostlim,
      :is_system => category.is_system
    }
  end

  SETTING_HOST_NAME = Setting.host_name
  SETTING_PROTOCOL = Setting.protocol

  def hostname_from_setting_hostname
    hostname = SETTING_HOST_NAME
    if hostname.match(/\//) then
      #ホスト名に/redmine等のパスがが入っていると、urlが/remine/redmineのようになり、正しくならないので修正する
      hostname = hostname.split(/\//, -1).first
    end
    hostname
  end

  def get_host_detail_link(sidfmd_host_id, opts = {})
    data = {
      :protocol => SETTING_PROTOCOL,
      :controller => :sidfm_host,
      :action => :index,
      :id => sidfmd_host_id
    }
    hostname = hostname_from_setting_hostname
    if hostname then
      data[:host] = hostname
    end
    if opts.size > 0 then
      data.merge!(opts)
    end
    url_for(data)
  end

  def get_vul_detail_link(html_id, opts = {})
    data = {
      :protocol => SETTING_PROTOCOL,
      :controller => :sidfm_vulnerability,
      :action => :index,
      :id => html_id,
    }
    hostname = hostname_from_setting_hostname
    if hostname then
      data[:host] = hostname
    end
    if opts.size > 0 then
      data.merge!(opts)
    end
    url_for(data)
  end

  def get_vul_infomation_link(html_id)
    if html_id.to_i < 0 then
      return nil #htmlがマイナスは詳細がない
    end
    data = {
      :protocol => SETTING_PROTOCOL,
      :controller => :sidfm_informations,
      :action => :vuln,
      :id => html_id,
    }
    hostname = hostname_from_setting_hostname
    if hostname then
      data[:host] = hostname
    end
    url_for(data)
  end

  def get_ctag_data(ctag)
    {
      :id => ctag.id,
      :keyword => ctag.keyword,
      :category_id => ctag.sidfmd_category_id,
      :color => ctag.color,
      :background_color => ctag.bgcolor,
      :priority => ctag.priority
    }
  end

  #脆弱性毎のカスタムタグ
  def get_vul_ctags(vul_view_ids, host_id = nil)
    vul_ctags = {}
    ctags = SidfmdCtag.joins(:sidfmd_vv_ctags)
      .where(:sidfmd_vv_ctags => {
        :sidfmd_vul_view_id => vul_view_ids,
      })
      .order(%| sidfmd_ctags.priority DESC |)
      .order(%| LOWER(sidfmd_ctags.keyword COLLATE "C") |)
      .select(
        "sidfmd_ctags.id",
        "sidfmd_ctags.keyword",
        "sidfmd_ctags.sidfmd_category_id",
        "sidfmd_ctags.color",
        "sidfmd_ctags.bgcolor",
        "sidfmd_ctags.priority",
        "sidfmd_vv_ctags.sidfmd_vul_view_id AS view_id"
      )
    if host_id then
      ctags = ctags.where(:sidfmd_vv_ctags => { :sidfmd_host_id => host_id })
    end
    unless @role_admin then
      ctags = ctags.where("sidfmd_ctags.sidfmd_category_id IS NULL OR sidfmd_ctags.sidfmd_category_id IN (?)",  @user_category_ids)
    end
    ctags.each do |ctag|
      vul_ctags[ctag.view_id] = [] unless vul_ctags[ctag.view_id]
      vul_ctags[ctag.view_id] << get_ctag_data(ctag)
    end
    
    vul_ctags
  end

  # ホスト毎のカスタムタグ
  def get_host_ctags(host_view_ids, html_id = nil)
    host_ctags = {}
    ctags = SidfmdCtag.joins(:sidfmd_hv_ctags)
      .where(:sidfmd_hv_ctags => { :sidfmd_host_view_id => host_view_ids })
      .order(%| sidfmd_ctags.priority DESC |)
      .order(%| LOWER(sidfmd_ctags.keyword COLLATE "C") |)
    if html_id then
      ctags = ctags.where(:sidfmd_hv_ctags => { :html_id => html_id })
    end
    unless @role_admin then
      ctags = ctags.where("sidfmd_ctags.sidfmd_category_id IS NULL OR sidfmd_ctags.sidfmd_category_id IN (?)",  @user_category_ids)
    end
    ctags.select(
      "sidfmd_ctags.id",
      "sidfmd_ctags.keyword",
      "sidfmd_ctags.sidfmd_category_id",
      "sidfmd_ctags.color",
      "sidfmd_ctags.bgcolor",
      "sidfmd_ctags.priority",
      "sidfmd_hv_ctags.sidfmd_host_view_id AS view_id"
    ).each do |ctag|
      host_ctags[ctag.view_id] = [] unless host_ctags[ctag.view_id]
      host_ctags[ctag.view_id] << get_ctag_data(ctag)
    end
    host_ctags
  end

  # SidfmdIssueのカスタムタグ
  def get_issue_ctags(sidfmd_issue_ids)
    issue_ctags = {}
    ctags = SidfmdIssueCtag.joins(:sidfmd_ctag)
      .where(:sidfmd_issue_ctags => { :sidfmd_issue_id => sidfmd_issue_ids })
    unless @role_admin then
      ctags = ctags.where("sidfmd_ctags.sidfmd_category_id IS NULL OR sidfmd_ctags.sidfmd_category_id IN (?)", @user_category_ids)
    end
    ctags
      .order("sidfmd_ctags.priority DESC")
      .order(%| LOWER(sidfmd_ctags.keyword COLLATE "C") |)
      .select(
        "sidfmd_issue_id",
        "sidfmd_ctags.id",
        "sidfmd_ctags.keyword",
        "sidfmd_ctags.sidfmd_category_id",
        "sidfmd_ctags.priority",
        "sidfmd_ctags.color",
        "sidfmd_ctags.bgcolor",
      )
      .each do |issue_ctag|
        issue_ctags[issue_ctag.sidfmd_issue_id] = [] unless issue_ctags[issue_ctag.sidfmd_issue_id]
        issue_ctags[issue_ctag.sidfmd_issue_id] << get_ctag_data(issue_ctag)
      end
    issue_ctags
  end

  def convert_filter_product_name_into_id
    if @filters[:products][:product_all_version_names] then
      @filters[:products][:product_all_version_ids] = SidfmdProduct.where(:name => @filters[:products][:product_all_version_names])
                                                                   .pluck(:pid)
    end
    if @filters[:products][:product_names] then
      @filters[:products][:product_ids] = SidfmdProduct.where(:name => @filters[:products][:product_names])
                                                       .pluck(:pid)
    end
    if @filters[:products][:version_names] then
      @filters[:products][:version_ids] = SidfmdVersion.where(:fullname => @filters[:products][:version_names])
                                                       .pluck(:vid)
    end
    if @filters[:products][:lproduct_all_version_names] then
      @filters[:products][:lproduct_all_version_ids] = SidfmdLproduct.where(:name => @filters[:products][:lproduct_all_version_names])
                                                                     .pluck(:id)
    end
    if @filters[:products][:lproduct_names] then
      @filters[:products][:lproduct_ids] = SidfmdLproduct.where(:name => @filters[:products][:lproduct_names])
                                                         .pluck(:id)
    end
    if @filters[:products][:lversion_names] then
      @filters[:products][:lversion_ids] = SidfmdLversion.where(:fullname => @filters[:products][:lversion_names])
                                                         .pluck(:id)
    end
  end

  def convert_filter_status_name_into_id
    status_names = SidfmdIssue::STATUS_NAMES
    if @filters[:status].is_a?(String) && !@filters[:status].match(/[0-9]/) then
      @filters[:status] = status_names.index(@filters[:status].strip).to_s
    elsif @filters[:status].is_a?(Array) then
      @filters[:status] = @filters[:status].map{|status_name|
        status_id = status_names.index(status_name.strip)
        ret = status_id ? status_id.to_s : nil
      }.compact.join(':')
    end
  end

  def convert_filter_sri_name_into_id
    sris = {
      SidfmdIssueSri::RISK_NONE => SidfmdIssueSri::VALUE_NONE,
      SidfmdIssueSri::RISK_LOW => SidfmdIssueSri::VALUE_LOW,
      SidfmdIssueSri::RISK_MEDIUM => SidfmdIssueSri::VALUE_MEDIUM,
      SidfmdIssueSri::RISK_HIGH => SidfmdIssueSri::VALUE_HIGH,
      SidfmdIssueSri::RISK_CRITICAL => SidfmdIssueSri::VALUE_CRITICAL,
      SidfmdIssueSri::RISK_FIXED => SidfmdIssueSri::VALUE_FIXED,
    }
    @filters[:sris] = @filters[:sris].map{|sri| sris[sri.strip] }.compact
  end

  def convert_filter_area_name_into_id
    if @filters[:area].to_s.size > 0 then
      area = SidfmdArea.find_by_name(@filters[:area])
      @filters[:area] = area ? area.area_id.to_s : ''
    end
  end

  def convert_filter_category_name_into_id
    if @filters[:categories] && @filters[:categories][:names] then
      @filters[:categories][:ids] = SidfmdCategory.where(:name => @filters[:categories][:names]).pluck(:id)
    end
  end

  def convert_filter_category_id_type_into_int
    if @filters[:categories] && @filters[:categories][:ids] then
      @filters[:categories][:ids] = @filters[:categories][:ids].map(&:to_i)
    end
  end

  def convert_filter_ctag_name_into_id
    if @filters[:ctags].is_a?(Array) then
      @filters[:ctags] = { :ids => SidfmdCtag.where(:keyword => @filters[:ctags]).pluck(:id) }
    end
  end

  def convert_filter_role_name_into_id
    if @filters[:roles] && @filters[:roles][:names] then
      @filters[:roles][:ids] = Role.joins(:member_roles).joins("
        INNER JOIN members ON members.id = member_roles.member_id
      ")
      .where(:members => { :project_id => @sidfm_project.id })
      .where(:roles => { :name => @filters[:roles][:names]})
      .pluck(:id)
    end
  end

  def convert_filter_user_login_into_id
    @filters[:users][:ids] = [] unless @filters[:users][:ids]
    if @filters[:users] && @filters[:users][:logins] then
      @filters[:users][:ids] +=
        User.joins(:members)
            .where(:members => { :project_id => @sidfm_project.id })
            .where(:users => { :login => @filters[:users][:logins] })
            .pluck(:id)
    end
  end

  def convert_filter_user_mail_into_id
    @filters[:users][:ids] = [] unless @filters[:users][:ids]
    if @filters[:users] && @filters[:users][:mail_addresses] then
      @filters[:users][:ids] +=
        User.joins(:members)
            .joins(:email_addresses)
            .where(:members => { :project_id => @sidfm_project.id })
            .where(:email_addresses => { :address => @filters[:users][:mail_addresses] })
            .pluck(:id)
    end
  end

  def convert_filter_user_name_into_id
    @filters[:users][:ids] = [] unless @filters[:users][:ids]
    if @filters[:users] && @filters[:users][:names] then
      @filters[:users][:ids] +=
        User.joins(:members)
            .where(:members => { :project_id => @sidfm_project.id })
            .where("(users.lastname || ' ' || users.firstname) IN (?)", @filters[:users][:names])
            .pluck(:id)
    end
  end

  def convert_filter_hostname_into_id
    @filters[:hosts][:ids] = [] unless @filters[:hosts][:ids]
    if @filters[:hosts] && @filters[:hosts][:hostnames] then
      @filters[:hosts][:ids] += SidfmdHostView.where(:hostname => @filters[:hosts][:hostnames])
                                              .pluck(:sidfmd_host_id)
    end
  end

  def convert_filter_displayname_into_id
    @filters[:hosts][:ids] = [] unless @filters[:hosts][:ids]
    if @filters[:hosts] && @filters[:hosts][:display_names] then
      @filters[:hosts][:ids] += SidfmdHostView.where(:display_name => @filters[:hosts][:display_names])
                                              .pluck(:sidfmd_host_id)
    end
  end

  def set_filter_duedate
    duedate = '::'
    if @filters[:due_date_begin] && @filters[:due_date_end] then
      duedate = ":#{@filters[:due_date_begin]}:#{@filters[:due_date_end]}"
    elsif @filters[:due_date_begin] then
      duedate = ":#{@filters[:due_date_begin]}:"
    elsif @filters[:due_date_end] then
      duedate = "::#{@filters[:due_date_end]}"
    end
    if @filters[:due_date_nodate].to_s == 'true' then
      duedate = "notset#{duedate}"
    end
    @filters[:due_date] = duedate
  end

  def get_filter_date_param(date_begin, date_end, outside_period)
    date = '::'
    if date_begin && date_end then
      date = ":#{date_begin}:#{date_end}"
    elsif date_begin then
      date = ":#{date_begin}:"
    elsif date_end then
      date = "::#{date_end}"
    end
    if date != '::' then
      if outside_period.to_s == 'true' then
        date = "rev#{date}"
      else
        date = "in#{date}"
      end
    end
    date
  end

  def set_filter_date
    if @filters[:date].is_a?(Hash) then
      date = get_filter_date_param(@filters[:date_begin], @filters[:date_end], @filters[:date_outside_period])
      if @filters[:date_target] then
        case @filters[:date_target]
        when 'detected_date' then
          @filters[:date][:detected_date] = date
        when 'created_date' then
          @filters[:date][:created_date] = date
        when 'updated_date' then
          @filters[:date][:updated_date] = date
        end
      else
        @filters[:date][:detected_date] = date #デフォルトは検出日
      end
    elsif @filters[:date].is_a?(String) then
      @filters[:date] = get_filter_date_param(@filters[:date_begin], @filters[:date_end], @filters[:date_outside_period])
    end
  end

  def set_filter_journal
    if @filters[:journal].is_a?(Hash) then
      @filters[:journal][:period] = get_filter_date_param(@filters[:journal_date_begin], @filters[:journal_date_end], @filters[:journal_outside_period])
      if @filters[:journal_status_change].to_s == 'true' then
        @filters[:journal][:filter_status_change] = true
      end
      if @filters[:journal_duedate_change].to_s == 'true' then
        @filters[:journal][:filter_duedate_change] = true
      end
    end
  end

  public

  # 導通確認
  # 認証不要で、常に success: true を返す。
  def index
    res = create_response
    res["success"] = true
    render :json => res
  end

  # アクセスキー認証チェック
  def auth
    res = create_response
    res["login"] = User.current.login
    res["allow_update"] = @apikey.allow_update
    res["success"] = true
    render :json => res
  end

  # SIDfmVMバージョン取得
  def version
    res = create_response
    res["version"] = @vm_version
    res["success"] = true
    render :json => res
  end

  # カテゴリ一覧
  def categories
    res = create_response
    begin
      content = SidfmdCategory
        .order("CASE WHEN is_system IS TRUE THEN id ELSE -1 END")
        .order('name COLLATE "C"')
      unless @role_admin then
        content = content.where(:id => @user_category_ids)
      end
      if @filters['name'] then
        if @filters['name'].is_a?(Array) then
          @filters['name'].each do |i|
            content = content.where("name ILIKE ?", "%#{ActiveRecord::Base.send(:sanitize_sql_like, i.to_s)}%")
          end
        else
          content = content.where("name ILIKE ?", "%#{ActiveRecord::Base.send(:sanitize_sql_like, @filters['name'].to_s)}%")
        end
      end
      if @filters['is_system'] then
        content = content.where(:is_system => ("true" == @filters['is_system']))
      end
      if @filters['id'] then
        if @filters['id'].is_a?(Array) then
          content = content.where(:id => @filters['id'].map(&:to_i).uniq)
        else
          content = content.where(:id => @filters['id'].to_i)
        end
      end
      res["total_size"] = content.count
      content = content.offset(@offset).limit(@per_page)
      res["categories"] = content.collect {|x| get_category_data(x)}
      res["success"] = true
    rescue Exception => e
      res["success"] = false
      res["response"]["message"] = e.message
      logger.error(e.backtrace)
    end
    render :json => res
  end

  # ユーザ一覧
  def users
    res = create_response
    begin
      content = User.create_all_content(
        :sidfm_roles => @sidfm_roles,
        :current_user_id => User.current.id,
        :user_category_ids => @user_category_ids,
        :sidfm_project_id => @sidfm_project.id
      )
      # フィルタ
      filters = {}
      @filters.each do |key, value|
        key = key.to_sym
        case key
        when :login, :name then
          filters[key] = value
        when :auth then
          if value == 'password' then
            filters[key] = -1.to_s
          elsif value == 'openidconnect' then
            auth = AuthSource.find_by_name(OPENID_CONNECT_AUTH_NAME)
            filters[key] = auth.id.to_s if auth
          else
            auth = AuthSource.find_by_name(value)
            filters[key] = (auth ? auth.id : -2).to_s #存在しない場合は何も該当させない
          end
        when :role then
          role = Role.joins(:member_roles).joins("
            INNER JOIN members ON members.id = member_roles.member_id
          ")
          .where(:members => { :project_id => @sidfm_project.id })
          .where("roles.name = ?", value)
          .first
          filters[key] = (role ? role.id : -2).to_s #存在しない場合は何も該当させない
        when :category then
          category = SidfmdCategory.find_by_name(value)
          filters[key] = (category ? category.id : -2).to_s #存在しない場合は何も該当させない
        when :category_group then
          if value[:id] then
            filters[:group] = value[:id].to_i.to_s
          elsif value[:no_group].to_s == 'true' then
            filters[:group] = -1.to_s
          end
        when :email_enable then
          filters[:mail_notification] = value.to_s
        when :lock_enable then
          filters[:locked] = value.to_s
        when :login_date_begin then
          filters[:last_login_on] = [] unless filters[:last_login_on]
          filters[:last_login_on][0] = value.to_s
          filters[:last_login_on][1] = '' unless filters[:last_login_on][1]
        when :login_date_end then
          filters[:last_login_on] = [] unless filters[:last_login_on]
          filters[:last_login_on][0] = '' unless filters[:last_login_on][0]
          filters[:last_login_on][1] = value.to_s
        when :login_date_nologin then
          if value.to_s == 'true' then
            filters[:last_login_on] = [] unless filters[:last_login_on]
            filters[:last_login_on][0] = 'not-login-users'
          end
        end
      end
      content = filter_users_content(
        :content => content,
        :filters => filters,
        :sidfm_project_id => @sidfm_project.id,
        :current_user => User.current,
        :sidfm_roles => @sidfm_roles,
      )

      # ソート
      if @api_params[:order] == 'auth_asc' || @api_params[:order] == 'auth_desc' then
        @api_params[:order] = @api_params[:order].sub('auth', 'auth_source_id')
      end
      content = sort_users_content(
        :content => content,
        :order => @api_params[:order] || 'login_asc',
        :sidfm_roles => @sidfm_roles
      )
   
      # データ
      res["total_size"] = content.count
      content = content.offset(@offset).limit(@per_page)
      current_user_group_ids = User.current.sidfmd_cat_groups.pluck(:id)
      res["users"] = content.collect {|user|
        auth = user.auth_source
        role = user.roles.where(:members => { :project_id => @sidfm_project.id }).first
        data = {
          :id => user.id,
          :login => user.login,
          :firstname => user.firstname,
          :lastname => user.lastname,
          :mail_address => user.mail,
          :last_login_on => user.last_login_on ? user.last_login_on.strftime('%Y/%m/%d %H:%M:%S') : '',
          :auth => auth ? auth.name : 'パスワード認証',
          :role => role ? role.name : '',
          :email_enable => user.mail_notification == "only_my_events",
          :categories => (
            @role_admin ?
            user.sidfmd_categories :
            user.sidfmd_categories.where(:id => @user_category_ids)
          ).collect {|c| get_category_data(c) }
        }
        if @sidfm_roles.include?(:user_admin) then
          data[:must_change_password] = user.must_change_passwd
          data[:lock_enable] = user.status == User::STATUS_LOCKED 
          data[:category_groups] = (
            @role_admin ?
            user.sidfmd_cat_groups :
            user.sidfmd_cat_groups.where(:id => current_user_group_ids)
          ).collect {|grp| { :id => grp.id, :name => grp.name } }
        end
        data
      }
      res["success"] = true
    rescue Exception => e
      res["success"] = false
      res["response"]["message"] = e.message
      logger.error(e.backtrace)
    end
    render :json => res
    
  end

  # ホスト一覧
  def hosts
    params_fetched = params.fetch(:q, {}).permit(:index, :prior_category, :category)
    res = create_response
    begin
      status_names = SidfmdIssue::STATUS_NAMES
      category = params_fetched[:category] ? SidfmdCategory.find_by_name(params_fetched[:category]) : nil
      category_id = category ? category.id : 0
      if !@role_admin && category && !category.is_system && !@user_category_ids.include?(category_id) then
        category_id = -1 #範囲外のカテゴリのデータは常になし
      end
      index = (params_fetched[:index] || 'sri').to_sym
      prior_category = params_fetched[:prior_category] || 'system'
      content = SidfmdHostView.create_all_content(
        :role_admin => @role_admin,
        :user_host_ids => @user_host_ids,
        :user_category_ids => @user_category_ids,
        :category_id => category_id
      )
      @filters = @filters.deep_symbolize_keys
      default_hosts_filters(hosts_dist_versions(content, true)).each do |key, value|
        @filters[key] = value[:default] unless @filters.has_key?(key)
      end
      convert_filter_category_id_type_into_int
      convert_filter_product_name_into_id
      convert_filter_status_name_into_id
      convert_filter_sri_name_into_id
      convert_filter_area_name_into_id
      convert_filter_category_name_into_id
      set_filter_duedate
      set_filter_date
      # フィルタ
      content = filter_hosts_content(
        :content => content,
        :filters => @filters,
        :role_admin => @role_admin,
        :user_category_ids => @user_category_ids,
        :user_host_ids => @user_host_ids,
        :prior_category => prior_category,
        :index => index
      )
      # ソート
      content = sort_hosts_content(
        :content => content,
        :order => @api_params[:order] || 'hostname_asc',
        :role_admin => @role_admin,
        :user_category_ids => @user_category_ids,
        :prior_category => prior_category,
        :index => index
      )

      # データ
      areas = SidfmdArea.all.map{|area| [ area.area_id, area.name ]}.to_h
      res["total_size"] = content.count
      content = content.offset(@offset).limit(@per_page)
      res["hosts"] = content.collect {|host|
        host_categories = SidfmdCategory.joins(:sidfmd_categories_hosts)
          .where(:sidfmd_categories_hosts => { :sidfmd_host_id => host.sidfmd_host_id })
          .order("CASE WHEN is_system IS TRUE THEN sidfmd_categories.id ELSE #{prior_category == 'custom' ? '0' : 'NULL'} END")
          .order('LOWER(name COLLATE "C")')
        unless @role_admin then
          host_categories = host_categories.where("sidfmd_categories.id IN (?) OR sidfmd_categories.is_system IS TRUE", @user_category_ids)
        end
        data = {
          :id => host.sidfmd_host_id,
          :hostname => host.hostname,
          :ext_name => host.ext_name,
          :display_name => host.display_name,
          :url => get_host_detail_link(host.sidfmd_host_id),
          :warning => host[:"warning_#{index.to_s}"],
          :min_status => status_names[host.min_status_id],
          :max_sri => SidfmdIssueSri::RISKS[host.max_sri],
          :max_cvss_score => (host.max_cvss_label.to_s == 'notreg' ? nil : sprintf('%.1f', (host.max_cvss_score ? host.max_cvss_score : 0))),
          :area => areas[host.max_area_id],
          :exploitable => host.max_cvss_label.to_s == 'notreg' ? nil : host.exploitable,
          :categories => host_categories.collect {|c| get_category_data(c) },
          :os => host.osname,
          :duedates_percent => host.duedates_percent ? "#{sprintf('%.1f', host.duedates_percent)}%" : '',
          :finished_percent => host.fixed_percent ? "#{sprintf('%.1f', host.fixed_percent)}%" : '',
        }
      }
      res["success"] = true
    rescue Exception => e
      res["success"] = false
      res["response"]["message"] = e.message
      logger.error(e.backtrace)
    end
    render :json => res
  end

  # ホスト詳細の一覧
  def vulnerabilities_by_host
    params_fetched = params.fetch(:q, {}).permit(:index)
    res = create_response
    begin
      host_id = @api_params[:host_id].to_i
      unless SidfmdHost.where(:id => host_id).exists? then
        res['response']['message'] = 'not found.'
        render :json => res and return
      end
      if !@role_admin && !@user_host_ids.include?(host_id) then
        res['response']['message'] = 'forbidden access.'
        render :json => res and return
      end
      status_names = SidfmdIssue::STATUS_NAMES
      index = (params_fetched[:index] || 'sri').to_sym
      corresp_required = Sidfm.corresp_required
      content = SidfmdVulView.create_host_content(
        :role_admin => @role_admin,
        :user_host_ids => @user_host_ids,
        :host_id => host_id
      )

      # フィルタ
      @filters = @filters.deep_symbolize_keys
      default_host_filters.each do |key, value|
        @filters[key] = value[:default] unless @filters.has_key?(key)
        if key == :cvss_vector then
          @filters[key][:cvss_v2] = value[:default][:cvss_v2] unless @filters[key][:cvss_v2]
          @filters[key][:cvss_v3] = value[:default][:cvss_v3] unless @filters[key][:cvss_v3]
        end
      end
      set_filter_date
      convert_filter_product_name_into_id
      convert_filter_status_name_into_id
      convert_filter_sri_name_into_id
      convert_filter_area_name_into_id
      convert_filter_ctag_name_into_id
      set_filter_duedate
      set_filter_journal
      content = filter_host_content(
        :content => content,
        :filters => @filters,
        :role_admin => @role_admin,
        :user_category_ids => @user_category_ids,
        :user_host_ids => @user_host_ids,
        :host_id => host_id,
        :index => index,
        :corresp_required => corresp_required
      )
      # ソート
      content = sort_host_content(
        :content => content,
        :order => @api_params[:order] || 'title_asc',
        :host_id => host_id,
        :index => index,
        :corresp_required => corresp_required
      )

      # データ
      areas = SidfmdArea.all.map{|area| [ area.area_id, area.name ]}.to_h
      htmlid_dates = SidfmdHtmlidDate.create_htmlid_create_update_time(content.pluck(:html_id))
      res["total_size"] = content.count
      content = content.offset(@offset).limit(@per_page)

      vul_ctags = get_vul_ctags(content.pluck(:id), host_id)
      vul_issues = SidfmdIssue.where(
        :sidfmd_host_id => host_id, :html_id => content.pluck(:html_id)
      ).map{|issue| [ issue.html_id, issue ]}.to_h
      vul_scsids = SidfmdVvScsid.where(:sidfmd_host_id => host_id)
        .map{|scsid| [ scsid.sidfmd_vul_view_id, scsid ]}.to_h
       
      res["vulnerabilities"] = []
      content.each do |vul|
        issue = vul_issues[vul.html_id]
        host_vul_scsid = vul_scsids[vul.id]
        next if !issue || !host_vul_scsid
        cvss_label = SidfmdIssueCve.labelling(host_vul_scsid.cvss).to_s
        elapsed_created_on = host_vul_scsid.created_on ? host_vul_scsid.created_on : Time.current
        cves = vul.get_summary_cves
        cvss_2_cves = []
        cvss_3_cves = []
        cve_ids = []
        cves.each do |cve|
          cve_ids << cve.cve_id
          cvss_2_cves << { :id => cve.cve_id, :vector => cve.c2_vector_index }
          cvss_3_cves << { :id => cve.cve_id, :vector => cve.c3_vector_index }
        end
        journals = []
        Journal.where(:journalized_id => issue.issue_id).order(:created_on => :desc).each do |journal|
          j_item = {}
          j_item[:id] = journal[:id]
          j_item[:created_on] = journal[:created_on] ? journal[:created_on].strftime('%Y/%m/%d %H:%M:%S') : ''
          user = User.find(journal[:user_id])
          j_item[:user] = {
            :id => journal[:user_id],
            :fullname => user.anonymous? ? '削除済みユーザ' : (user ? "#{user.lastname} #{user.firstname}" : '（存在しないユーザ）'),
            :mail_address => user.anonymous? ? nil : user.mail
          }
          j_item[:notes] = Journal.find(journal[:id]).notes
          j_item[:details] = []
          JournalDetail.where(:journal_id => journal[:id]).order(:id).each do |detail|
            jd_item = {}
            jd_item[:id] = detail[:id]
            case detail[:prop_key]
            when 'status_id' then
              jd_item[:action] = 'ステータス変更'
              jd_item[:from] = SidfmdIssue.status_id_to_s(detail[:old_value].to_i)
              jd_item[:to] = SidfmdIssue.status_id_to_s(detail[:value].to_i)
            when 'due_date' then
              jd_item[:action] = '対策予定日の変更'
              jd_item[:from] = SidfmdIssue.due_date_to_s(detail[:old_value])
              jd_item[:to] = SidfmdIssue.due_date_to_s(detail[:value])
            else
              next
            end
            j_item[:details] << jd_item
          end
          journals << j_item
        end
        create_date = (htmlid_dates[vul.html_id] && htmlid_dates[vul.html_id][:create_datetime]) ? htmlid_dates[vul.html_id][:create_datetime].strftime('%Y/%m/%d') : ''
        update_date = (htmlid_dates[vul.html_id] && htmlid_dates[vul.html_id][:update_datetime]) ? htmlid_dates[vul.html_id][:update_datetime].strftime('%Y/%m/%d') : ''
        res["vulnerabilities"] << {
          :html_id => vul.html_id,
          :title => SidfmdVulView.display_title(vul.title),
          :short_description => vul.short_description,
          :create_date => create_date,
          :update_date => create_date == update_date ? '' : update_date,
          :url => get_vul_detail_link(vul.html_id),
          :information_detail_url => get_vul_infomation_link(vul.html_id),
          :warning => if SidfmdIssue::STATUS_UNFINISHED.include?(host_vul_scsid.status_id) then
                        if index == :sri then
                          if host_vul_scsid.sri >= corresp_required[:sri] then
                            true
                          else
                            false
                          end
                        else
                          if host_vul_scsid.cvss.nil? || host_vul_scsid.cvss >= corresp_required[:cvss] then
                            true
                          else
                            false
                          end
                         end
                       else
                         false
                       end,
           :status => status_names[host_vul_scsid.status_id],
           :sri => SidfmdIssueSri::RISKS[host_vul_scsid.sri],
           :cvss_score => (cvss_label == 'notreg' ? nil : sprintf('%.1f', (host_vul_scsid.cvss ? host_vul_scsid.cvss : 0).to_f)),
           :area => areas[vul.softek_area_id],
           :exploitable => (cvss_label != 'notreg' and SidfmdIssue::STATUS_UNFINISHED.include?(host_vul_scsid.status_id)) ? vul.exploitable : nil,
           :cve_ids => cve_ids,
           :cvss_v2 => cvss_2_cves,
           :cvss_v3 => cvss_3_cves,
           :elapsed_days => (SidfmdIssue::STATUS_UNFINISHED.include?(host_vul_scsid.status_id) ? (((Time.current - elapsed_created_on) / 24.0 / 60.0 / 60.0).to_i) : nil),
           :created_on => elapsed_created_on.strftime('%Y/%m/%d'),
           :due_date => (host_vul_scsid.due_date ? host_vul_scsid.due_date.strftime('%Y/%m/%d') : ''),
           :journals => journals,
           :affect_packages => issue.affect_packages(''),
           :affect_products => issue.affect_products + issue.affect_lproducts,
           :ctags => vul_ctags[vul.id] || [],   
        }
      end
      res["success"] = true
    rescue Exception => e
      res["success"] = false
      res["response"]["message"] = e.message
      logger.error(e.backtrace)
    end
    render :json => res
  end

  # 脆弱性一覧
  def vulnerabilities
    params_fetched = params.fetch(:q, {}).permit(:index, :prior_category, :category)
    res = create_response
    begin
      status_names = SidfmdIssue::STATUS_NAMES
      category = params_fetched[:category] ? SidfmdCategory.find_by_name(params_fetched[:category]) : nil
      category_id = category ? category.id : 0
      if !@role_admin && category && !category.is_system && !@user_category_ids.include?(category_id) then
        category_id = -1 #範囲外のカテゴリのデータは常になし
      end
      index = (params_fetched[:index] || 'sri').to_sym
      prior_category = params_fetched[:prior_category] || 'system'
      corresp_required = Sidfm.corresp_required
      content = SidfmdVulView.create_all_content(
        :role_admin => @role_admin,
        :user_host_ids => @user_host_ids,
        :category_id => category_id
      )
      filter_category_host_ids =
        category_id > 0 ?
          SidfmdCategoriesHost.category_host_ids(
            :category_id => category_id,
            :role_admin => @role_admin,
            :user_host_ids => @user_host_ids
          ) :
          nil
      @filters = @filters.deep_symbolize_keys
      default_vuls_filters.each do |key, value|
        @filters[key] = value[:default] unless @filters.has_key?(key)
        if key == :cvss_vector then
          @filters[key][:cvss_v2] = value[:default][:cvss_v2] unless @filters[key][:cvss_v2]
          @filters[key][:cvss_v3] = value[:default][:cvss_v3] unless @filters[key][:cvss_v3]
        end
      end
      set_filter_date
      convert_filter_category_id_type_into_int
      convert_filter_product_name_into_id
      convert_filter_status_name_into_id
      convert_filter_sri_name_into_id
      convert_filter_area_name_into_id
      convert_filter_category_name_into_id
      convert_filter_ctag_name_into_id
      set_filter_duedate
      set_filter_date
      # フィルタ
      content = filter_vuls_content(
        :content => content,
        :filters => @filters,
        :role_admin => @role_admin,
        :user_category_ids => @user_category_ids,
        :user_host_ids => @user_host_ids,
        :category_host_ids => filter_category_host_ids,
        :index => index,
        :corresp_required => corresp_required
      )
      # ソート
      content = sort_vuls_content(
        :content => content,
        :role_admin => @role_admin,
        :user_host_ids => @user_host_ids,
        :user_category_ids => @user_category_ids,
        :order => @api_params[:order] || 'title_asc',
        :index => index,
        :corresp_required => corresp_required,
        :category_host_ids => filter_category_host_ids,
        :prior_category => prior_category
      )
      # データ
      res["total_size"] = content.count
      content = content.offset(@offset).limit(@per_page)

      #vul_ctags = get_vul_ctags(content.pluck(:id))
      areas = SidfmdArea.all.map{|area| [ area.area_id, area.name ]}.to_h
      status_names = SidfmdIssue::STATUS_NAMES
      res["vulnerabilities"] = []
      content.each do |vul|
        vul_values = vul.create_summary_values(@role_admin, category_id, @user_category_ids, @user_host_ids, corresp_required)
        next unless vul_values
        item = {}
        item[:html_id] = vul_values[:html_id] > 0 ? "#{vul_values[:html_id]}" : "KB#{-vul_values[:html_id]}"
        item[:title] = vul_values[:title]
        item[:short_description] = (vul_values[:short_description] ? vul_values[:short_description] : '')
        item[:url] = get_vul_detail_link(vul_values[:html_id])
        item[:information_detail_url] = get_vul_infomation_link(vul_values[:html_id])
        item[:warning] = vul_values[:"warning_#{index}"]
        count_status_ids = SidfmdIssue.count_status_ids(vul_values[:status_ids])
        item[:statuses] = count_status_ids
          .sort {|x, y| x[:id] <=> y[:id] }
          .collect {|x| { :name => x[:name], :value => x[:value] } }
        item[:max_sri] = SidfmdIssueSri::RISKS[vul_values[:max_sri]]
        item[:max_cvss] = (vul_values[:cvss_label] == 'notreg' ? nil : sprintf('%.1f', (vul_values[:cvss_score] ? vul_values[:cvss_score] : 0).to_f))
        cves = vul.get_summary_cves
        cvss_2_cves = []
        cvss_3_cves = []
        cve_ids = []
        cves.each do |cve|
          cve_ids << cve.cve_id
          cvss_2_cves << { :id => cve.cve_id, :vector => cve.c2_vector_index }
          cvss_3_cves << { :id => cve.cve_id, :vector => cve.c3_vector_index }
        end
        item[:cve_ids] = cve_ids
        item[:cvss_v2] = cvss_2_cves
        item[:cvss_v3] = cvss_3_cves
        item[:max_area] = areas[vul_values[:softek_area_id].to_i]
        item[:exploitable] = vul_values[:exploitable]
        item[:min_elapsed_days] = (vul_values[:min_elapsed_created_on_display] >= 0 ? vul_values[:min_elapsed_created_on_display] : nil)
        item[:min_created_on] = vul_values[:min_created_on_display]
        item[:hosts] = vul_values[:hosts].map{|host|
          {
            :id => host.sidfmd_host_id,
            :hostname => host.hostname,
            :ext_name => host.ext_name,
            :display_name => host.display_name,
            :url => get_host_detail_link(host.sidfmd_host_id)
          }
        }
        item[:categories] = SidfmdCategory.where(
          "id IN (?) OR id IN (?)",
          vul_values[:custom_categories].select(:sidfmd_category_id),
          vul_values[:system_categories].select(:sidfmd_category_id)
        )
        .order("CASE WHEN is_system IS TRUE THEN id ELSE #{prior_category == 'custom' ? '0' : 'NULL'} END")
        .order('LOWER(name COLLATE "C")')
        .collect {|c| get_category_data(c) }

        item[:due_dates_percent] = display_percent(vul_values[:duedates_percent])
        item[:statuses_percent] = display_percent(vul_values[:fixed_percent])
        item[:ctags] = vul_values[:ctags]#vul_ctags[vul_values[:id]] || []
        item[:create_date] = vul_values[:create_date] || ''
        item[:update_date] = vul_values[:update_date] || ''
        item[:update_date] = '' if item[:update_date] == item[:create_date]
        res["vulnerabilities"] << item
      end
      res["success"] = true 
    rescue Exception => e
      res["success"] = false
      res["response"]["message"] = e.message
      logger.error(e.backtrace)
    end
    render :json => res
  end

  # 脆弱性詳細の一覧
  def hosts_by_vulnerability
    params_fetched = params.fetch(:q, {}).permit(:index, :prior_category, :category)
    res = create_response
    begin
      html_id = @api_params[:html_id].to_i
      unless SidfmdIssue.where(:html_id => html_id).exists? then
        res['response']['message'] = 'not found.'
        render :json => res and return
      end
      if !@role_admin && !SidfmdIssue.where(:sidfmd_host_id => @user_host_ids).where(:html_id => html_id).exists? then
        res['response']['message'] = 'forbidden access.'
        render :json => res and return
      end
      category = params_fetched[:category] ? SidfmdCategory.find_by_name(params_fetched[:category]) : nil
      category_id = category ? category.id : 0
      if !@role_admin && category && !category.is_system && !@user_category_ids.include?(category_id) then
        category_id = -1 #範囲外のカテゴリのデータは常になし
      end
      index = (params_fetched[:index] || 'sri').to_sym
      prior_category = params_fetched[:prior_category] || 'system'
      corresp_required = Sidfm.corresp_required

      vul_view = SidfmdVulView.find_by(:html_id => html_id)
     
      content = SidfmdHostView.create_vul_content(
        :role_admin => @role_admin,
        :user_host_ids => @user_host_ids,
        :vul_view_id => vul_view ? vul_view.id : -1,
        :category_id => category_id
      )

      # フィルタ
      if vul_view then
        @filters = @filters.deep_symbolize_keys
        default_vul_filters(vul_dist_versions(content)).each do |key, value|
          @filters[key] = value[:default] unless @filters.has_key?(key)
          if key == :cvss_vector then
            @filters[key][:cvss_v2] = value[:default][:cvss_v2] unless @filters[key][:cvss_v2]
            @filters[key][:cvss_v3] = value[:default][:cvss_v3] unless @filters[key][:cvss_v3]
          end
        end
        convert_filter_category_id_type_into_int
        convert_filter_product_name_into_id
        convert_filter_status_name_into_id
        convert_filter_sri_name_into_id
        convert_filter_category_name_into_id
        convert_filter_ctag_name_into_id
        set_filter_duedate
        set_filter_date
        set_filter_journal
        content = filter_vul_content(
          :content => content,
          :filters => @filters,
          :vul_view => vul_view,
          :role_admin => @role_admin,
          :user_category_ids => @user_category_ids,
          :index => index,
          :corresp_required => corresp_required
        )
  
        # ソート
        content = sort_vul_content(
          :content => content,
          :order => @api_params[:order] || 'hostname_asc',
          :vul_view => vul_view,
          :role_admin => @role_admin,
          :user_category_ids => @user_category_ids,
          :prior_category => prior_category,
          :index => index,
          :corresp_required => corresp_required
        )
      end

      # データ
      res["total_size"] = content.count
      content = content.offset(@offset).limit(@per_page)

      areas = SidfmdArea.all.map{|area| [ area.area_id, area.name ]}.to_h
      res["hosts"] = []
      if vul_view then
        scsids = SidfmdVvScsid.where(:sidfmd_host_id => content.select(:sidfmd_host_id))
                              .where(:sidfmd_vul_view_id => vul_view.id)
                              .map{|scsid| [ scsid.sidfmd_host_id, scsid ]}
                              .to_h
        host_ctags = get_host_ctags(content.pluck(:id), html_id)

        content.each do |host|
          issue = SidfmdIssue.where(:sidfmd_host_id => host.sidfmd_host_id, :html_id => vul_view.html_id).first
          host_scsid = scsids[host.sidfmd_host_id]
          next if !issue or !host_scsid
          item = {}
          ctags = (host_ctags[host.id] || [])
          item[:id] = host[:sidfmd_host_id]
          item[:hostname] = host[:hostname]
          item[:ext_name] = host[:ext_name]
          item[:display_name] = host[:display_name]
          item[:url] = get_host_detail_link(host.sidfmd_host_id)
          item[:warning] = if SidfmdIssue::STATUS_UNFINISHED.include?(host_scsid.status_id) then
                             if index == :sri then
                               if host_scsid.sri >= corresp_required[:sri] then
                                 true
                               else
                                 false
                               end
                             else
                               if host_scsid.cvss.nil? || host_scsid.cvss >= corresp_required[:cvss] then
                                 true
                               else
                                 false
                               end
                             end
                           else
                             false
                           end
          item[:status] = SidfmdIssue.status_id_to_s(host_scsid[:status_id])
          item[:sri] = SidfmdIssueSri::RISKS[host_scsid[:sri]]
          cvss_label = SidfmdIssueCve.labelling(host_scsid[:cvss]).to_s
          item[:cvss] = (cvss_label == 'notreg' ? nil : sprintf('%.1f', (host_scsid[:cvss] ? host_scsid[:cvss] : 0).to_f))
          item[:area] = areas[vul_view[:softek_area_id]]
          item[:exploitable] = (cvss_label != 'notreg' and SidfmdIssue::STATUS_UNFINISHED.include?(host_scsid[:status_id])) ? vul_view[:exploitable] : nil
          elapsed_created_on = host_scsid[:created_on] || Time.current
          item[:elapsed_days] = SidfmdIssue::STATUS_UNFINISHED.include?(host_scsid[:status_id]) ?
            (((Time.current - elapsed_created_on) / 24.0 / 60.0 / 60.0).to_i) :
            nil
          item[:created_on] = host_scsid[:created_on] ? host_scsid[:created_on].localtime.strftime('%Y/%m/%d') : ''
          ns_cats = (
            @role_admin ?
            host.sidfmd_hv_categories :
            host.sidfmd_hv_categories.where(:sidfmd_category_id => @user_category_ids)
          )
          item[:categories] = SidfmdCategory
            .where("id IN (?) OR id = ?", ns_cats.select(:sidfmd_category_id), host[:system_category_id])
            .order("CASE WHEN is_system IS TRUE THEN id ELSE #{prior_category == 'custom' ? '0' : 'NULL'} END")
            .order('LOWER(name COLLATE "C")')
            .collect {|c| get_category_data(c) }
          item[:os] = ((host[:osname] and host[:osname].size > 0) ? host[:osname] : '')
          item[:journals] =  []
          Journal.where(:journalized_id => issue.issue_id).order(:created_on => :desc).each do |journal|
            j_item = {}
            j_item[:id] = journal[:id]
            j_item[:created_on]  = journal[:created_on] ? journal[:created_on].strftime('%Y/%m/%d %H:%M:%S') : ''
            user = User.find(journal[:user_id])
            j_item[:user] = {
              :id => journal[:user_id],
              :fullname => user.anonymous? ? '削除済みユーザ' : (user ? "#{user.lastname} #{user.firstname}" : '（存在しないユーザ）'),
              :mail_address => user.anonymous? ? nil : user.mail
            }
            j_item[:notes] = Journal.find(journal[:id]).notes
            j_item[:details] = []
            JournalDetail.where(:journal_id => journal[:id]).order(:id).each do |detail|
              jd_item = {}
              jd_item[:id] = detail[:id]
              case detail[:prop_key]
              when 'status_id' then
                jd_item[:action] = 'ステータス変更'
                jd_item[:from] = SidfmdIssue.status_id_to_s(detail[:old_value].to_i)
                jd_item[:to] = SidfmdIssue.status_id_to_s(detail[:value].to_i)
              when 'due_date' then
                jd_item[:action] = '対策予定日の変更'
                jd_item[:from] = SidfmdIssue.due_date_to_s(detail[:old_value])
                jd_item[:to] = SidfmdIssue.due_date_to_s(detail[:value])
              end
              j_item[:details] << jd_item
            end
            item[:journals] << j_item
          end
          item[:due_date] = host_scsid[:due_date] ? host_scsid[:due_date].strftime('%Y/%m/%d') : ''

          item[:affect_packages] = []
          item[:affect_products] = []

          item[:issue] = { :id => issue.id }
          item[:affect_packages] = issue.affect_packages('')
          item[:affect_products] = issue.affect_products + issue.affect_lproducts
          item[:ctags] = ctags
          item[:issue][:title] = issue.title
          item[:issue][:html_id] = issue.html_id > 0 ? "#{issue.html_id}" : "KB#{-issue.html_id}"
          res["hosts"] << item
        end
      end
      res["success"] = true
    rescue Exception => e
      res["success"] = false
      res["response"]["message"] = e.message
      logger.error(e.backtrace)
    end
    render :json => res
  end

  # 履歴一覧
  def journals
    res = create_response
    begin
      content = SidfmdIssue.create_journals_content(
        :role_admin => @role_admin,
        :user_host_ids => @user_host_ids
      )
      # フィルタ
      @filters = @filters.deep_symbolize_keys
      default_histories_filters.each do |key, value|
        @filters[key] = value unless @filters.has_key?(key)
      end
      convert_filter_category_id_type_into_int
      convert_filter_status_name_into_id
      convert_filter_category_name_into_id
      convert_filter_ctag_name_into_id
      convert_filter_role_name_into_id
      convert_filter_user_login_into_id
      convert_filter_user_mail_into_id
      convert_filter_user_name_into_id
      convert_filter_hostname_into_id
      convert_filter_displayname_into_id
      set_filter_duedate
      set_filter_date
      if @filters[:operations] && @filters[:operations][:items] then
        @filters[:opes][:targets] = []
        @filters[:operations][:items].each do |op|
          case op.strip
          when 'ステータス変更' then
            @filters[:opes][:targets] << 'status_id'
          when '対策予定日変更' then
            @filters[:opes][:targets] << 'due_date'
          else
            # invalid
          end
        end
        @filters[:opes][:is_except] = @filters[:operations][:is_except] if @filters[:operations].has_key?(:is_except)
      end
      if @filters[:vulnerabilities] then
        @filters[:vuls] = @filters[:vulnerabilities].to_s
      end
      content = filter_histories_content(
        :content => content,
        :filters => @filters,
        :role_admin => @role_admin,
        :user_category_ids => @user_category_ids
      )
      # ソート
      content = sort_histories_content(
        :content => content,
        :order => @api_params[:order] || 'created_on_desc', 
      )
      # データ
      res["total_size"] = content.pluck(:id).size #countはクエリのエラーになるため、このように処理
      content = content.offset(@offset).limit(@per_page)

      htmlid_dates = SidfmdHtmlidDate.create_htmlid_create_update_time(content.pluck(:html_id))
      issue_ctags = get_issue_ctags(content.pluck(:id))
      status_names = SidfmdIssue::STATUS_NAMES
      
      res["journals"] = content.map{|history|
        item = {}
        item[:journal_id] = history.journal_id
        item[:created_on] = history.journal_created_on.strftime('%Y/%m/%d %H:%M:%S')
        item[:user] = {
          :id => history.user_id,
          :login => history.is_anonymous ? '削除済みユーザ' : history.login_name,
          :fullname => history.is_anonymous ? '' : history.user_fullname,
          :mail_address => history.is_anonymous ? '' : history.email_address,
        }
        item[:operation] = {
          :title => history.operation ? history.operation : '',
          :content => if history.operation == 'ステータス変更' and history.old_value and history.current_value then
                        status_names[history.old_value.to_i] + ' から ' + status_names[history.current_value.to_i] + 'へ'
                      elsif history.operation == '対策予定日の変更' then
                        (history.old_value ? history.old_value : '未設定') + ' から ' + (history.current_value ? history.current_value : '未設定') + 'へ'
                      else
                        ''
                      end
        }
        item[:comment] = history.comment
        create_date = (htmlid_dates[history.html_id] && htmlid_dates[history.html_id][:create_datetime]) ?
          htmlid_dates[history.html_id][:create_datetime].strftime('%Y/%m/%d') :
          ''
        update_date = (htmlid_dates[history.html_id] && htmlid_dates[history.html_id][:update_datetime]) ?
          htmlid_dates[history.html_id][:update_datetime].strftime('%Y/%m/%d') :
          ''
        item[:vulnerability] = {
          :html_id => history.html_id,
          :title => "[#{history.html_id > 0 ? history.html_id : "KB#{-history.html_id}"}] #{history.vul_title}",
          :short_description => history.vul_short_description,
          :url => get_vul_detail_link(history.html_id, { :'form[history]' => [ history.host_id, history.html_id ].to_json }),
          :create_date => create_date,
          :update_date => create_date == update_date ? '' : update_date
        }
        item[:host] = {
          :id => history.host_id,
          :hostname => history.hostname,
          :ext_name => history.ext_name,
          :display_name => history.display_name,
          :url => get_host_detail_link(history.host_id, { :'form[history]' => [ history.html_id, history.host_id ].to_json })
        }
        item[:status] = status_names[history.status_id]
        item[:due_date] = history.due_date ? history.due_date.strftime('%Y/%m/%d') : ''
        item[:ctags] = issue_ctags[history.id] || []
        item
      }
      res["success"] = true
    rescue Exception => e
      res["success"] = false
      res["response"]["message"] = e.message
      logger.error(e.backtrace)
    end
    render :json => res
  end

end
