# -*- coding: utf-8 -*-

class SidfmdRule < ActiveRecord::Base
  belongs_to :sidfmd_category

  has_one :sidfmd_rule_trigger, dependent: :destroy
  has_many :sidfmd_rule_conds, dependent: :destroy
  has_many :sidfmd_rule_actions, dependent: :destroy

  scope :enables, -> { where(:flg_enable => true) }
  scope :filter_category, -> (category_id) { where(:sidfmd_category_id => category_id) }

  #キーワードの変換時とhtmlのタイトル表示に使用する
  PAGE_TITLE = "ルール設定"

  #ルールの最大登録数
  MAXIMUM_LIMIT = 500

  #ルールのpositionの最大値
  MAX_POSITION_NUM = 100000
  
  #ルールのpositionの最小値
  MIN_POSITION_NUM = -100000

  #triggerの一覧。パラメータに使用する。
  RULE_TRIGGER_NAMES = {
      'periodic_hour' => "毎日実行-時間指定",
      'periodic_interval'=> "定期実行-間隔指定",
      'vulsync' => "脆弱性同期時 実行",
  }

  #condの一覧。htmlの要素やパラメータに使用する。
  RULE_COND_NAMES = {
      'issue-update' => '脆弱性の更新',
      'cvss-exploitable' => '攻撃コード指定',
      'sri-above' => 'SRI指定値以上',
      'sri-below' => 'SRI指定値以下',
      'sri-warning' => 'SRI値要対応',
      'cvss-above' => 'CVSS指定値以上',
      'cvss-below' => 'CVSS指定値以下',
      'cvss-warning' => 'CVSS値要対応',
      'cvss-notreg' => 'CVSS値未定義',
      'cvss-vector' => 'CVSS指標',
      'status-match' => 'ステータスの一致',
      'product-match' => 'プロダクト限定',
      'product-unmatch' => 'プロダクト除外',
      'due-date-expired' => '対策予定日超過',
      'due-date-approach' => '対策予定日接近',
      'due-date-undef' => '対策予定日未設定',
      'history-create' => '履歴追加',
      'category-match' => 'カテゴリ限定',
      'category-unmatch' => 'カテゴリ除外',
      'host-match' => 'ホスト限定',
      'host-unmatch' => 'ホスト除外',
      'host-syncing' => '同期中ホスト',
      'timerange' => '期間指定',
      'ctag' => 'カスタムタグ'
  }

  #action一覧。htmlの要素やパラメータに使用する。
  RULE_ACTION_NAMES = {
    'send-mail-user' => 'メール送信(ユーザ)',
    'send-mail-category'=> 'メール送信(カテゴリ)',
    'status-change' => 'ステータス変更',
    'due-date-change' => '対策予定日設定',
    'send-syslog' => 'syslog送信',
    'set-ctag' => 'カスタムタグ設定'
  }

  #操作のsyslog送信で指定可能なログの種類
  SYSLOG_FACILITIES = SyslogRuby::Facility.constants.find_all {|x| /^LOG/ !~ x.to_s }.map(&:to_s)

  #操作のsyslog送信で指定可能なログの深刻度
  SYSLOG_SEVERITIES = SyslogRuby::Severity.constants.find_all {|x| /^(LOG|WARNING)/ !~ x.to_s }.map(&:to_s)
  
  # ログでのTrigger, Cond, Actionの名称
  LOG_TRIGGER = 'trigger'.freeze
  LOG_COND = 'cond'.freeze
  LOG_ACTION = 'action'.freeze
  LOG_KINDS = {
    LOG_TRIGGER => 'タイミング'.freeze,
    LOG_COND => '条件'.freeze,
    LOG_ACTION => '操作'.freeze,
  }.freeze

  # ログの種類
  LOG_INFO = 'info'.freeze
  LOG_WARN = 'warn'.freeze
  LOG_ERR = 'err'.freeze
  LOG_TYPES = {
    LOG_INFO => '情報'.freeze,
    LOG_WARN => '警告'.freeze,
    LOG_ERR => 'エラー'.freeze,
  }.freeze

  # Trigger, Cond, Actionのそれぞれのlogsに保存できるログの最大件数
  # 1つのルールには1つのTriggerと複数のCond, Actionがあるので、合計するとそれなりの
  # 量になる。
  LOG_MAXNUM = 30

  # ログの追加処理
  # Trigger, Cond, Actionで利用
  def self.log_append(logs, type, kind, message)
    ret = nil
    begin
      ret = JSON.parse(logs)
    rescue JSON::ParserError
      ret = []
    end
    ret << { 'type' => type, 'kind' => kind, 'message' => message, 'created_at' => Time.current.iso8601(3) }
    ret.compact!
    ret.sort! {|x, y|
      (y['created_at'] && x['created_at']) ?
      y['created_at'] <=> x['created_at'] :
      (y['created_at'] ? 1 : -1)
    }
    JSON.generate(ret[0, LOG_MAXNUM])
  end

  # 指定した現在の条件に適合するルールのIDの配列を返す。
  def self.match_rule_ids(vulsync: false, periodic: false)
    ret = []
    # 定期実行
    if periodic then
      # hourは、前回実行時刻から20時間以上経過及び
      # 現在時刻がhour時になっていることが条件。
      # 始めての実行時(sidfmd_rule_triggers.last_checked_at=nil)は、後者の条件のみ。
      # current_timestampは、sidfmd_rule_triggers.last_checked_atがwithout time zone
      # だったため、整合性を採るために細工したもの。
      current_timestamp = "'#{Time.current.localtime.strftime('%Y/%m/%d %H:%M:%S')}'::timestamp without time zone"
      SidfmdRuleTpHour.enables
        .where("
          sidfmd_rule_tp_hours.hour IS NOT NULL AND (
            (
              sidfmd_rule_triggers.last_checked_at IS NULL AND
              TO_CHAR(#{current_timestamp}, 'HH24')::integer = sidfmd_rule_tp_hours.hour
            ) OR (
              sidfmd_rule_triggers.last_checked_at IS NOT NULL AND
              (#{current_timestamp} - sidfmd_rule_triggers.last_checked_at) >= '20 hours'::interval AND
              TO_CHAR(#{current_timestamp}, 'HH24')::integer = sidfmd_rule_tp_hours.hour
            )
          )
      ").each do |tph|
        ret << tph.sidfmd_rule_trigger.sidfmd_rule_id
      end
      # intervalは前回実行時間から、interval時間以上経過していることが条件
      # 初めての実行時(sidfmd_rule_triggers.last_checked_atがnil)は、無条件で適合。
      SidfmdRuleTpInterval.enables
        .where("
           (
             sidfmd_rule_tp_intervals.unit IS NOT NULL AND
             sidfmd_rule_tp_intervals.value IS NOT NULL ) AND (
             sidfmd_rule_triggers.last_checked_at IS NULL OR
             (#{current_timestamp} - sidfmd_rule_triggers.last_checked_at) >=
              (sidfmd_rule_tp_intervals.value::text || ' ' || sidfmd_rule_tp_intervals.unit || 's')::interval
           )
      ").each do |tpi|
        ret << tpi.sidfmd_rule_trigger.sidfmd_rule_id
      end
    end
    # 脆弱性同期時
    if vulsync then
      # vulsyncは無条件で適合。
      SidfmdRuleTVulsync.enables.each do |i|
        ret << i.sidfmd_rule_trigger.sidfmd_rule_id
      end
    end
    ### FIXME: 有効なruleのtriggerについて、logsに成否の結果を記録。
    #実行優先度順にソート。優先度実装前のルールと優先度設定なしのルールは、positionが全て0で設定されている
    ret = SidfmdRule.where(:id => ret).order(:position => :desc).pluck(:id)
  end

  # ルールのTrigger, Cond, Actionのログをまとめて時系列でソートし返す。
  # 以下のHashの配列を返す。
  # {
  #   'type'       LOG_TYPES.keysのいずれか。LOG_TYPESで日本語名称を取得して出力すること。
  #   'kind'       LOG_KINDS.keysのいずれか。LOG_KINDSで日本語名称を朱徳して出力すること。
  #   'message'    ログ内容
  #   'created_at' 記録日時
  # }
  def logs(maxlogs = 200)
    ret = []
    if self.sidfmd_rule_trigger then
      begin
        r = JSON.parse(self.sidfmd_rule_trigger.logs)
        r.each do |i|
          ret << i.merge('kind' => SidfmdRule::LOG_TRIGGER)
        end
      rescue JSON::ParserError => e
        logger.error(e.message)
        e.backtrace.each do |i|
          logger.error(i)
        end
      end
    end
    self.sidfmd_rule_conds.each do |cond|
      begin
        r = JSON.parse(cond.logs)
        r.each do |i|
          ret << i.merge('kind' => SidfmdRule::LOG_COND)
        end
      rescue JSON::ParserError => e
        logger.error(e.message)
        e.backtrace.each do |i|
          logger.error(i)
        end
      end
    end
    self.sidfmd_rule_actions.each do |action|
      begin
        r = JSON.parse(action.logs)
        r.each do |i|
          ret << i.merge('kind' => SidfmdRule::LOG_ACTION)
        end
      rescue JSON::ParserError => e
        logger.error(e.message)
        e.backtrace.each do |i|
          logger.error(i)
        end
      end
    end
    ret.compact!
    ret.sort! {|x, y|
      (y['created_at'] && x['created_at']) ?
      y['created_at'] <=> x['created_at'] :
      (y['created_at'] ? 1 : -1)
    }
    # 時刻のフォーマットの変更
    ret.each do |i|
      if i['created_at'] then
        i['created_at'].gsub!(/^(\d{4}-\d{2}-\d{2})T(\d{2}:\d{2}:\d{2}\.\d+)[+-]\d{2}:\d{2}$/, '\1 \2')
      end
    end
    ret[0, maxlogs].reverse
  end

  #指定したユーザに関連しているルールの情報を返す。
  #sidfm_users_controller/sidfm_internal_api_controllerで使用
  #関係するのは今のところユーザ削除とカテゴリ変更のみ
  #actionは今のところimport/category_change/bulk_category_change/delete/auth_change/bulk_auth_changeのいずれか
  #:change_category_user_idsにはカテゴリ変更するユーザのidsを、:delete_user_idsには削除するユーザのidsを入れて渡す
  def self.users_related_rule_info(**opts)
    #初めの文字がアンダースコアの変数にしてあるものは再定義しないこと
    user_ids = []
    
    unless opts[:action] then
      raise ArgumentError, "must be action" unless opts.has_key?(:action)
    end

    case opts[:action]
    when 'import' then
      #インポート
      if !opts[:delete_user_ids] or !opts[:change_category_user_ids] then
         raise ArgumentError, "must be delete_user_ids and change_category_user_ids in case of the import"
      end
    when  'delete' then
      #削除
      if !opts[:delete_user_ids] then
        raise ArgumentError, "must be delete_user_ids in case of the delete" 
      end
    when 'category_change', 'bulk_category_change' then
      #カテゴリ(一括)変更
      if !opts[:change_category_user_ids] then
        raise ArgumentError, "must be delete_user_ids in case of the category_change" 
      end
    when 'auth_change', 'bulk_auth_change' then
      #権限(一括)変更
      if !opts[:change_role_user_ids] then
        raise ArgumentError, "must be change_role_user_ids in case of the auth_change"
      end
    else
      #invalid
      raise ArgumentError, "invalid action"
    end

    if opts[:delete_user_ids] then
      opts[:delete_user_ids] = opts[:delete_user_ids].map(&:to_i)
      user_ids |= opts[:delete_user_ids]
    end

    if opts[:change_category_user_ids] then
      user_ids |= opts[:change_category_user_ids].keys.map(&:to_i)
    end

    _category_related_admin_user_ids = opts[:change_category_user_ids] ?
      Member.joins(:roles)
        .where(:roles => { :name => 'SIDfmVM管理者' })
        .where(:members => { :user_id => opts[:change_category_user_ids].keys })
        .pluck('members.user_id')
        .map{|user_id| [ user_id, true ]}
        .to_h :
      {}
      
    if opts[:change_role_user_ids] then
      admin_role = Role.find_by(:name => 'SIDfmVM管理者')
      if admin_role then
        #異常がない限り、このスコープに入る
        admin_role_id = admin_role.id
        filtered_change_role_user_ids = {}
        Member.joins(:member_roles).where(:members => { 
          :user_id => opts[:change_role_user_ids].keys
        })
        .select('members.user_id AS user_id', 'member_roles.role_id AS current_role_id')
        .each do |member|
          update_role_id = opts[:change_role_user_ids][member.user_id.to_s]
          if update_role_id != member.current_role_id then
            if admin_role_id == update_role_id then
              #管理者に変更する場合はカテゴリの情報が必要
              _category_related_admin_user_ids[member.user_id] = true
              opts[:change_category_user_ids] = {} unless opts[:change_category_user_ids]
              if !opts[:change_category_user_ids][member.user_id.to_s] then
                user = User.find_by_id(member.user_id)
                if user then
                  opts[:change_category_user_ids][member.user_id.to_s] = user.sidfmd_categories.pluck(:id) #グループがなくなるので、ここはsidfmd_categoriesのみ
                end
              end
            else
              _category_related_admin_user_ids[member.user_id] = false
            end
            filtered_change_role_user_ids[member.user_id.to_s] = update_role_id
          end
        end
        opts[:change_role_user_ids] = filtered_change_role_user_ids
      else
        logger.error('There is not the role of SIDfmVM管理者 in DB')
        raise Exception
      end
      user_ids |= opts[:change_role_user_ids].keys.map(&:to_i)
    end

    unless opts[:role_admin] then
      unless opts[:user_category_ids] then
        raise ArgumentError, "must be user_category_ids in case the user's role is not admin"
      end
    end

    _rules = if user_ids.any? then
      SidfmdRule.where(%|
        id IN (?) OR id IN (?)
      |,SidfmdRuleAction.joins(%|
          LEFT OUTER JOIN sidfmd_rule_a_mail_categories ON sidfmd_rule_a_mail_categories.sidfmd_rule_action_id = sidfmd_rule_actions.id
        |)
        .joins(%|
          LEFT OUTER JOIN sidfmd_rule_a_mail_users ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id
        |).where(%|
          sidfmd_rule_a_mail_categories.from_user_id IN (?) OR
          sidfmd_rule_a_mail_users.from_user_id IN (?) OR
          sidfmd_rule_a_mail_users.user_id IN (?)
        |, user_ids, user_ids, user_ids)
        .select(:sidfmd_rule_id),#ユーザに関連する操作
        SidfmdRuleCond.joins(%|
          LEFT OUTER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.sidfmd_rule_cond_id = sidfmd_rule_conds.id
        |).joins(%|
          LEFT OUTER JOIN sidfmd_rule_chp_user_subs ON sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id = sidfmd_rule_c_history_posts.id
        |)
        .where(:sidfmd_rule_chp_user_subs => { :user_id => user_ids })
        .select(:sidfmd_rule_id) #ユーザに関連する条件
      )
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_a_mail_users
          INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id
          WHERE sidfmd_rule_id = sidfmd_rules.id
        ) AS count_to_user_ids /*SidfmdRuleAMailUserの送信先に設定されているユーザの数*/
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_actions
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_actions /*ルールのactionの数*/      
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_chp_user_subs
          INNER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.id = sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id
          INNER JOIN sidfmd_rule_conds ON sidfmd_rule_conds.id = sidfmd_rule_c_history_posts.sidfmd_rule_cond_id
          WHERE sidfmd_rule_id = sidfmd_rules.id
        ) AS count_history_cond_user_ids /*SidfmdRuleChpUserSubの数(SidfmdRuleCHistoryPostに設定されているユーザの数)*/
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_conds
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_conds /*ルールのcondの数*/
      ")
      .select(:id, :name, :sidfmd_category_id)
    else
      SidfmdRule.where(:id => 0)
    end

    if opts[:delete_user_ids] then
      opts[:delete_user_ids] = [ 0 ] if opts[:delete_user_ids].size == 0
      _rules = _rules.select(" 
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.user_id IN (SELECT unnest(ARRAY[#{opts[:delete_user_ids].join(',')}]))
          )
        ) AS orig_delete_mail_users_to_actions /*SidfmdRuleAMailUserの送信先に設定されている削除対象のユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.from_user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.from_user_id IN (SELECT unnest(ARRAY[#{opts[:delete_user_ids].join(',')}]))
          )
        ) AS orig_delete_mail_users_from_actions /*SidfmdRuleAMailUserの送信元に設定されている削除対象のユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_categories.from_user_id], ':')
            FROM sidfmd_rule_a_mail_categories
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_categories.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_categories.from_user_id IN (SELECT unnest(ARRAY[#{opts[:delete_user_ids].join(',')}]))
          )
        ) AS orig_delete_mail_categories_from_actions /*SidfmdRuleAMailCategoryの送信元に設定されている削除対象のユーザに関連するactions(通常は1つのみ)*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_chp_user_subs.id, sidfmd_rule_chp_user_subs.user_id, sidfmd_rule_conds.id], ':')
            FROM sidfmd_rule_chp_user_subs
            INNER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.id = sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id
            INNER JOIN sidfmd_rule_conds ON sidfmd_rule_conds.id = sidfmd_rule_c_history_posts.sidfmd_rule_cond_id
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_chp_user_subs.user_id IN (SELECT unnest(ARRAY[#{opts[:delete_user_ids].join(',')}]))
          )
        ) AS orig_delete_history_cond_chp_users /*SidfmdRuleChpUserSubのidとuser_idとsidfmd_rule_cond_id*/
      ") 
    end
    if opts[:change_category_user_ids] then
      target_user_ids = opts[:change_category_user_ids].keys.map(&:to_i)
      target_user_ids = [ 0 ] if target_user_ids.size == 0
      _rules = _rules.select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_category_mail_users_to_actions /*SidfmdRuleAMailUserの送信先に設定されているカテゴリ変更ユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.from_user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.from_user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_category_mail_users_from_actions /*SidfmdRuleAMailUserの送信元に設定されているカテゴリ変更ユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_categories.from_user_id], ':')
            FROM sidfmd_rule_a_mail_categories
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_categories.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_categories.from_user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_category_mail_categories_from_actions /*SidfmdRuleAMailCategoryの送信元に設定されているカテゴリ変更ユーザに関連するactions(通常は1つのみ)*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_chp_user_subs.id, sidfmd_rule_chp_user_subs.user_id, sidfmd_rule_conds.id], ':')
            FROM sidfmd_rule_chp_user_subs
            INNER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.id = sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id
            INNER JOIN sidfmd_rule_conds ON sidfmd_rule_conds.id = sidfmd_rule_c_history_posts.sidfmd_rule_cond_id
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_chp_user_subs.user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_category_history_cond_chp_users /*SidfmdRuleChpUserSubのidとuser_idとsidfmd_rule_cond_id*/
      ")
    end

    if opts[:change_role_user_ids] then
      target_user_ids = opts[:change_role_user_ids].keys.map(&:to_i)
      target_user_ids = [ 0 ] if target_user_ids.size == 0

      _rules = _rules.select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_role_mail_users_to_actions /*SidfmdRuleAMailUserの送信先に設定されている、権限変更されるユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.from_user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.from_user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_role_mail_users_from_actions /*SidfmdRuleAMailUserの送信元に設定されている、権限変更されるユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_categories.from_user_id], ':')
            FROM sidfmd_rule_a_mail_categories
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_categories.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_categories.from_user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_role_mail_categories_from_actions /*SidfmdRuleAMailCategoryの送信元に設定されている、権限変更されるユーザに関連するactions(通常は1つのみ)*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_chp_user_subs.id, sidfmd_rule_chp_user_subs.user_id, sidfmd_rule_conds.id], ':')
            FROM sidfmd_rule_chp_user_subs
            INNER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.id = sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id
            INNER JOIN sidfmd_rule_conds ON sidfmd_rule_conds.id = sidfmd_rule_c_history_posts.sidfmd_rule_cond_id
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_chp_user_subs.user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_role_history_cond_chp_users /*権限が変更されるユーザのSidfmdRuleChpUserSubのidとuser_idとsidfmd_rule_cond_id*/
      ")
    end

    _mail_cat_action_name = RULE_ACTION_NAMES['send-mail-category'];
    _mail_user_action_name = RULE_ACTION_NAMES['send-mail-user'];
    _history_cond_name = RULE_COND_NAMES['history-create']; 
    _rule_messages = {}
    _delete_rules = []
    _delete_actions = []
    _delete_conds = []
    _delete_chp_user_subs = []

    get_user_names = ->(objects){
       User.where(
         :id => objects.map{|action| action['user_id']}
       ).map{|user| "#{user.lastname} #{user.firstname}(ログイン名: #{user.login})"}
    }

    classify_action = ->(
      action, 
      action_name, 
      delete_mail_to_users_actions, 
      delete_mail_from_users_actions , 
      delete_mail_category_actions,
      user_category_ids,
      rule_category_id
    ){
      user_cat_ids = user_category_ids[action['user_id'].to_s]
      if user_cat_ids and (user_cat_ids.size == 0 or (rule_category_id and !user_cat_ids.include?(rule_category_id))) then
        if rule_category_id || !_category_related_admin_user_ids[action['user_id'].to_i] then
          #ルールのカテゴリが全体なら管理者は選択可能
          if action_name == 'mail_to_users' then
            delete_mail_to_users_actions << action
          elsif action_name == 'mail_from_users' then
            delete_mail_from_users_actions << action
          else
            delete_mail_category_actions << action
          end
        end
      end
    }

    classify_chp_user = ->(history_cond_chp_user, rule_delete_chp_user_subs, user_category_ids, rule_category_id){
      user_cat_ids = user_category_ids[history_cond_chp_user["user_id"].to_s]
      if user_cat_ids and (user_cat_ids.size == 0 or (rule_category_id and !user_cat_ids.include?(rule_category_id))) then
        if rule_category_id || !_category_related_admin_user_ids[history_cond_chp_user["user_id"].to_i] then
          #ルールのカテゴリが全体なら管理者は選択可能
          rule_delete_chp_user_subs << history_cond_chp_user
        end
      end
    }

    #ruleのObjectの生成とメソッドの追加
    create_rule_object = ->(rule){
      if rule.respond_to?(:orig_delete_mail_users_to_actions) then
        def rule.delete_mail_users_to_actions
          self.orig_delete_mail_users_to_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_delete_mail_users_from_actions) then
        def rule.delete_mail_users_from_actions
          self.orig_delete_mail_users_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end
   
      if rule.respond_to?(:orig_delete_mail_categories_from_actions) then
        def rule.delete_mail_categories_from_actions
          self.orig_delete_mail_categories_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_delete_history_cond_chp_users) then
        def rule.delete_history_cond_chp_users
          self.orig_delete_history_cond_chp_users.map{|cond|
            cols = cond.split(':')
            {
              'chp_user_sub_id' => cols[0].to_i,
              'user_id' => cols[1].to_i,
              'cond_id' => cols[2].to_i,
            }
          }
        end
      end

      if rule.respond_to?(:orig_change_category_mail_users_to_actions) then
        def rule.change_category_mail_users_to_actions
          self.orig_change_category_mail_users_to_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_change_category_mail_users_from_actions) then
        def rule.change_category_mail_users_from_actions
          self.orig_change_category_mail_users_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_change_category_mail_categories_from_actions) then
        def rule.change_category_mail_categories_from_actions
          self.orig_change_category_mail_categories_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end
     
      if rule.respond_to?(:orig_change_category_history_cond_chp_users) then
        def rule.change_category_history_cond_chp_users
          self.orig_change_category_history_cond_chp_users.map{|cond|
            cols = cond.split(':')
            {
              'chp_user_sub_id' => cols[0].to_i,
              'user_id' => cols[1].to_i,
              'cond_id' => cols[2].to_i,
            }
          }
        end
      end

      if rule.respond_to?(:orig_change_role_mail_users_to_actions) then
        def rule.change_role_mail_users_to_actions
          self.orig_change_role_mail_users_to_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_change_role_mail_users_from_actions) then
        def rule.change_role_mail_users_from_actions
          self.orig_change_role_mail_users_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_change_role_mail_categories_from_actions) then
        def rule.change_role_mail_categories_from_actions
          self.orig_change_role_mail_categories_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_change_role_history_cond_chp_users) then
        def rule.change_role_history_cond_chp_users
          self.orig_change_role_history_cond_chp_users.map{|cond|
            cols = cond.split(':')
            {
              'chp_user_sub_id' => cols[0].to_i,
              'user_id' => cols[1].to_i,
              'cond_id' => cols[2].to_i,
            }
          }
        end
      end

      rule
    }

    get_users_current_category_ids = -> (user_ids){
      User.where(:id => user_ids).map{|user|
        { :user_id => user.id, :category_ids => user.categories.pluck(:id) }
      }
    }

    set_rule_messages_of_actions_related_to_categories = -> (rule, _rule_messages, delete_mail_category_actions = [], delete_mail_from_users_actions = [], delete_mail_to_users_actions = []){
      content_msgs = []
      reason_msgs = []
      if delete_mail_category_actions.any? then
        content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
        reason_msgs << "「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
      end

      if delete_mail_from_users_actions.any? then
        content_msgs << "「#{_mail_user_action_name}」が削除されます。"
        reason_msgs << "「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
      elsif delete_mail_to_users_actions.any? then
        if delete_mail_to_users_actions.size == rule.count_to_user_ids then
          content_msgs << "「#{_mail_user_action_name}」が削除されます。"
          reason_msgs << "「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
        else
          user_names = get_user_names.(delete_mail_to_users_actions)
          if user_names.any? then
            content_msgs << "「#{_mail_user_action_name}」の送信先から、#{user_names.join(', ')}が削除されます。"
            reason_msgs <<  "ユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          end
        end
      end

      if content_msgs.any? then
        if _rule_messages[rule.id] then
          #既に条件のメッセージが設定されている場合。
          #条件のメッセージを先に入れること。
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "条件・操作",
            :content => _rule_messages[rule.id][:content] + "\n" + content_msgs.join("\n"),
            :reason => _rule_messages[rule.id][:reason] + "\n" + reason_msgs.join("\n")
          }
        else
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "操作",
            :content => content_msgs.join("\n"),
            :reason => reason_msgs.join("\n")
          }
        end
      end
    }

    case opts[:action]
    when 'delete' then 
      #ユーザ削除の場合
      _rules.each do |rule|
        rule = create_rule_object.(rule)
        #条件が全てなくなるかのチェック
        rule_delete_conds = []
        if rule.orig_delete_history_cond_chp_users.size > 0 and rule.orig_delete_history_cond_chp_users.size == rule.count_history_cond_user_ids then
          #履歴追加の条件に設定されているユーザが全て削除されるなら条件削除
          rule_delete_conds |= rule.delete_history_cond_chp_users
        end
        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          #条件がなくなるならルールを削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザが削除されるため。"
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end

        #操作が全てなくなるかのチェック
        rule_delete_actions = []
        rule_delete_actions |= rule.delete_mail_users_to_actions
        rule_delete_actions |= rule.delete_mail_users_from_actions
        rule_delete_actions |= rule.delete_mail_categories_from_actions
        if rule.count_actions == rule_delete_actions.uniq{|action| action['action_id']}.size then
          #操作が無くなるルールは削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_actions }
          if opts[:messages_necessary] then
            reason = if rule.count_actions == rule.orig_delete_mail_users_from_actions.size then
                       "ルールの操作が「#{_mail_user_action_name}」のみで、送信元に設定されているユーザが削除されるため。"
                     elsif rule.count_actions == rule.orig_delete_mail_users_to_actions.size then
                       "ルールの操作が「#{_mail_user_action_name}」のみで、送信先に設定されている全てのユーザが削除されるため。"
                     elsif rule.count_actions == rule.orig_delete_mail_categories_from_actions.size then
                       "ルールの操作が「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザが削除されるため。"
                     else
                       "ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、このユーザを削除すると有効な操作がなくなるため。"
                     end
            _rule_messages[rule.id] = {
              :rule_name => rule.name, 
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}#{reason}"
            }
          end
        end
        if _rule_messages[rule.id] or _delete_rules.find{|del_rule| del_rule[:rule_id] == rule.id} then
          next
        end
  
        #条件のチェック
        _delete_conds |= rule_delete_conds
        if rule.orig_delete_history_cond_chp_users.size > 0 then
          if rule_delete_conds.find{|cond| cond['cond_id'] == rule.delete_history_cond_chp_users.first['cond_id']} then
            if opts[:messages_necessary] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件",
                :content => "「#{_history_cond_name}」が削除されます。",
                :reason => "「#{_history_cond_name}」に設定されている全てのユーザが削除されるため。"
              }
            end
          else
            rule_delete_chp_user_subs = rule.delete_history_cond_chp_users
            _delete_chp_user_subs |= rule_delete_chp_user_subs
            if opts[:messages_necessary] then
              user_names = get_user_names.(rule_delete_chp_user_subs)
              if user_names.any? then
                _rule_messages[rule.id] = {
                  :rule_name => rule.name,
                  :item_name => "条件",
                  :content => "「#{_history_cond_name}」から、#{user_names.join(', ')}が削除されます。",
                  :reason => "ユーザが削除されるため。"
                }
              end
            end
          end
        end
  
        #操作のチェック
        _delete_actions |= rule_delete_actions
        if opts[:messages_necessary] then
          content_msgs = []
          reason_msgs = []
          if rule.orig_delete_mail_categories_from_actions.any? then
            #メール送信(カテゴリ)の送信先ユーザが削除されるならメール送信(カテゴリ)が削除されるメッセージを表示
            content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_cat_action_name}」の送信元に設定されているユーザが削除されるため。"
          end
          if rule.orig_delete_mail_users_from_actions.any? then
            #メール送信(ユーザ)の送信元ユーザが削除されるならメール送信(ユーザ)が削除されるメッセージを表示
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_user_action_name}」の送信元に設定されているユーザが削除されるため。"
          elsif rule.orig_delete_mail_users_to_actions.any? then
            if rule.orig_delete_mail_users_to_actions.size == rule.count_to_user_ids then
              #メール送信(ユーザ)の送信先ユーザが全て削除されるならメール送信(ユーザ)が削除されるメッセージを表示
              content_msgs << "「#{_mail_user_action_name}」が削除されます。"
              reason_msgs <<  "「#{_mail_user_action_name}」の送信先に設定されているユーザが全て削除されるため。"
            else
              #メール送信(ユーザ)の送信先ユーザが削除されるメッセージを表示
              user_names = get_user_names.(rule.delete_mail_users_to_actions) 
              if user_names.any? then
                content_msgs << "「#{_mail_user_action_name}」の送信先から、#{user_names.join(', ')}が削除されます。"
                reason_msgs <<  "ユーザが削除されるため。"
              end
            end
          end
          if content_msgs.any? then
            if _rule_messages[rule.id] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件・操作",
                :content => _rule_messages[rule.id][:content] + "\n" + content_msgs.join("\n"),
                :reason => _rule_messages[rule.id][:reason] + "\n" + reason_msgs.join("\n")
              }
            else
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "操作",
                :content => content_msgs.join("\n"),
                :reason => reason_msgs.join("\n")
              }
            end
          end
        end
      end
    when 'category_change', 'bulk_category_change' then
      user_category_ids = opts[:change_category_user_ids] #opts[:change_category_user_ids]のkeyは文字列のユーザid, valueは変更するカテゴリのids
      _rules.each do |rule|
        rule = create_rule_object.(rule)
        rule_delete_conds = []
        #条件がなくなるかのチェック
        rule_delete_chp_user_subs = []
        
        rule.change_category_history_cond_chp_users.each do |chp_user|
          classify_chp_user.(chp_user, rule_delete_chp_user_subs, user_category_ids, rule.sidfmd_category_id)
        end
        if rule_delete_chp_user_subs.size > 0 then
          if rule_delete_chp_user_subs.size == rule.count_history_cond_user_ids then
            #全てのユーザのカテゴリがルールのカテゴリから外れるなら条件削除
            rule_delete_conds |= rule.change_category_history_cond_chp_users
          end   
        end

        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          #全ての条件がなくなるならルール削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end

        #操作が無くなるかのチェック
        delete_mail_to_users_actions = []
        delete_mail_from_users_actions = []
        delete_mail_category_actions = []
        rule.change_category_mail_categories_from_actions.each do |action|
          classify_action.(
            action,
            'mail_category',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_category_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_category_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule.change_category_mail_users_from_actions.each do |action|
          classify_action.(
            action,
            'mail_from_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_from_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_from_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_from_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_from_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule.change_category_mail_users_to_actions.each do |action|
          classify_action.(
            action,
            'mail_to_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_to_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_to_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_to_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_to_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元と「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        if _rule_messages[rule.id] or _delete_rules.find{|del_rule| del_rule[:rule_id] == rule.id} then
          next
        end

        _delete_conds |= rule_delete_conds
        #条件のチェック
        if rule_delete_chp_user_subs.size > 0 then
          if rule_delete_conds.find{|cond| cond['cond_id'] == rule.change_category_history_cond_chp_users.first['cond_id']} then
            if opts[:messages_necessary] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件",
                :content => "「#{_history_cond_name}」が削除されます。",
                :reason => "「#{_history_cond_name}」に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
              }
            end
          else
            _delete_chp_user_subs |= rule_delete_chp_user_subs
            if opts[:messages_necessary] then
              user_names = get_user_names.(rule_delete_chp_user_subs)
              if user_names.any? then
                _rule_messages[rule.id] = {
                  :rule_name => rule.name,
                  :item_name => "条件",
                  :content => "「#{_history_cond_name}」から、#{user_names.join(', ')}が削除されます。",
                  :reason => "ユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                }
              end
            end
          end
        end

        #操作のチェック
        _delete_actions |= (delete_mail_category_actions + delete_mail_to_users_actions + delete_mail_from_users_actions)
        
        if opts[:messages_necessary] then
          set_rule_messages_of_actions_related_to_categories.(rule, _rule_messages, delete_mail_category_actions, delete_mail_from_users_actions, delete_mail_to_users_actions) 
        end
      end
    when 'import' then
      user_category_ids = opts[:change_category_user_ids] #opts[:change_category_user_ids]のkeyは文字列のユーザid, valueは変更するカテゴリのids
      _auth_change_related_user_category_ids = {} #権限変更するユーザIDに対応するカテゴリID。権限削除ならカテゴリは空。カテゴリ変更があり、権限削除でないユーザなら、変更後のカテゴリを参照する
      get_users_current_category_ids.(opts[:change_role_user_ids].keys).each do |categories_user|
        _auth_change_related_user_category_ids[categories_user[:user_id].to_s] = (
          user_category_ids[categories_user[:user_id].to_s] || categories_user[:category_ids]
        ) 
      end

      _rules.each do |rule|
        rule = create_rule_object.(rule)
        rule_delete_conds = []
        #条件が全てなくなるかのチェック
        if rule.orig_delete_history_cond_chp_users.size > 0 and rule.orig_delete_history_cond_chp_users.size == rule.count_history_cond_user_ids then
          #履歴追加の条件に設定されているユーザが全て削除されるなら条件削除
          rule_delete_conds |= rule.delete_history_cond_chp_users
        end
        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          #条件がなくなるならルールを削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザが削除されるため。"
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end

        #操作が全てなくなるかのチェック
        delete_user_actions = []
        #ユーザ削除によって操作が無くなるルールは削除
        delete_user_actions |= rule.delete_mail_users_to_actions
        delete_user_actions |= rule.delete_mail_users_from_actions
        delete_user_actions |= rule.delete_mail_categories_from_actions
        if rule.count_actions == delete_user_actions.uniq{|action| action['action_id']}.size then
          _delete_rules << { :rule_id => rule.id, :opts => delete_user_actions }
          if opts[:messages_necessary] then
            reason = if rule.count_actions == rule.orig_delete_mail_users_from_actions.size then
                       "ルールの操作が「#{_mail_user_action_name}」のみで、送信元に設定されているユーザが削除されるため。"
                     elsif rule.count_actions == rule.orig_delete_mail_users_to_actions.size then
                       "ルールの操作が「#{_mail_user_action_name}」のみで、送信先に設定されている全てのユーザが削除されるため。"
                     elsif rule.count_actions == rule.orig_delete_mail_categories_from_actions.size then
                       "ルールの操作が「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザが削除されるため。"
                     else
                       "ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、関連するユーザが全て削除されるため。"
                     end
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}#{reason}"
            }
          end
          next
        end

        if _rule_messages[rule.id] or _delete_rules.find{|del_rule| del_rule[:rule_id] == rule.id} then
          next
        end

        #カテゴリ変更チェック
        #条件が全てなくなるかのチェック
        rule_delete_chp_user_subs = [] #カテゴリ変更によって削除する履歴追加の条件のchp_user_sub_ids
        rule.change_category_history_cond_chp_users.each do |chp_user|
          classify_chp_user.(chp_user, rule_delete_chp_user_subs, user_category_ids, rule.sidfmd_category_id)
        end
        if rule_delete_chp_user_subs.size > 0 then
          if (rule.orig_delete_history_cond_chp_users.size + rule_delete_chp_user_subs.size) == rule.count_history_cond_user_ids then
            #全てのユーザのカテゴリがルールのカテゴリから外れる、または削除されるなら条件削除
            rule_delete_conds |= (rule.change_category_history_cond_chp_users + rule.delete_history_cond_chp_users)
          end
        end

        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          #全ての条件がなくなるならルール削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = if rule.orig_delete_history_cond_chp_users.size == rule.count_history_cond_user_ids then
                       "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザが削除されるため。"
                     elsif rule_delete_chp_user_subs.size == rule.count_history_cond_user_ids then
                       "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                     else
                       "ユーザの削除及びカテゴリ変更によって、有効な条件が無くなるため。"
                     end
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end
  
        delete_mail_to_users_actions = [] #カテゴリ変更によって削除するメール送信(ユーザ)の送信先のactionのids
        delete_mail_from_users_actions = [] #カテゴリ変更によって削除するメール送信(ユーザ)の送信元のactionのids      
        delete_mail_category_actions = [] #カテゴリ変更によって削除するメール送信(カテゴリ)のactionのids
        
        rule.change_category_mail_categories_from_actions.each do |action|
          classify_action.(
            action,
            'mail_category',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_category_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_category_actions }
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          }
          next
        end

        rule.change_category_mail_users_from_actions.each do |action|
          classify_action.(
            action,
            'mail_from_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_from_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_from_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_from_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_from_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule.change_category_mail_users_to_actions.each do |action|
          classify_action.(
            action,
            'mail_to_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_to_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_to_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_to_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_to_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元と「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        merged_delete_mail_to_user_actions = (rule.delete_mail_users_to_actions + delete_mail_to_users_actions).uniq
        all_rule_delete_actions = (merged_delete_mail_to_user_actions + delete_user_actions + delete_mail_category_actions + delete_mail_from_users_actions).uniq
        #カテゴリ変更とユーザ削除によって操作が無くなるルールを削除
        if all_rule_delete_actions.uniq{|action| action['action_id']}.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => all_rule_delete_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ユーザの削除及びカテゴリ変更によって、有効な操作がなくなるため。"
            }
          end
          next
        end

        if _rule_messages[rule.id] or _delete_rules.find{|del_rule| del_rule[:rule_id] == rule.id} then
          next
        end

        #権限変更のチェック
        #条件のチェック
        rule_change_role_history_cond_chp_users = rule.change_role_history_cond_chp_users
        rule_delete_chp_user_subs_by_role = [] #権限変更によって削除されるchp_users
        rule_change_role_history_cond_chp_users.each do |chp_user|
          classify_chp_user.(chp_user, rule_delete_chp_user_subs_by_role, _auth_change_related_user_category_ids, rule.sidfmd_category_id)
        end

        if rule_delete_chp_user_subs_by_role.size > 0 then
          if (rule.delete_history_cond_chp_users + rule_delete_chp_user_subs + rule_delete_chp_user_subs_by_role).uniq{|u_sub| u_sub['user_id']}.size == rule.count_history_cond_user_ids then
            #全てのユーザのカテゴリがルールのカテゴリから外れているなら条件削除
            rule_delete_conds |= (rule.change_category_history_cond_chp_users + rule.delete_history_cond_chp_users + rule_delete_chp_user_subs_by_role)
          end
        end

        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = if rule_delete_chp_user_subs_by_role.size == rule.count_history_cond_user_ids or
                        (rule_delete_chp_user_subs_by_role.size + rule_delete_chp_user_subs.size) == rule.count_history_cond_user_ids
                     then
                       #権限変更のみ、もしくは権限変更とカテゴリ変更で条件削除されるパターン
                       "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                     else
                       #権限変更とカテゴリ変更で条件削除されるパターン
                       "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザが、削除される、またはルールカテゴリと一致しなくなるため。"
                     end
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end

        #操作のチェック
        delete_mail_to_users_actions_by_role = [] #権限変更に関連するメール送信(ユーザ)
        rule_change_role_mail_users_to_actions = rule.change_role_mail_users_to_actions
        rule_change_role_mail_users_to_actions.each do |action|
          classify_action.(
            action,
            'mail_to_users',
            delete_mail_to_users_actions_by_role,
            [],
            [],
            _auth_change_related_user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_to_users_actions_by_role.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_to_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule_change_role_mail_users_from_actions = rule.change_role_mail_users_from_actions
        rule_change_role_mail_categories_from_actions = rule.change_role_mail_categories_from_actions

        delete_mail_from_users_actions_by_role = []
        delete_mail_category_actions_by_role = []
        rule_change_role_mail_categories_from_actions.each do |action|
          classify_action.(
            action,
            'mail_category',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions_by_role,
            delete_mail_category_actions_by_role,
            _auth_change_related_user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_category_actions_by_role.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_category_actions_by_role }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule_change_role_mail_users_from_actions.each do |action|
          classify_action.(
            action,
            'mail_from_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions_by_role,
            delete_mail_category_actions_by_role,
            _auth_change_related_user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_from_users_actions_by_role.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_from_users_actions_by_role }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_from_users_actions_by_role.size + delete_mail_category_actions_by_role.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_from_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        merged_delete_mail_to_user_actions = (merged_delete_mail_to_user_actions + delete_mail_to_users_actions_by_role).uniq
        all_rule_delete_actions = (all_rule_delete_actions + delete_mail_to_users_actions_by_role + delete_mail_from_users_actions_by_role + delete_mail_category_actions_by_role).uniq

        #権限変更と、カテゴリ変更/ユーザ削除によって操作が無くなるルールを削除
        if all_rule_delete_actions.uniq{|action| action['action_id']}.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => all_rule_delete_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ユーザの権限変更、ユーザの削除、またはカテゴリ変更によって、有効な操作がなくなるため。"
            }
          end
          next
        end

        if _rule_messages[rule.id] or _delete_rules.find{|del_rule| del_rule[:rule_id] == rule.id} then
          next
        end

        _delete_conds |= rule_delete_conds
        #条件のチェック
        is_history_cond_deleted_by_user_delete = (rule.orig_delete_history_cond_chp_users.size > 0 and (rule_delete_conds.find{|cond| cond['cond_id'] == rule.delete_history_cond_chp_users.first['cond_id']} != nil))

        is_history_cond_deleted_by_user_cat_change = (rule.orig_change_category_history_cond_chp_users.size > 0 and (rule_delete_conds.find{|cond| cond['cond_id'] == rule.change_category_history_cond_chp_users.first['cond_id']} != nil))
 
        is_history_cond_deleted_by_role = (rule_delete_chp_user_subs_by_role.size > 0 and (rule_delete_conds.find{|cond| cond['cond_id'] == rule_delete_chp_user_subs_by_role.first['cond_id']} != nil))        

        if is_history_cond_deleted_by_user_delete or is_history_cond_deleted_by_user_cat_change or is_history_cond_deleted_by_role then
          if opts[:messages_necessary] then
            content_message = if is_history_cond_deleted_by_user_delete and is_history_cond_deleted_by_user_cat_change and is_history_cond_deleted_by_role then
                                "削除、カテゴリ変更、及び権限変更によって、「#{_history_cond_name}」に設定されている全てのユーザが無効になるため。"
                              elsif is_history_cond_deleted_by_user_delete and is_history_cond_deleted_by_role then
                                "削除及び権限変更によって、「#{_history_cond_name}」に設定されている全てのユーザが無効になるため。"
                              elsif is_history_cond_deleted_by_user_delete and is_history_cond_deleted_by_user_cat_change then
                                "削除及びカテゴリ変更によって、「#{_history_cond_name}」に設定されている全てのユーザが無効になるため。"
                              elsif is_history_cond_deleted_by_user_delete then
                                "「#{_history_cond_name}」に設定されている全てのユーザが削除されるため。"
                              else
                                "「#{_history_cond_name}」に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                              end
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "条件",
              :content => "「#{_history_cond_name}」が削除されます。",
              :reason => "#{content_message}"
            } 
          end
        else
          merged_rule_delete_chp_user_subs = (rule.delete_history_cond_chp_users + rule_delete_chp_user_subs + rule_delete_chp_user_subs_by_role) 
          if merged_rule_delete_chp_user_subs.size > 0 then
            _delete_chp_user_subs |= merged_rule_delete_chp_user_subs
            if opts[:messages_necessary] then 
              user_names = get_user_names.(merged_rule_delete_chp_user_subs)
              if user_names.size > 0 then
                content_message = if rule.orig_delete_history_cond_chp_users.size > 0 and
                                    (rule_delete_chp_user_subs.size > 0 or rule_delete_chp_user_subs_by_role.size > 0)
                                  then
                                    "ユーザが削除される、またはユーザのカテゴリがルールのカテゴリと一致しなくなるため。"
                                  elsif rule.orig_delete_history_cond_chp_users.size > 0 then
                                    "ユーザが削除されるため。"
                                  else
                                    "ユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                                  end
                _rule_messages[rule.id] = {
                  :rule_name => rule.name,
                  :item_name => "条件",
                  :content => "「#{_history_cond_name}」から、#{user_names.join(', ')}が削除されます。",
                  :reason => "#{content_message}"
                }
              end
            end
          end
        end

        #操作のチェック
        _delete_actions |= all_rule_delete_actions

        if opts[:messages_necessary] then
          content_msgs = []
          reason_msgs = []

          if rule.orig_delete_mail_categories_from_actions.any? then
            content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_cat_action_name}」の送信元に設定されているユーザが削除されるため。"
          elsif delete_mail_category_actions.any? then
            content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          elsif delete_mail_category_actions_by_role.any? then
            content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          end

          if rule.orig_delete_mail_users_from_actions.any? then
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_user_action_name}」の送信元に設定されているユーザが削除されるため。"
          elsif delete_mail_from_users_actions.any? then
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          elsif delete_mail_from_users_actions_by_role.any? then
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          elsif rule.count_to_user_ids > 0 and merged_delete_mail_to_user_actions.size == rule.count_to_user_ids then
            #送信先ユーザがなくなるなら、メール送信(ユーザ)の操作を削除
            reason = if rule.count_to_user_ids == rule.orig_delete_mail_users_to_actions.size then
                       "送信先に設定されている全てのユーザが削除されるため。"
                     elsif rule.count_to_user_ids == delete_mail_to_users_actions.size or rule.count_to_user_ids == delete_mail_to_users_actions_by_role.size then
                       "送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                     else
                       "送信先に設定されている全てのユーザが無効になるため。"
                     end
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << reason
          elsif merged_delete_mail_to_user_actions.any? then
            user_names = get_user_names.(merged_delete_mail_to_user_actions)
            if user_names.any? then
              content_msgs << "「#{_mail_user_action_name}」の送信先から、#{user_names.join(', ')}が削除されます。"
              reason_msgs << if rule.orig_delete_mail_users_to_actions.any? and (delete_mail_to_users_actions.any? or delete_mail_to_users_actions_by_role.any?) then
                               "ユーザが削除される、またはユーザのカテゴリがルールのカテゴリと一致しなくなるため。"
                             elsif rule.orig_delete_mail_users_to_actions.any? then
                               "ユーザが削除されるため。"
                             else
                               "ユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                             end
            end
          end
          if content_msgs.any? then
            if _rule_messages[rule.id] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件・操作",
                :content => _rule_messages[rule.id][:content] + "\n" + content_msgs.join("\n"),
                :reason => _rule_messages[rule.id][:reason] + "\n" + reason_msgs.join("\n")
              }
            else
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "操作",
                :content => content_msgs.join("\n"),
                :reason => reason_msgs.join("\n")
              }
            end
          end
        end
      end
    when 'auth_change', 'bulk_auth_change' then
      #権限変更
      #事前に、変更前が管理者でないものと、変更後が管理者であるものは除外されていることに注意

      #ユーザのカテゴリ。
      #インポート時は、カテゴリ変更があるユーザなら、変更後のカテゴリを参照するようにする
      #ただし、インポート時は権限削除では、カテゴリを空にすることが強制されるため、カテゴリ変更のチェックのみで良い
      _auth_change_related_user_category_ids = {}
      get_users_current_category_ids.(opts[:change_role_user_ids].keys).each do |categories_user|
        _auth_change_related_user_category_ids[categories_user[:user_id].to_s] =
          if opts[:change_category_user_ids] && opts[:change_category_user_ids][categories_user[:user_id].to_s] then
            opts[:change_category_user_ids][categories_user[:user_id].to_s]
          else
            categories_user[:category_ids]
          end
      end

      #ルールのチェック
      _rules.each do |rule|
        rule_delete_conds = []
        rule = create_rule_object.(rule)
        rule_change_role_history_cond_chp_users = rule.change_role_history_cond_chp_users
        #条件がなくなるかのチェック
        rule_delete_chp_user_subs = []
        rule_change_role_history_cond_chp_users.each do |chp_user|
          classify_chp_user.(chp_user, rule_delete_chp_user_subs, _auth_change_related_user_category_ids, rule.sidfmd_category_id)
        end
        if rule_delete_chp_user_subs.size > 0 then
          if rule_delete_chp_user_subs.size == rule.count_history_cond_user_ids then
            #全てのユーザのカテゴリがルールのカテゴリから外れるなら条件削除
            rule_delete_conds |= rule_change_role_history_cond_chp_users
          end
        end

        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          #全ての条件がなくなるならルール削除
           _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end
      
        #操作が無くなるかのチェック
        delete_mail_to_users_actions = []
        delete_mail_from_users_actions = []
        delete_mail_category_actions = []

        rule_change_role_mail_users_to_actions = rule.change_role_mail_users_to_actions
        rule_change_role_mail_users_from_actions = rule.change_role_mail_users_from_actions
        rule_change_role_mail_categories_from_actions = rule.change_role_mail_categories_from_actions

        rule_change_role_mail_categories_from_actions.each do |action|
          classify_action.(
            action,
            'mail_category',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            _auth_change_related_user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_category_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_category_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule_change_role_mail_users_from_actions.each do |action|
          classify_action.(
            action,
            'mail_from_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            _auth_change_related_user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_from_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_from_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_from_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_from_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule_change_role_mail_users_to_actions.each do |action|
          classify_action.(
            action,
            'mail_to_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions,
            delete_mail_category_actions,
            _auth_change_related_user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_to_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_to_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_to_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_to_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元と「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        if _rule_messages[rule.id] or _delete_rules.find{|del_rule| del_rule[:rule_id] == rule.id} then
          next     
        end

        _delete_conds |= rule_delete_conds
        #条件のチェック
        if rule_delete_chp_user_subs.size > 0 then
          if rule_delete_conds.find{|cond| cond['cond_id'] == rule_change_role_history_cond_chp_users.first['cond_id']} then
            if opts[:messages_necessary] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件",
                :content => "「#{_history_cond_name}」が削除されます。",
                :reason => "「#{_history_cond_name}」に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
              }
            end
          else
            _delete_chp_user_subs |= rule_delete_chp_user_subs
            if opts[:messages_necessary] then
              user_names = get_user_names.(rule_delete_chp_user_subs)
              if user_names.any? then
                _rule_messages[rule.id] = {
                  :rule_name => rule.name,
                  :item_name => "条件",
                  :content => "「#{_history_cond_name}」から、#{user_names.join(', ')}が削除されます。",
                  :reason => "ユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                }
              end
            end
          end
        end
  
        #操作のチェック
        _delete_actions |= (delete_mail_category_actions + delete_mail_from_users_actions + delete_mail_to_users_actions)

        if opts[:messages_necessary] then
          set_rule_messages_of_actions_related_to_categories.(rule, _rule_messages, delete_mail_category_actions, delete_mail_from_users_actions, delete_mail_to_users_actions)
        end
      end
    end
    #ここまでルールのチェック

    delete_chp_user_sub_ids = _delete_chp_user_subs.map{|chp_user| chp_user['chp_user_sub_id']}.uniq
    unchangeable_users = []
    is_unchangeable_user = false

    if opts[:role_admin] then
      delete_rule_ids = _delete_rules.map{|rule| rule[:rule_id]}.uniq
      delete_action_ids = _delete_actions.map{|action| action['action_id']}.uniq
      delete_cond_ids = _delete_conds.map{|cond| cond['cond_id']}.uniq
    else
      #管理者で無ければ、カテゴリ範囲外のルールのルール自体が削除される、または操作・条件自体が削除される場合は変更できないユーザを取得する
      unchangeable_user_ids = []
      t_rules = {}
      _delete_rules.each do |rule|
        if t_rules[rule[:rule_id]] then
          t_rules[rule[:rule_id]] += rule[:opts]
        else
          t_rules[rule[:rule_id]] = rule[:opts]
        end
      end
      t_actions = {}
      t_action_ids = []
      _delete_actions.each do |action|
        t_actions[action['action_id']] = [] unless t_actions[action['action_id']]
        t_actions[action['action_id']] << action['user_id']
        t_action_ids << action['action_id']
      end
      t_conds = {}
      _delete_conds.each do |cond|
        t_conds[cond['cond_id']] = [] unless t_conds[cond['cond_id']]
        t_conds[cond['cond_id']] << cond['user_id']
      end
      
      unchangeable_rules = (opts[:user_category_ids].size > 0 ? SidfmdRule.where(%|sidfmd_category_id NOT IN (?) OR sidfmd_category_id IS NULL|, opts[:user_category_ids]) : SidfmdRule.all)
      unchangeable_rules.where(:id => t_rules.keys).each do |rule|
        if t_rules[rule.id] then
          #ルール自体が削除されるなら、ユーザの変更はできない
          unchangeable_user_ids |= t_rules[rule.id].map{|opt| opt['user_id']}
          t_rules.delete(rule.id)
          _rule_messages.delete(rule.id) if opts[:messages_necessary]
        end
      end

      unchangeable_rules.where(
        :id => SidfmdRuleAction.where(:id => t_action_ids).select(:sidfmd_rule_id) 
      )
      .select("
        (
          ARRAY(
            SELECT sidfmd_rule_actions.id 
            FROM sidfmd_rule_a_mail_users
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id
          )
        ) AS mail_to_action_ids
      ")
      .select(:id)
      .each do |rule|
        if (rule.mail_to_action_ids - t_action_ids).size == 0 then
          #操作自体が削除されるなら、ユーザを変更できない
          rule.mail_to_action_ids.each do |action_id|
            unchangeable_user_ids |= t_actions[action_id] if t_actions[action_id]
            t_actions.delete(action_id)
          end
          _rule_messages.delete(rule.id) if opts[:messages_necessary]
        end
      end
      
      SidfmdRuleAction.joins(:sidfmd_rule_a_mail_category).where(%|
        sidfmd_rule_actions.id IN (?) AND sidfmd_rule_actions.sidfmd_rule_id IN (?)
      |,t_action_ids, unchangeable_rules.select(:id)
      )
      .select('sidfmd_rule_actions.id AS action_id', 'sidfmd_rule_actions.sidfmd_rule_id AS rule_id')
      .each do |action|
        unchangeable_user_ids |= t_actions[action.action_id] if t_actions[action.action_id]
        t_actions.delete(action.action_id)
        _rule_messages.delete(action.rule_id) if opts[:messages_necessary] 
      end
 
      SidfmdRuleCond.where(%|
        id IN (?) AND sidfmd_rule_id IN (?)
      |, t_conds.keys, unchangeable_rules.select(:id)
      )
      .select('id AS cond_id', 'sidfmd_rule_id AS rule_id')
      .each do |cond|
        unchangeable_user_ids |= t_conds[cond.cond_id] if t_conds[cond.cond_id]
        t_conds.delete(cond.cond_id)
        _rule_messages.delete(cond.rule_id) if opts[:messages_necessary]  
      end
      delete_rule_ids = t_rules.keys
      delete_action_ids = t_actions.keys
      delete_cond_ids = t_conds.keys
      if unchangeable_user_ids.any? then
        is_unchangeable_user = true
        if opts[:messages_necessary] then
          unchangeable_users = User.where(
            :id => unchangeable_user_ids
          ).map{|user|
            {
              :login => user.login,
              :lastname => user.lastname, 
              :firstname => user.firstname,
              :mail => user.mail
            }
         }
        end
      end
      if opts[:messages_necessary] then
        unchangeable_rules.where(:id => _rule_messages.keys).each do |rule|
          _rule_messages.delete(rule.id)
        end
      end
    end 

    ret = {
      :delete_rule_ids => delete_rule_ids,
      :delete_action_ids => delete_action_ids,
      :delete_cond_ids => delete_cond_ids,
      :delete_chp_user_sub_ids => delete_chp_user_sub_ids,
      :is_unchangeable_user => is_unchangeable_user,
      :unchangeable_users => unchangeable_users
    }

    ret[:rule_messages] = _rule_messages if opts[:messages_necessary]
    ret


  end

  #ホストに関連するルールを返す。
  #actionは今のところimport/category_change/bulk_category_change/deleteのいずれか
  #:change_category_host_idsにはカテゴリ変更するホストのidsを、:delete_host_idsには削除するホストのidsを入れて渡す
  def self.hosts_related_rule_info(**opts)
    #初めの文字がアンダースコアの変数にしてあるものは再定義しないこと
    host_ids = []
    unless opts[:role_admin] then
      unless opts[:user_category_ids] then
        raise ArgumentError, "must be user_category_ids in case the user's role is not admin"
      end
    end

    case opts[:action]
    when 'import' then
      if !opts[:delete_host_ids] or !opts[:change_category_host_ids] then
        raise ArgumentError, "must be delete_host_ids and change_category_host_ids in case of the import"
      end
    when  'delete' then
      if !opts[:delete_host_ids] then
        raise ArgumentError, "must be delete_host_ids in case of the delete"
      end
    when 'category_change', 'bulk_category_change' then
      if !opts[:change_category_host_ids] then
        raise ArgumentError, "must be change_category_host_ids in case of the category_change"
      end
    else
      #invalid
      raise ArgumentError, "invalid action"
    end

    if opts[:delete_host_ids] then
      opts[:delete_host_ids] = opts[:delete_host_ids].map(&:to_i)
      host_ids |= opts[:delete_host_ids]
    end
    if opts[:change_category_host_ids] then
      host_ids |= opts[:change_category_host_ids].keys.map(&:to_i)
      #opts[:change_category_host_ids]のkeyはStringで統一し、
      #values(カテゴリID)からシステムカテゴリを排除
      key_string_change_category_host_ids = {}
      system_category_ids = SidfmdCategory.where(:is_system => true).pluck(:id)
      opts[:change_category_host_ids].each do |hostid, category_ids|
        key_string_change_category_host_ids[hostid.to_s] = (category_ids - system_category_ids)
      end
      opts[:change_category_host_ids] = key_string_change_category_host_ids
    end

    _rules = if host_ids.any? then
      SidfmdRule.where(%|
        id IN (?) OR id IN (?)
      |,SidfmdRuleTrigger.joins(:sidfmd_rule_t_vulsync)
                         .where.not(:sidfmd_rule_t_vulsyncs => { :sidfmd_host_id => nil })
                         .select(:sidfmd_rule_id), #ホストが設定されている脆弱性同期時実行のタイミング
        SidfmdRuleCond.joins(:sidfmd_rule_c_match_host).select(:sidfmd_rule_id), #ホスト関連の条件
      )
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_conds
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_conds /*ルールのcondの数*/
      ")
      .select("
         (
           SELECT COUNT (*)
           FROM sidfmd_rule_conds
           INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
           INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
           WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_hosts.inv IS FALSE /*ホスト限定に設定されているホストの数*/
         ) AS count_match_sub_host_conds
      ")
      .select("
        (
          SELECT COUNT (*)
          FROM sidfmd_rule_conds
          INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
          INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
          WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_hosts.inv IS TRUE /*ホスト除外に設定されているホストの数*/
        ) AS count_unmatch_sub_host_conds
      ") 
      .select(:id, :name, :sidfmd_category_id)
    else
      SidfmdRule.where(:id => 0)
    end

    if opts[:delete_host_ids] then
      opts[:delete_host_ids] = [ 0 ] if opts[:delete_host_ids].size == 0
      _rules = _rules.select("
        (
          SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_triggers.id, (CASE WHEN sidfmd_rule_t_vulsyncs.sidfmd_category_id IS NULL THEN 0 ELSE sidfmd_rule_t_vulsyncs.sidfmd_category_id END), sidfmd_rule_t_vulsyncs.sidfmd_host_id], ':')
          FROM sidfmd_rule_triggers INNER JOIN sidfmd_rule_t_vulsyncs ON sidfmd_rule_triggers.id = sidfmd_rule_t_vulsyncs.sidfmd_rule_trigger_id
          WHERE sidfmd_rule_triggers.sidfmd_rule_id = sidfmd_rules.id AND 
                sidfmd_rule_t_vulsyncs.sidfmd_host_id IN (SELECT unnest (ARRAY[#{opts[:delete_host_ids].join(', ')}]))
        ) AS delete_trigger_vulsync /*設定されているホストが削除される脆弱性同期時実行のタイミング*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_host_subs.id, sidfmd_rule_cm_host_subs.sidfmd_host_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
            INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND 
                  sidfmd_rule_c_match_hosts.inv IS FALSE AND
                  sidfmd_rule_cm_host_subs.sidfmd_host_id IN (SELECT unnest (ARRAY[#{opts[:delete_host_ids].join(', ')}])) /*設定されているホストが削除されるホスト限定*/
          )
        ) AS orig_delete_match_host_conds
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_host_subs.id, sidfmd_rule_cm_host_subs.sidfmd_host_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
            INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND 
                  sidfmd_rule_c_match_hosts.inv IS TRUE AND
                  sidfmd_rule_cm_host_subs.sidfmd_host_id IN (SELECT unnest (ARRAY[#{opts[:delete_host_ids].join(', ')}])) /*設定されているホストが削除されるホスト除外*/
          )
        ) AS orig_delete_unmatch_host_conds
      ")
    end

    if opts[:change_category_host_ids] then
      target_host_ids = opts[:change_category_host_ids].keys.map(&:to_i)
      target_host_ids = [ 0 ] if target_host_ids.size == 0
      _rules = _rules.select("
        (
          SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_triggers.id, (CASE WHEN sidfmd_rule_t_vulsyncs.sidfmd_category_id IS NULL THEN 0 ELSE sidfmd_rule_t_vulsyncs.sidfmd_category_id END), sidfmd_rule_t_vulsyncs.sidfmd_host_id], ':')
          FROM sidfmd_rule_triggers INNER JOIN sidfmd_rule_t_vulsyncs ON sidfmd_rule_triggers.id = sidfmd_rule_t_vulsyncs.sidfmd_rule_trigger_id
          WHERE sidfmd_rule_triggers.sidfmd_rule_id = sidfmd_rules.id AND 
                sidfmd_rule_t_vulsyncs.sidfmd_host_id IN (SELECT unnest (ARRAY[#{target_host_ids.join(', ')}]))
        ) AS category_change_host_trigger_vulsync /*設定されているホストのカテゴリが変更される脆弱性同期時実行のタイミング*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_host_subs.id, sidfmd_rule_cm_host_subs.sidfmd_host_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
            INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND 
                  sidfmd_rule_c_match_hosts.inv IS FALSE AND
                  sidfmd_rule_cm_host_subs.sidfmd_host_id IN (SELECT unnest (ARRAY[#{target_host_ids.join(', ')}])) /*設定されているホストのカテゴリが変更されるホスト限定*/
          )
        ) AS orig_category_change_host_match_host_conds
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_host_subs.id, sidfmd_rule_cm_host_subs.sidfmd_host_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
            INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND 
                  sidfmd_rule_c_match_hosts.inv IS TRUE AND
                  sidfmd_rule_cm_host_subs.sidfmd_host_id IN (SELECT unnest (ARRAY[#{target_host_ids.join(', ')}])) /*設定されているホストカテゴリが変更されるホスト除外*/
          )
        ) AS orig_category_change_host_unmatch_host_conds
      ") 
    end

    _match_host_cond_name = RULE_COND_NAMES['host-match']
    _unmatch_host_cond_name = RULE_COND_NAMES['host-unmatch']
    _rule_messages = {}
    _delete_rules = []
    _delete_conds = []
    _delete_sub_host_conds = []

    create_m_host_cond_object = ->(conds){
      obj = conds.map{|cond|
        cols = cond.split(':')
        {
          :cond_id => cols[0].to_i,
          :sub_host_cond_id => cols[1].to_i,
          :sidfmd_host_id => cols[2].to_i,
        }
      }
    }

    create_trigger_vulsync_object = -> (trigger){
      cols = trigger.split(':')
      obj = {
        :trigger_id => cols[0].to_i,
        :sidfmd_category_id => cols[1].to_i,
        :sidfmd_host_id => cols[2].to_i,
      }
    }

    create_delete_rule_object = -> (rule, opts){
      { :rule_id => rule.id, :opts => Array.wrap(opts) }
    }

    get_hostnames = -> (host_ids){
      SidfmdHost.where(:id => host_ids).order(%|
        CASE WHEN ext_name IS NULL THEN LOWER(hostname COLLATE "C") ELSE LOWER(ext_name COLLATE "C") END
      |).map{|host|
        "「#{(host.ext_name ? (host.ext_name + '(ホスト名: ' + host.hostname + ')') : host.hostname)}」"
      }
    }

    check_host_belongs_to_rule = -> (host_id, category_id){
      category_id = category_id.to_i
      ret = true
      target_host_cat_ids = opts[:change_category_host_ids][host_id.to_s]
      if target_host_cat_ids then
        if category_id == 0 then
          ret = target_host_cat_ids.any?
        else
          ret = target_host_cat_ids.include?(category_id)
        end
      end
      ret
    }

    add_remove_conds_by_category_change = -> (conds, rule_category_id, remove_conds){
      conds.each do |cond|
        is_in_category = check_host_belongs_to_rule.(cond[:sidfmd_host_id], rule_category_id)
        if !is_in_category then
          remove_conds << cond
        end
      end
    }

    get_delete_rule_reason_item_names = -> (host_match_conds, host_unmach_conds){
      item_name = if host_match_conds.any? and host_unmach_conds.any? then
        "「#{_match_host_cond_name}」と「#{_unmatch_host_cond_name}」"
      elsif host_match_conds.any? then
        "「#{_match_host_cond_name}」"
      else
        "「#{_unmatch_host_cond_name}」"
      end
      item_name
    }

    case opts[:action]
    when 'delete' then
      #ホスト削除の場合
      _rules.each do |rule|
        if rule.delete_trigger_vulsync then
          trigger_vulsync = create_trigger_vulsync_object.(rule.delete_trigger_vulsync)
          _delete_rules << create_delete_rule_object.(rule, trigger_vulsync)
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "タイミングに削除対象のホストが設定されているため。"
            }
          end
          next
        end

        delete_match_host_conds = create_m_host_cond_object.(rule.orig_delete_match_host_conds)
        delete_unmatch_host_conds = create_m_host_cond_object.(rule.orig_delete_unmatch_host_conds)
        rule_m_host_conds_size = 0
        rule_m_host_conds_size += 1 if delete_match_host_conds.any?
        rule_m_host_conds_size += 1 if delete_unmatch_host_conds.any?       
 
        if rule.count_conds == rule_m_host_conds_size and
           delete_match_host_conds.size == rule.count_match_sub_host_conds and 
           delete_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds
        then
          _delete_rules << create_delete_rule_object.(rule, (delete_match_host_conds + delete_unmatch_host_conds))
          if opts[:messages_necessary] then
            reason_item_names = get_delete_rule_reason_item_names.(delete_match_host_conds, delete_unmatch_host_conds)
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "条件が#{reason_item_names}のみで、設定されている全てのホストが削除されるため。"
            }
          end
          next
        end
 
        item_names = []
        content_msgs = []
        reason_msgs = []

        if delete_match_host_conds.size > 0 then
          if delete_match_host_conds.size == rule.count_match_sub_host_conds then
            _delete_conds += delete_match_host_conds
            if opts[:messages_necessary] then
              item_names << "条件"
              content_msgs << "「#{_match_host_cond_name}」が削除されます。"
              reason_msgs << "「#{_match_host_cond_name}」に設定されている全てのホストが削除されるため。"
            end
          else
            _delete_sub_host_conds += delete_match_host_conds
            if opts[:messages_necessary] then
              hostnames = get_hostnames.(delete_match_host_conds.map{|c| c[:sidfmd_host_id]})
              item_names << "条件"
              content_msgs << "「#{_match_host_cond_name}」から、#{hostnames.join(', ')}が削除されます。"
              reason_msgs << "ホストが削除されるため。"
            end
          end
        end
        if delete_unmatch_host_conds.size > 0 then
          if delete_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds then
            _delete_conds += delete_unmatch_host_conds
            if opts[:messages_necessary] then
              item_names << "条件"
              content_msgs << "「#{_unmatch_host_cond_name}」が削除されます。"
              reason_msgs << "「#{_unmatch_host_cond_name}」に設定されている全てのホストが削除されるため。"
            end
          else
            _delete_sub_host_conds += delete_unmatch_host_conds
            if opts[:messages_necessary] then
              hostnames = get_hostnames.(delete_unmatch_host_conds.map{|c| c[:sidfmd_host_id]})
              item_names << "条件"
              content_msgs << "「#{_unmatch_host_cond_name}」から、#{hostnames.join(', ')}が削除されます。"
              reason_msgs << "ホストが削除されるため。"
            end
          end
        end
   
        if opts[:messages_necessary] and item_names.any? then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => item_names.uniq.join("\n"),
            :content => content_msgs.join("\n"),
            :reason => reason_msgs.join("\n")
          } 
        end
      end
    when 'category_change', 'bulk_category_change' then
      #カテゴリ変更/カテゴリ一括変更の場合
      _rules.each do |rule|
        if rule.category_change_host_trigger_vulsync then
          trigger_vulsync = create_trigger_vulsync_object.(rule.category_change_host_trigger_vulsync)
          is_in_category = check_host_belongs_to_rule.(trigger_vulsync[:sidfmd_host_id], trigger_vulsync[:sidfmd_category_id])
          if !is_in_category then
            _delete_rules << create_delete_rule_object.(rule, trigger_vulsync)
            if opts[:messages_necessary] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "ルール",
                :content => "ルールが削除されます。",
                :reason =>  "タイミングに設定しているホストが、タイミングの対象カテゴリに属さなくなるため。"
              }
            end
            next
          end
        end
        
        cat_change_match_host_conds = create_m_host_cond_object.(rule.orig_category_change_host_match_host_conds)
        cat_change_unmatch_host_conds = create_m_host_cond_object.(rule.orig_category_change_host_unmatch_host_conds)

        remove_match_host_conds = []
        remove_unmatch_host_conds = []

        add_remove_conds_by_category_change.(cat_change_match_host_conds, rule.sidfmd_category_id, remove_match_host_conds)
        add_remove_conds_by_category_change.(cat_change_unmatch_host_conds, rule.sidfmd_category_id, remove_unmatch_host_conds)
        rule_m_host_conds_size = 0
        rule_m_host_conds_size += 1 if cat_change_match_host_conds.any?
        rule_m_host_conds_size += 1 if cat_change_unmatch_host_conds.any?
    
        if rule.count_conds == rule_m_host_conds_size and
           remove_match_host_conds.size == rule.count_match_sub_host_conds and
           remove_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds
        then
          _delete_rules << create_delete_rule_object.(rule, (remove_match_host_conds + remove_unmatch_host_conds))
          if opts[:messages_necessary] then
            reason_item_names = get_delete_rule_reason_item_names.(cat_change_match_host_conds, cat_change_unmatch_host_conds)
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "条件が#{reason_item_names}のみで、設定されている全てのホストのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        item_names = []
        content_msgs = []
        reason_msgs = []

        if remove_match_host_conds.any? then
          if remove_match_host_conds.size == rule.count_match_sub_host_conds then
            _delete_conds += remove_match_host_conds
            if opts[:messages_necessary] then
              item_names << "条件"
              content_msgs << "「#{_match_host_cond_name}」が削除されます。"
              reason_msgs << "「#{_match_host_cond_name}」に設定されている全てのホストのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            end
          else
            _delete_sub_host_conds += remove_match_host_conds
            if opts[:messages_necessary] then
              hostnames = get_hostnames.(remove_match_host_conds.map{|c| c[:sidfmd_host_id]})
              item_names << "条件"
              content_msgs << "「#{_match_host_cond_name}」から、#{hostnames.join(', ')}が削除されます。"
              reason_msgs << "ホストのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            end
          end
        end

        if remove_unmatch_host_conds.any? then
          if remove_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds then
            _delete_conds += remove_unmatch_host_conds
            if opts[:messages_necessary] then
              item_names << "条件"
              content_msgs << "「#{_unmatch_host_cond_name}」が削除されます。"
              reason_msgs << "「#{_unmatch_host_cond_name}」に設定されている全てのホストのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            end
          else
            _delete_sub_host_conds += remove_unmatch_host_conds
            if opts[:messages_necessary] then
              hostnames = get_hostnames.(remove_unmatch_host_conds.map{|c| c[:sidfmd_host_id]})
              item_names << "条件"
              content_msgs << "「#{_unmatch_host_cond_name}」から、#{hostnames.join(', ')}が削除されます。"
              reason_msgs << "ホストのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            end
          end
        end

        if opts[:messages_necessary] and item_names.any? then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => item_names.uniq.join("\n"),
            :content => content_msgs.join("\n"),
            :reason => reason_msgs.join("\n")
          }
        end
      end
    when 'import' then
      _rules.each do |rule|
        if rule.delete_trigger_vulsync then
          #タイミングに設定されているホスト削除ならルール削除
          trigger_vulsync = create_trigger_vulsync_object.(rule.delete_trigger_vulsync)
          _delete_rules << create_delete_rule_object.(rule, trigger_vulsync)
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "タイミングに削除対象のホストが設定されているため。"
            }
          end
          next
        end

        if rule.category_change_host_trigger_vulsync then
          #タイミングに設定されているホストが、タイミングのカテゴリから外れるならルール削除
          trigger_vulsync = create_trigger_vulsync_object.(rule.category_change_host_trigger_vulsync)
          is_in_category = check_host_belongs_to_rule.(trigger_vulsync[:sidfmd_host_id], trigger_vulsync[:sidfmd_category_id])
          if !is_in_category then
            _delete_rules << create_delete_rule_object.(rule, trigger_vulsync)
            if opts[:messages_necessary] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "ルール",
                :content => "ルールが削除されます。",
                :reason =>  "タイミングに設定しているホストが、タイミングの対象カテゴリに属さなくなるため。"
              }
            end
            next
          end
        end

        delete_match_host_conds = create_m_host_cond_object.(rule.orig_delete_match_host_conds) #削除されるホストのホスト限定
        delete_unmatch_host_conds = create_m_host_cond_object.(rule.orig_delete_unmatch_host_conds) #削除されるホストのホスト除外

        rule_delete_m_host_conds_size = 0
        rule_delete_m_host_conds_size += 1 if delete_match_host_conds.any?
        rule_delete_m_host_conds_size += 1 if delete_unmatch_host_conds.any?

        if rule.count_conds == rule_delete_m_host_conds_size and
          delete_match_host_conds.size == rule.count_match_sub_host_conds and
          delete_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds
        then
          #ホスト限定/除外から全てのホスト削除で且つ、ルールの条件がホスト限定/除外のみならルール削除
          _delete_rules << create_delete_rule_object.(rule, (delete_match_host_conds + delete_unmatch_host_conds))
          if opts[:messages_necessary] then
            reason_item_names = get_delete_rule_reason_item_names.(delete_match_host_conds, delete_unmatch_host_conds)
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "条件が#{reason_item_names}のみで、設定されている全てのホストが削除されるため。"
            }
          end
          next
        end
        
        cat_change_match_host_conds = create_m_host_cond_object.(rule.orig_category_change_host_match_host_conds) #カテゴリ変更されるホストのホスト限定
        cat_change_unmatch_host_conds = create_m_host_cond_object.(rule.orig_category_change_host_unmatch_host_conds) #カテゴリ変更されるホストのホスト除外
        remove_match_host_conds = [] #カテゴリ変更でルールから外れるホスト限定の条件
        remove_unmatch_host_conds = [] #カテゴリ変更でルールから外れるホスト除外の条件
       
        add_remove_conds_by_category_change.(cat_change_match_host_conds, rule.sidfmd_category_id, remove_match_host_conds)
        add_remove_conds_by_category_change.(cat_change_unmatch_host_conds, rule.sidfmd_category_id, remove_unmatch_host_conds)
     
        rule_cat_change_m_host_conds_size = 0
        rule_cat_change_m_host_conds_size += 1 if cat_change_match_host_conds.any?
        rule_cat_change_m_host_conds_size += 1 if cat_change_unmatch_host_conds.any?

        if rule.count_conds == rule_cat_change_m_host_conds_size and
           remove_match_host_conds.size == rule.count_match_sub_host_conds and
           remove_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds
        then
          #ホスト限定/除外の全てのホストがカテゴリ変更でルールのカテゴリから外れ、且つルールの条件がホスト限定/除外のみならルール削除
          _delete_rules << create_delete_rule_object.(rule, (remove_match_host_conds + remove_unmatch_host_conds))
          if opts[:messages_necessary] then
            reason_item_names = get_delete_rule_reason_item_names.(cat_change_match_host_conds, cat_change_unmatch_host_conds)
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "条件が#{reason_item_names}のみで、設定されている全てのホストのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end
       
        rule_m_host_conds_size = 0
        rule_m_host_conds_size += 1 if delete_match_host_conds.any? or cat_change_match_host_conds.any?
        rule_m_host_conds_size += 1 if delete_unmatch_host_conds.any? or cat_change_unmatch_host_conds.any?
        
        rule_all_delete_match_host_conds = delete_match_host_conds + remove_match_host_conds
        rule_all_delete_unmatch_host_conds = delete_unmatch_host_conds + remove_unmatch_host_conds

        if rule.count_conds == rule_m_host_conds_size and
           rule_all_delete_match_host_conds.size == rule.count_match_sub_host_conds and
           rule_all_delete_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds
        then
          #ホスト限定/除外の全てのホストが削除またはカテゴリ変更でルールのカテゴリから外れ、且つルールの条件がホスト限定/除外のみならルール削除
          _delete_rules << create_delete_rule_object.(rule, (rule_all_delete_match_host_conds + rule_all_delete_unmatch_host_conds))
          if opts[:messages_necessary] then
            reason_item_names = get_delete_rule_reason_item_names.(rule_all_delete_match_host_conds, rule_all_delete_unmatch_host_conds)
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "条件が#{reason_item_names}のみで、設定されている全てのホストについて、削除される、またはホストのカテゴリがルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        item_names = []
        content_msgs = []
        reason_msgs = []

        if rule_all_delete_match_host_conds.any? then
          if rule_all_delete_match_host_conds.size == rule.count_match_sub_host_conds then
            _delete_conds += rule_all_delete_match_host_conds
            if opts[:messages_necessary] then
              item_names << "条件"
              content_msgs << "「#{_match_host_cond_name}」が削除されます。"
              reason_msgs << "「#{_match_host_cond_name}」に設定されている全てのホストについて、削除される、またはホストのカテゴリがルールのカテゴリと一致しなくなるため。"
            end
          else
            _delete_sub_host_conds += rule_all_delete_match_host_conds
            if opts[:messages_necessary] then
              hostnames = get_hostnames.(rule_all_delete_match_host_conds.map{|c| c[:sidfmd_host_id]})
              item_names << "条件"
              content_msgs << "「#{_match_host_cond_name}」から、#{hostnames.join(', ')}が削除されます。"
              reason_msgs << "ホストが削除される、またはホストのカテゴリがルールのカテゴリと一致しなくなるため。"
            end
          end
        end

        if rule_all_delete_unmatch_host_conds.any? then
          if rule_all_delete_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds then
            _delete_conds += rule_all_delete_unmatch_host_conds
            if opts[:messages_necessary] then
              item_names << "条件"
              content_msgs << "「#{_unmatch_host_cond_name}」が削除されます。"
              reason_msgs << "「#{_unmatch_host_cond_name}」に設定されている全てのホストについて、削除される、またはホストのカテゴリがルールのカテゴリと一致しなくなるため。"
            end
          else
            _delete_sub_host_conds += rule_all_delete_unmatch_host_conds
            if opts[:messages_necessary] then
              hostnames = get_hostnames.(rule_all_delete_unmatch_host_conds.map{|c| c[:sidfmd_host_id]})
              item_names << "条件"
              content_msgs << "「#{_unmatch_host_cond_name}」から、#{hostnames.join(', ')}が削除されます。"
              reason_msgs << "ホストが削除される、またはホストのカテゴリがルールのカテゴリと一致しなくなるため。"
            end
          end
        end

        if opts[:messages_necessary] and item_names.any? then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => item_names.uniq.join("\n"),
            :content => content_msgs.join("\n"),
            :reason => reason_msgs.join("\n")
          }
        end

      end
    end

    is_unchangeable_host = false
    unchangeable_hosts = []
 
    delete_sub_host_cond_ids = _delete_sub_host_conds.map{|cond| cond[:sub_host_cond_id]}.uniq
    if opts[:role_admin] then
      delete_rule_ids = _delete_rules.map{|rule| rule[:rule_id]}.uniq
      delete_cond_ids = _delete_conds.map{|cond| cond[:cond_id]}.uniq
    else
      #管理者で無ければ、カテゴリ範囲外のルールのルール自体が削除される、または操作・条件自体が削除される場合は変更できないホストを取得する
      unchangeable_host_ids = []
      t_rules = {}
      _delete_rules.each do |rule|
        if t_rules[rule[:rule_id]] then
          t_rules[rule[:rule_id]] += rule[:opts]
        else
          t_rules[rule[:rule_id]] = rule[:opts]
        end
      end
      t_conds = {}
      _delete_conds.each do |cond|
        t_conds[cond[:cond_id]] = [] unless t_conds[cond[:cond_id]]
        t_conds[cond[:cond_id]] << cond[:sidfmd_host_id]
      end
      unchangeable_rules = (opts[:user_category_ids].size > 0 ? SidfmdRule.where(%|sidfmd_category_id NOT IN (?) OR sidfmd_category_id IS NULL|, opts[:user_category_ids]) : SidfmdRule.all)

      unchangeable_rules.where(:id => t_rules.keys).each do |rule|
        #ルール自体が削除されるなら、ホストの変更はできない
        unchangeable_host_ids |= t_rules[rule.id].map{|opt| opt[:sidfmd_host_id]}
        t_rules.delete(rule.id)
        _rule_messages.delete(rule.id) if opts[:messages_necessary]  
      end

      SidfmdRuleCond.where(%|
        id IN (?) AND sidfmd_rule_id IN (?)
      |, t_conds.keys, unchangeable_rules.select(:id)
      )
      .select('id AS cond_id', 'sidfmd_rule_id AS rule_id')
      .each do |cond|
        #条件が削除されるなら、ホストの変更はできない
        unchangeable_host_ids |= t_conds[cond.cond_id]
        t_conds.delete(cond.cond_id)
        _rule_messages.delete(cond.rule_id) if opts[:messages_necessary]   
      end

      delete_rule_ids = t_rules.keys
      delete_cond_ids = t_conds.keys
      if unchangeable_host_ids.any? then
        is_unchangeable_host = true
        if opts[:messages_necessary] then
          unchangeable_hosts = SidfmdHost.where(
            :id => unchangeable_host_ids
          ).map{|host|
            { :hostname => host.hostname, :ext_name => host.ext_name }
          }
        end
      end
      if opts[:messages_necessary] then
        unchangeable_rules.where(:id => _rule_messages.keys).each do |rule|
          _rule_messages.delete(rule.id)
        end
      end
    end
   
    ret = {
      :delete_rule_ids => delete_rule_ids,
      :delete_cond_ids => delete_cond_ids,
      :delete_sub_host_cond_ids => delete_sub_host_cond_ids,
      :rule_messages => _rule_messages,
      :is_unchangeable_host => is_unchangeable_host,
      :unchangeable_hosts => unchangeable_hosts
    }
   
    ret[:rule_messages] = _rule_messages if opts[:messages_necessary]
    ret
   
  end

  #カテゴリに関連するルールを返す。
  #actionは今のところimport/deleteのみ
  #:delete_category_idsに削除するカテゴリのidsを入れて渡す
  def self.categories_related_rule_info(**opts)
    #初めの文字がアンダースコアの変数にしてあるものは再定義しないこと
    category_ids = []
    unless opts[:role_admin] then
      #カテゴリは、管理者権限以外操作できない
      raise Exception, 'only admin users can use the function "categories_related_rule_info"'
      #unless opts[:user_category_ids] then
      #  raise ArgumentError, "must be user_category_ids in case the user's role is not admin"
      #end
    end

    case opts[:action]
    when 'import' then
      if !opts[:delete_category_ids] then
        raise ArgumentError, "must be delete_host_ids in case of the import"
      end
    when  'delete' then
      if !opts[:delete_category_ids] then
        raise ArgumentError, "must be delete_host_ids in case of the delete"
      end
    else
      #invalid
      raise ArgumentError, "invalid action"
    end

    _no_category_host_ids = Set.new #カテゴリ数が0になるホストのID
    _no_category_user_ids = Set.new #カテゴリ数が0になるユーザのID

    # = 操作・条件をチェックする前に、ルールのカテゴリと脆弱性同期時のタイミングに設定されているカテゴリをチェックする。
    # このため、ホストとユーザのチェックをする段階では、ルールのカテゴリが「全体」であることが確定しているので、
    # 初めにカテゴリ削除によってカスタムカテゴリが無くなるホストとユーザのみ取得する

    #なお、メール送信(ユーザ)・(カテゴリ)の操作の送信先と送信元、履歴追加の条件のユーザには、次のように設定可能ユーザが決められている。
    #ルールのカテゴリが全体の場合: 全ての管理者権限ユーザと、いずれかのカテゴリに属するユーザのみ選択可能。
    #ルールのカテゴリが全体の以外の場合: いずれかのカテゴリに属するユーザのみ選択可能。

    if opts[:delete_category_ids] then
      opts[:delete_category_ids] = opts[:delete_category_ids].map(&:to_i)
      category_ids |= opts[:delete_category_ids]

      SidfmdCategoriesHost.where(
        :sidfmd_host_id => SidfmdCategoriesHost.where(:sidfmd_category_id => category_ids).select(:sidfmd_host_id)
      )
      .where.not(:sidfmd_category_id => SidfmdCategory.where(:is_system => true).select(:id))
      .group(:sidfmd_host_id)
      .select('ARRAY_AGG(DISTINCT sidfmd_category_id) AS cat_ids', :sidfmd_host_id)
      .each do |cats_host|
        _no_category_host_ids.add(cats_host.sidfmd_host_id) if (cats_host.cat_ids - opts[:delete_category_ids]).size == 0
      end

      User.only_category_user_ids(category_ids).each do |user_id|
        _no_category_user_ids.add(user_id)
      end
    end

    _no_category_host_ids = _no_category_host_ids.to_a
    _no_category_user_ids = _no_category_user_ids.to_a
    #管理者権限のユーザを除外した、カテゴリの数0になるユーザのID
    _no_category_user_ids_without_admin_user_ids = 
      _no_category_user_ids - Member.joins(:member_roles)
                                    .joins(%| INNER JOIN roles ON roles.id = member_roles.role_id|)
                                    .where(:roles => { :name => 'SIDfmVM管理者' })
                                    .distinct.pluck('members.user_id')

    # 今後あるかもしれない変更に対応できるように、変更の対象(原因)となるホスト名やユーザ名やid等を取れるようにしておく
    # 他のページでは、報告者権限ではログインユーザのカテゴリ範囲外のルールに影響して変更できないホストやユーザを表示しているため。
    _rules = if category_ids.any? then
      SidfmdRule.where(%|
        sidfmd_category_id IN (?) OR id IN (?) OR id IN (?) OR id IN (?) OR id IN (?) OR id IN (?)
      |,category_ids,      
        SidfmdRuleTrigger.joins(:sidfmd_rule_t_vulsync)
                         .where(%| sidfmd_rule_t_vulsyncs.sidfmd_category_id IN (?) OR sidfmd_rule_t_vulsyncs.sidfmd_host_id IS NOT NULL |, category_ids)
                         .select(:sidfmd_rule_id), #対象のカテゴリが設定されているか、ホストが設定されている脆弱性同期時実行のタイミング
        SidfmdRuleCond.where(%|
          id IN (?) OR id IN (?) OR id IN (?)
        |,SidfmdRuleCMatchCategory.joins(:sidfmd_rule_cm_category_subs)
                                  .where(:sidfmd_rule_cm_category_subs => { :sidfmd_category_id => category_ids })
                                  .distinct
                                  .select(:sidfmd_rule_cond_id), #対象カテゴリが設定されているカテゴリ限定/除外の条件
          SidfmdRuleCMatchHost.joins(:sidfmd_rule_cm_host_subs)
                              .where(:sidfmd_rule_cm_host_subs => { :sidfmd_host_id => _no_category_host_ids })
                              .distinct
                              .select(:sidfmd_rule_cond_id), #カスタムカテゴリ数が0になるホストが設定されているホスト限定/ホスト除外の条件
          SidfmdRuleCHistoryPost.joins(:sidfmd_rule_chp_user_subs)
                                .where(:sidfmd_rule_chp_user_subs => { :user_id => _no_category_user_ids_without_admin_user_ids } )
                                .distinct
                                .select(:sidfmd_rule_cond_id) #管理者権限を除くカスタムカテゴリ数が0になるユーザが設定されている履歴追加の条件
        ).select(:sidfmd_rule_id), #対象カテゴリに関連する条件
        SidfmdRuleAction.where(%|
          id IN (?) OR id IN (?)
        |,SidfmdRuleAMailCategory.where(%| sidfmd_category_id IN (?) OR from_user_id IN (?)|, category_ids, _no_category_user_ids)
                                 .distinct
                                 .select(:sidfmd_rule_action_id), #送信先が対象カテゴリか、送信元にスタムカテゴリ数が0になるホストが設定されているメール送信(カテゴリ)
          SidfmdRuleAMailUser.where(%| user_id IN (?) OR from_user_id IN (?)|, _no_category_user_ids_without_admin_user_ids, _no_category_user_ids)
                             .distinct
                             .select(:sidfmd_rule_action_id) #送信先に管理者権限を除くカスタムカテゴリ数が0になるユーザ、または送信元でカスタムカテゴリ数が0になるユーザが設定されているメール送信の操作
        ).select(:sidfmd_rule_id), #対象カテゴリに関連する操作
        SidfmdRuleCond.joins(:sidfmd_rule_c_ctag)
                      .joins("INNER JOIN sidfmd_rule_cm_ctags ON sidfmd_rule_cm_ctags.sidfmd_rule_c_ctag_id = sidfmd_rule_c_ctags.id")
                      .where(:sidfmd_rule_cm_ctags => { :sidfmd_ctag_id => SidfmdCtag.where(:sidfmd_category_id => category_ids).select(:id) })
                      .select(:sidfmd_rule_id), #カスタムタグ条件
        SidfmdRuleAction.joins(:sidfmd_rule_a_ctag)
                        .joins("INNER JOIN sidfmd_rule_am_ctags ON sidfmd_rule_am_ctags.sidfmd_rule_a_ctag_id = sidfmd_rule_a_ctags.id")
                        .where(:sidfmd_rule_am_ctags => { :sidfmd_ctag_id => SidfmdCtag.where(:sidfmd_category_id => category_ids).select(:id) })
                        .select(:sidfmd_rule_id) #カスタムタグ操作
      )
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_conds
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_conds /*ルールのcondの数*/
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_actions
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_actions /*ルールのactionの数*/      
      ")
      .select("
         (
           SELECT COUNT (*)
           FROM sidfmd_rule_conds
           INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
           INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
           WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_hosts.inv IS FALSE /*ホスト限定に設定されているホストの数*/
         ) AS count_match_sub_host_conds
      ")
      .select("
        (
          SELECT COUNT (*)
          FROM sidfmd_rule_conds
          INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
          INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
          WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_hosts.inv IS TRUE /*ホスト除外に設定されているホストの数*/
        ) AS count_unmatch_sub_host_conds
      ")
      .select("
         (
           SELECT COUNT (*)
           FROM sidfmd_rule_conds
           INNER JOIN sidfmd_rule_c_match_categories ON sidfmd_rule_conds.id = sidfmd_rule_c_match_categories.sidfmd_rule_cond_id
           INNER JOIN sidfmd_rule_cm_category_subs ON sidfmd_rule_c_match_categories.id = sidfmd_rule_cm_category_subs.sidfmd_rule_c_match_category_id
           WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_categories.inv IS FALSE /*カテゴリ限定に設定されているカテゴリの数*/
         ) AS count_match_sub_category_conds
      ")
      .select("
         (
           SELECT COUNT (*)
           FROM sidfmd_rule_conds
           INNER JOIN sidfmd_rule_c_match_categories ON sidfmd_rule_conds.id = sidfmd_rule_c_match_categories.sidfmd_rule_cond_id
           INNER JOIN sidfmd_rule_cm_category_subs ON sidfmd_rule_c_match_categories.id = sidfmd_rule_cm_category_subs.sidfmd_rule_c_match_category_id
           WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_categories.inv IS TRUE /*カテゴリ除外に設定されているカテゴリの数*/
         ) AS count_unmatch_sub_category_conds
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_chp_user_subs
          INNER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.id = sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id
          INNER JOIN sidfmd_rule_conds ON sidfmd_rule_conds.id = sidfmd_rule_c_history_posts.sidfmd_rule_cond_id
          WHERE sidfmd_rule_id = sidfmd_rules.id
        ) AS count_history_cond_user_ids /*SidfmdRuleChpUserSubの数(SidfmdRuleCHistoryPostに設定されているユーザの数)*/
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_a_mail_users
          INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id
          WHERE sidfmd_rule_id = sidfmd_rules.id
        ) AS count_mail_to_user_ids /*SidfmdRuleAMailUserの送信先に設定されているユーザの数*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, COUNT (sidfmd_rule_cm_ctags.id)], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_ctags ON sidfmd_rule_c_ctags.sidfmd_rule_cond_id = sidfmd_rule_conds.id
            INNER JOIN sidfmd_rule_cm_ctags ON sidfmd_rule_cm_ctags.sidfmd_rule_c_ctag_id = sidfmd_rule_c_ctags.id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id
            GROUP BY sidfmd_rule_conds.id
          )
        ) AS cond_ctag_counts /*ルールのカスタムタグ条件毎の、設定されているカスタムタグの数*/
      ")
      .select("
        (
          SELECT COUNT(sidfmd_rule_am_ctags.id)
          FROM sidfmd_rule_actions
          INNER JOIN sidfmd_rule_a_ctags ON sidfmd_rule_a_ctags.sidfmd_rule_action_id = sidfmd_rule_actions.id
          INNER JOIN sidfmd_rule_am_ctags ON sidfmd_rule_am_ctags.sidfmd_rule_a_ctag_id = sidfmd_rule_a_ctags.id
          WHERE sidfmd_rules.id = sidfmd_rule_actions.sidfmd_rule_id
          GROUP BY sidfmd_rule_actions.id
        ) AS action_ctag_count /*ルールのカスタムタグの操作に設定されているカスタムタグの数*/
      ")
      .select(:id, :name, :sidfmd_category_id)
    else
      SidfmdRule.where(:id => 0)
    end

    if opts[:delete_category_ids] then
      opts[:delete_category_ids] = [ 0 ] if opts[:delete_category_ids].size == 0
      _rules = _rules.select("
        (
          SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_triggers.id, (CASE WHEN sidfmd_rule_t_vulsyncs.sidfmd_category_id IS NULL THEN 0 ELSE sidfmd_rule_t_vulsyncs.sidfmd_category_id END), sidfmd_rule_t_vulsyncs.sidfmd_host_id], ':')
          FROM sidfmd_rule_triggers INNER JOIN sidfmd_rule_t_vulsyncs ON sidfmd_rule_triggers.id = sidfmd_rule_t_vulsyncs.sidfmd_rule_trigger_id
          WHERE sidfmd_rule_triggers.sidfmd_rule_id = sidfmd_rules.id AND
                (
                  sidfmd_rule_t_vulsyncs.sidfmd_host_id IN (SELECT unnest (ARRAY[#{(_no_category_host_ids.any? ? _no_category_host_ids : [ 0 ]).join(', ')}])) OR
                  sidfmd_rule_t_vulsyncs.sidfmd_category_id IN (SELECT unnest (ARRAY[#{opts[:delete_category_ids].join(', ')}]))
                )
        ) AS delete_category_trigger_vulsync /*設定されているカテゴリか、ホストのカスタムカテゴリが無くなる脆弱性同期時実行のタイミング*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_category_subs.id, sidfmd_rule_cm_category_subs.sidfmd_category_id], ':') 
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_categories ON sidfmd_rule_conds.id = sidfmd_rule_c_match_categories.sidfmd_rule_cond_id
            INNER JOIN sidfmd_rule_cm_category_subs ON sidfmd_rule_c_match_categories.id = sidfmd_rule_cm_category_subs.sidfmd_rule_c_match_category_id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND
                  sidfmd_rule_c_match_categories.inv IS FALSE AND
                  sidfmd_rule_cm_category_subs.sidfmd_category_id IN (SELECT unnest(ARRAY[#{opts[:delete_category_ids].join(', ')}]))
          )
        ) AS orig_delete_match_category_conds /*削除対象のカテゴリが設定されているカテゴリ限定*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_category_subs.id, sidfmd_rule_cm_category_subs.sidfmd_category_id], ':') 
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_categories ON sidfmd_rule_conds.id = sidfmd_rule_c_match_categories.sidfmd_rule_cond_id
            INNER JOIN sidfmd_rule_cm_category_subs ON sidfmd_rule_c_match_categories.id = sidfmd_rule_cm_category_subs.sidfmd_rule_c_match_category_id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND
                  sidfmd_rule_c_match_categories.inv IS TRUE AND
                  sidfmd_rule_cm_category_subs.sidfmd_category_id IN (SELECT unnest(ARRAY[#{opts[:delete_category_ids].join(', ')}]))
          )
        ) AS orig_delete_unmatch_category_conds /*削除対象のカテゴリが設定されているカテゴリ除外*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_host_subs.id, sidfmd_rule_cm_host_subs.sidfmd_host_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
            INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND 
                  sidfmd_rule_c_match_hosts.inv IS FALSE AND
                  sidfmd_rule_cm_host_subs.sidfmd_host_id IN (SELECT unnest (ARRAY[#{(_no_category_host_ids.any? ? _no_category_host_ids : [ 0 ]).join(', ')}]))
          )
        ) AS orig_no_category_match_host_conds /*カテゴリ数が0になるホスト限定*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_host_subs.id, sidfmd_rule_cm_host_subs.sidfmd_host_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_hosts ON sidfmd_rule_conds.id = sidfmd_rule_c_match_hosts.sidfmd_rule_cond_id
            INNER JOIN sidfmd_rule_cm_host_subs ON sidfmd_rule_c_match_hosts.id = sidfmd_rule_cm_host_subs.sidfmd_rule_c_match_host_id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND 
                  sidfmd_rule_c_match_hosts.inv IS TRUE AND
                  sidfmd_rule_cm_host_subs.sidfmd_host_id IN (SELECT unnest (ARRAY[#{(_no_category_host_ids.any? ? _no_category_host_ids : [ 0 ]).join(', ')}]))
          )
        ) AS orig_no_category_unmatch_host_conds /*カテゴリ数が0になるホスト除外*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_chp_user_subs.id, sidfmd_rule_chp_user_subs.user_id], ':')
            FROM sidfmd_rule_chp_user_subs
            INNER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.id = sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id
            INNER JOIN sidfmd_rule_conds ON sidfmd_rule_conds.id = sidfmd_rule_c_history_posts.sidfmd_rule_cond_id
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_chp_user_subs.user_id IN (SELECT unnest(ARRAY[#{(_no_category_user_ids_without_admin_user_ids.any? ? _no_category_user_ids_without_admin_user_ids : [ 0 ]).join(',')}]))
          )
        ) AS orig_no_category_history_cond_chp_users /*管理者権限を除いて、カテゴリが無くなるユーザが設定されているSidfmdRuleChpUserSubのidとuser_idとsidfmd_rule_cond_id*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.user_id IN (SELECT unnest(ARRAY[#{(_no_category_user_ids_without_admin_user_ids.any? ? _no_category_user_ids_without_admin_user_ids : [ 0 ]).join(',')}]))
          )
        ) AS orig_no_category_mail_users_to_actions /*管理者権限を除いて、SidfmdRuleAMailUserの送信先に設定されているカテゴリ数が0になるユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.from_user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.from_user_id IN (SELECT unnest(ARRAY[#{(_no_category_user_ids_without_admin_user_ids.any? ? _no_category_user_ids_without_admin_user_ids : [ 0 ]).join(',')}]))
          )
        ) AS orig_no_category_mail_users_from_actions /*管理者権限を除いて、SidfmdRuleAMailUserの送信元に設定されているカテゴリ数が0になるユーザに関連するactions*/ 
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_categories.sidfmd_category_id], ':')
            FROM sidfmd_rule_a_mail_categories
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_categories.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_categories.sidfmd_category_id IN (SELECT unnest(ARRAY[#{opts[:delete_category_ids].join(',')}]))
          )
        ) AS orig_no_category_mail_categories_to_actions /*SidfmdRuleAMailCategoryの送信先が削除対象のカテゴリになっているactions(通常は1つのみ)*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_categories.from_user_id], ':')
            FROM sidfmd_rule_a_mail_categories
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_categories.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_categories.from_user_id IN (SELECT unnest(ARRAY[#{(_no_category_user_ids_without_admin_user_ids.any? ? _no_category_user_ids_without_admin_user_ids : [ 0 ]).join(',')}]))
          )
        ) AS orig_no_category_mail_categories_from_actions /*管理者権限を除いて、SidfmdRuleAMailCategoryの送信元に設定されているカテゴリ数が0になるユーザに関連するactions(通常は1つのみ)*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_ctags.id, sidfmd_rule_cm_ctags.sidfmd_ctag_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_ctags ON sidfmd_rule_c_ctags.sidfmd_rule_cond_id = sidfmd_rule_conds.id
            INNER JOIN sidfmd_rule_cm_ctags ON sidfmd_rule_cm_ctags.sidfmd_rule_c_ctag_id = sidfmd_rule_c_ctags.id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND
              sidfmd_rule_cm_ctags.sidfmd_ctag_id IN (
                SELECT id FROM sidfmd_ctags WHERE sidfmd_category_id IN (SELECT unnest (ARRAY[#{opts[:delete_category_ids].join(', ')}]))
              )
          )
        ) AS delete_cond_ctags /*削除されるカスタムタグとそのカスタムタグ条件*/
      ")
      .select("
        ARRAY(
          (
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_am_ctags.id, sidfmd_rule_am_ctags.sidfmd_ctag_id], ':')
            FROM sidfmd_rule_actions
            INNER JOIN sidfmd_rule_a_ctags ON sidfmd_rule_a_ctags.sidfmd_rule_action_id = sidfmd_rule_actions.id
            INNER JOIN sidfmd_rule_am_ctags ON sidfmd_rule_am_ctags.sidfmd_rule_a_ctag_id = sidfmd_rule_a_ctags.id
            WHERE sidfmd_rules.id = sidfmd_rule_actions.sidfmd_rule_id AND
              sidfmd_rule_am_ctags.sidfmd_ctag_id IN (
                SELECT id FROM sidfmd_ctags WHERE sidfmd_category_id IN (SELECT unnest (ARRAY[#{opts[:delete_category_ids].join(', ')}]))
              )
          )
        ) AS delete_action_ctags /*削除されるカスタムタグとそのカスタムタグ操作*/
      ")
    end

    _match_host_cond_name = RULE_COND_NAMES['host-match']
    _unmatch_host_cond_name = RULE_COND_NAMES['host-unmatch']
    _match_category_cond_name = RULE_COND_NAMES['category-match']
    _unmatch_category_cond_name = RULE_COND_NAMES['category-unmatch']
    _mail_cat_action_name = RULE_ACTION_NAMES['send-mail-category']
    _mail_user_action_name = RULE_ACTION_NAMES['send-mail-user']
    _history_cond_name = RULE_COND_NAMES['history-create']
    _ctag_cond_name = RULE_COND_NAMES['ctag']
    _ctag_action_name = RULE_ACTION_NAMES['set-ctag']
    _rule_messages = {}
    _delete_rules = []
    _delete_conds = []
    _delete_actions = []
    _delete_category_cond_subs = []
    _delete_host_cond_subs = []
    _delete_chp_user_cond_subs = []
    _delete_cm_ctags = []
    _delete_am_ctags = []
   
    create_m_host_cond_object = ->(conds){
      obj = conds.map{|cond|
        cols = cond.split(':')
        {
          :cond_id => cols[0].to_i,
          :sub_host_cond_id => cols[1].to_i,
          :sidfmd_host_id => cols[2].to_i,
        }
      }
    }

    create_m_category_cond_object = ->(conds){
      obj = conds.map{|cond|
        cols = cond.split(':')
        {
          :cond_id => cols[0].to_i,
          :sub_category_cond_id => cols[1].to_i,
          :sidfmd_category_id => cols[2].to_i,
        }
      }
    }

    create_history_cond_object = ->(conds){
      obj = conds.map{|cond|
        cols = cond.split(':')
        {
          :cond_id => cols[0].to_i,
          :chp_user_sub_id => cols[1].to_i,
          :user_id => cols[2].to_i,
        }
      }
    }

    create_mail_action_object = ->(actions){
      obj = actions.map{|action|
        actionid_userid = action.split(':')
        { :action_id => actionid_userid[0].to_i, :user_id => actionid_userid[1].to_i }
      }
    }

    create_trigger_vulsync_object = -> (trigger){
      cols = trigger.split(':')
      obj = {
        :trigger_id => cols[0].to_i,
        :sidfmd_category_id => cols[1].to_i,
        :sidfmd_host_id => cols[2].to_i,
      }
    }

    create_delete_rule_object = -> (rule, opts){
      { :rule_id => rule.id, :opts => Array.wrap(opts) }
    }

    create_delete_category_rule_opts = -> (delete_category_id){
      { :sidfmd_category_id => delete_category_id }
    }

    get_category_names = ->(category_ids){
      SidfmdCategory.where(:id => category_ids).order('LOWER(name COLLATE "C")').map{|cat| cat.name}
    }

    get_hostnames = -> (host_ids){
      SidfmdHost.where(:id => host_ids).order(%|
        CASE WHEN ext_name IS NULL THEN LOWER(hostname COLLATE "C") ELSE LOWER(ext_name COLLATE "C") END
      |).map{|host|
        "「#{(host.ext_name ? (host.ext_name + '(ホスト名: ' + host.hostname + ')') : host.hostname)}」"
      }
    }
 
    get_user_names = ->(user_ids){
       User.where(:id => user_ids)
           .order('LOWER(lastname COLLATE "C")')
           .map{|user| "#{user.lastname} #{user.firstname}(ログイン名: #{user.login})"}
    }

    get_reason_of_delete_trigger_vulsync = ->(trigger_vulsync){
      reason = if opts[:delete_category_ids].include?(trigger_vulsync[:sidfmd_category_id]) then
        "タイミングに、削除対象のカテゴリが設定されているため。"
      else
        "カテゴリの削除によって、タイミングに設定されているホストのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
      end
    }

    get_reason_of_delete_category_cond = ->(){
      "条件に設定されている全てのカテゴリが削除されるため。"
    }

    get_reason_of_delete_category_related_cond = ->(items){
      "カテゴリの削除によって、条件に設定されている全ての#{items.join('と')}のカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
    }

    get_reason_of_delete_mail_users_action = ->(is_from){
      reason = if is_from then
        "カテゴリの削除によって、送信元に設定されているユーザのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
      else
        "カテゴリの削除によって、送信先に設定されている全てのユーザのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
      end
    }

    get_reason_of_delete_mail_category_action = ->(is_from){
      reason = if is_from then
        "カテゴリの削除によって、送信元に設定されているユーザのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
      else
        "送信先に設定されているカテゴリが削除されるため。"
      end
    }

    get_reason_of_delete_rule_by_conds = -> (match_category_conds: [], unmatch_category_conds: [], match_host_conds: [], unmatch_host_conds: [], history_user_conds: [], ctag_conds: []){
      reason_item_names = []
      is_m_category_cond = (match_category_conds.any? or unmatch_category_conds.any?)
      is_m_host_cond = (match_host_conds.any? or unmatch_host_conds.any?)
      is_history_user_cond = history_user_conds.any?
      is_ctag_cond = ctag_conds.any?
      reason_item_names << "「#{_match_category_cond_name}」" if match_category_conds.any?
      reason_item_names << "「#{_unmatch_category_cond_name}」" if unmatch_category_conds.any?
      reason_item_names << "「#{_match_host_cond_name}」" if match_host_conds.any?
      reason_item_names << "「#{_unmatch_host_cond_name}」" if unmatch_host_conds.any?
      reason_item_names << "「#{_history_cond_name}」" if history_user_conds.any?
      reason_item_names << "「#{_ctag_cond_name}」" if is_ctag_cond
      return '' if reason_item_names.size == 0
      items = []
      items << 'ホスト' if is_m_host_cond
      items << 'ユーザ' if is_history_user_cond
      reason = "ルールの条件が#{reason_item_names.join('・')}のみで、"
      if is_m_category_cond then
        reason += get_reason_of_delete_category_cond.()
      end
      if is_m_host_cond or is_history_user_cond then
        reason += "#{is_m_category_cond ? 'また、' : ''}#{get_reason_of_delete_category_related_cond.(items)}"
      end
      if is_ctag_cond then
        reason += "カテゴリの削除によって、条件で選択されている全てのカスタムタグが削除されるため。"
      end
      reason
    }

    get_cond_ctag_data = ->(cond_ctag){
      cond_ctag = cond_ctag.split(':')
      return {
        :cond_id => cond_ctag[0].to_i,
        :cm_ctag_id => cond_ctag[1].to_i,
        :ctag_id => cond_ctag[2].to_i,
      }
    }

    get_action_ctag_data = ->(action_ctag){
      action_ctag = action_ctag.split(':')
      return {
        :action_id => action_ctag[0].to_i,
        :am_ctag_id => action_ctag[1].to_i,
        :ctag_id => action_ctag[2].to_i,
      }
    }

    get_delete_each_cond_ctags = ->(rule_delete_cond_ctags){
      rule_delete_each_cond_ctags = {}
      rule_delete_cond_ctags.each do |ctag|
        rule_delete_each_cond_ctags[ctag[:cond_id]] = [] unless rule_delete_each_cond_ctags[ctag[:cond_id]]
        rule_delete_each_cond_ctags[ctag[:cond_id]] << ctag
      end
      return rule_delete_each_cond_ctags
    }

    get_each_cond_ctags_count = ->(rule_cond_ctag_counts){
      rule_each_cond_ctags_counts = {}
      rule_cond_ctag_counts.each do |cond_ctag_counts|
        cond_ctag_counts = cond_ctag_counts.split(':')
        cond_id = cond_ctag_counts[0].to_i
        ctag_count = cond_ctag_counts[1].to_i
        rule_each_cond_ctags_counts[cond_id] = ctag_count
      end
      return rule_each_cond_ctags_counts
    }

    get_ctag_names = ->(ctags){
      SidfmdCtag.where(:id => ctags.map{|ctag| ctag[:ctag_id]})
                .order(:priority => :desc)
                .order('LOWER(keyword COLLATE "C") ASC')
                .map{|c| "「#{c.keyword}」" }
    }

    #今のところ削除時のみルールをチェックするため、
    #deleteでもimportでも同じチェックをする。
    #他にチェックが必要になったら、opts[:action]で処理を分ける
    _rules.each do |rule|
      #ルールのカテゴリのカテゴリが削除されるならルール削除
      if opts[:delete_category_ids].include?(rule.sidfmd_category_id) then
        _delete_rules << create_delete_rule_object.(rule, create_delete_category_rule_opts.(rule.sidfmd_category_id))
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => "ルールのカテゴリに、削除対象のカテゴリが設定されているため。"
          }
        end
        next
      end
      
      if rule.delete_category_trigger_vulsync then
        trigger_vulsync = create_trigger_vulsync_object.(rule.delete_category_trigger_vulsync)
        _delete_rules << create_delete_rule_object.(rule, trigger_vulsync)
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => get_reason_of_delete_trigger_vulsync.(trigger_vulsync)
          }
        end
        next
      end

      delete_match_category_conds = create_m_category_cond_object.(rule.orig_delete_match_category_conds) #削除されるカテゴリのカテゴリ限定
      delete_unmatch_category_conds = create_m_category_cond_object.(rule.orig_delete_unmatch_category_conds) #削除されるカテゴリのカテゴリ除外
      rule_delete_m_category_conds_size = 0
      rule_delete_m_category_conds_size += 1 if delete_match_category_conds.any?
      rule_delete_m_category_conds_size += 1 if delete_unmatch_category_conds.any?

      if rule.count_conds == rule_delete_m_category_conds_size and
         delete_match_category_conds.size == rule.count_match_sub_category_conds and 
         delete_unmatch_category_conds.size == rule.count_unmatch_sub_category_conds
      then
        #条件がカテゴリ限定/カテゴリ除外のみで、削除されて条件がなくなるなら、ルール削除
        _delete_rules << create_delete_rule_object.(rule, (delete_match_category_conds + delete_unmatch_category_conds))
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => get_reason_of_delete_rule_by_conds.(
              match_category_conds: delete_match_category_conds, unmatch_category_conds: delete_unmatch_category_conds
            )
          }
        end
        next
      end
 
      delete_match_host_conds = create_m_host_cond_object.(rule.orig_no_category_match_host_conds) #カスタムカテゴリの数が0になるホストのホスト限定
      delete_unmatch_host_conds = create_m_host_cond_object.(rule.orig_no_category_unmatch_host_conds) #カスタムカテゴリの数が0になるホストのホスト除外
      rule_delete_m_host_conds_size = 0
      rule_delete_m_host_conds_size += 1 if delete_match_host_conds.any?
      rule_delete_m_host_conds_size += 1 if delete_unmatch_host_conds.any?

      if rule.count_conds == rule_delete_m_host_conds_size and
         delete_match_host_conds.size == rule.count_match_sub_host_conds and
         delete_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds 
      then
        #条件がホスト限定/ホスト除外のみで、全てのホストのカスタムカテゴリが0になるなら、ルール削除
        _delete_rules << create_delete_rule_object.(rule, (delete_match_host_conds + delete_unmatch_host_conds))
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => get_reason_of_delete_rule_by_conds.(
              match_host_conds: delete_match_host_conds, unmatch_host_conds: delete_unmatch_host_conds
            )
          }
        end
        next
      end

      if rule.count_conds == (rule_delete_m_category_conds_size + rule_delete_m_host_conds_size) and
         delete_match_category_conds.size == rule.count_match_sub_category_conds and
         delete_unmatch_category_conds.size == rule.count_unmatch_sub_category_conds and
         delete_match_host_conds.size == rule.count_match_sub_host_conds and
         delete_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds
      then
        #条件がカテゴリ限定/カテゴリ除外/ホスト限定/ホスト除外のみで、削除されて条件がなくなるなら、ルール削除
        _delete_rules << create_delete_rule_object.(
          rule,
          (delete_match_category_conds + delete_unmatch_category_conds + delete_match_host_conds + delete_unmatch_host_conds)
        )
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => get_reason_of_delete_rule_by_conds.(
              match_category_conds: delete_match_category_conds,
              unmatch_category_conds: delete_unmatch_category_conds,
              match_host_conds: delete_match_host_conds,
              unmatch_host_conds: delete_unmatch_host_conds
            )
          }
        end
        next
      end

      delete_history_cond_chp_users = create_history_cond_object.(rule.orig_no_category_history_cond_chp_users) #管理者権限を除いて、カスタムカテゴリが0になるユーザの履歴追加 
      rule_delete_history_chp_users_cond_size = (delete_history_cond_chp_users.any? ? 1 : 0)
     
      if rule.count_conds == rule_delete_history_chp_users_cond_size and
         delete_history_cond_chp_users.size == rule.count_history_cond_user_ids 
      then
        #条件が履歴追加のみで、削除されて条件がなくなるなら、ルール削除
        _delete_rules << create_delete_rule_object.(rule, delete_history_cond_chp_users)
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => get_reason_of_delete_rule_by_conds.(history_user_conds: delete_history_cond_chp_users)
          }
        end
        next
      end

      #カスタムタグ条件のルール削除チェック
      _rule_delete_cond_ctags = rule.delete_cond_ctags.map{|cond_ctag| get_cond_ctag_data.(cond_ctag) } #条件に設定されている削除するカスタムタグ
      _rule_delete_each_cond_ctags = get_delete_each_cond_ctags.(_rule_delete_cond_ctags) #カスタムタグ条件毎の削除するカスタムタグ
      _rule_each_cond_ctags_count = get_each_cond_ctags_count.(rule.cond_ctag_counts) #カスタムタグ条件毎のカスタムタグ数の数
      _rule_delete_ctag_cond_ids = [] #削除するカスタムタグ条件のid

      _rule_delete_each_cond_ctags.each do |cond_id, ctags|
        if _rule_each_cond_ctags_count[cond_id] == ctags.size then
          _rule_delete_ctag_cond_ids << cond_id
        end
      end

      _rule_delete_ctag_cond_size = _rule_delete_ctag_cond_ids.size
      if _rule_delete_ctag_cond_size == rule.count_conds then
        #条件がカスタムタグのみで、削除されて条件がなくなるなら、ルール削除
        _delete_rules << create_delete_rule_object.(rule, _rule_delete_cond_ctags)
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => get_reason_of_delete_rule_by_conds.(ctag_conds: _rule_delete_ctag_cond_ids)
          }
        end
        next
      end

      if rule.count_conds == (rule_delete_m_category_conds_size + rule_delete_m_host_conds_size + rule_delete_history_chp_users_cond_size + _rule_delete_ctag_cond_size) and
         delete_match_category_conds.size == rule.count_match_sub_category_conds and
         delete_unmatch_category_conds.size == rule.count_unmatch_sub_category_conds and
         delete_match_host_conds.size == rule.count_match_sub_host_conds and
         delete_unmatch_host_conds.size == rule.count_unmatch_sub_host_conds and
         delete_history_cond_chp_users.size == rule.count_history_cond_user_ids
      then
        #条件がカテゴリ限定/カテゴリ除外/ホスト限定/ホスト除外/履歴追加/カスタムタグのみで、削除されて条件がなくなるなら、ルール削除
        #ルールの条件数と削除する条件数が一致するかどうかをここでチェック。
        #カテゴリ関連のルールの条件が追加されて、条件数によるルールの削除をチェックする場合は、上記のif文に条件を足し、get_reason_of_delete_rule_by_condsに追加実装するようにすると楽
        _delete_rules << create_delete_rule_object.(
          rule,
          (delete_match_category_conds + delete_unmatch_category_conds + delete_match_host_conds + delete_unmatch_host_conds + delete_history_cond_chp_users + _rule_delete_cond_ctags)
        )
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => get_reason_of_delete_rule_by_conds.(
              match_category_conds: delete_match_category_conds,
              unmatch_category_conds: delete_unmatch_category_conds,
              match_host_conds: delete_match_host_conds,
              unmatch_host_conds: delete_unmatch_host_conds,
              history_user_conds: delete_history_cond_chp_users,
              ctag_conds: _rule_delete_ctag_cond_ids
            )
          }
        end
        next
      end

      delete_mail_users_from_actions = create_mail_action_object.(rule.orig_no_category_mail_users_from_actions) #送信元に設定されているユーザのカテゴリ数が0になるメール送信(ユーザ)
      
      if rule.count_actions == delete_mail_users_from_actions.size then
        #操作がメール送信(ユーザ)のみで、送信元のユーザのカテゴリがなくなるならルール削除
        _delete_rules << create_delete_rule_object.(rule, delete_mail_users_from_actions)
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => "操作が「#{_mail_user_action_name}」のみで、#{get_reason_of_delete_mail_users_action.(true)}"
          }
        end
        next
      end

      delete_mail_users_to_actions = create_mail_action_object.(rule.orig_no_category_mail_users_to_actions) #管理者権限を除いて、送信先に設定されているユーザのカテゴリ数が0になるメール送信(ユーザ)
      
      if rule.count_actions == delete_mail_users_to_actions.size then
        #操作がメール送信(ユーザ)のみで、送信先の全てのユーザのカテゴリがなくなるならルール削除
        _delete_rules << create_delete_rule_object.(rule, delete_mail_users_to_actions)
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => "操作が「#{_mail_user_action_name}」のみで、#{get_reason_of_delete_mail_users_action.(false)}"
          }
        end
        next
      end

      delete_mail_category_from_actions = create_mail_action_object.(rule.orig_no_category_mail_categories_from_actions) #送信元に設定されているユーザのカテゴリ数が0になるメール送信(カテゴリ)
      
      if rule.count_actions == delete_mail_category_from_actions.size then
        #操作がメール送信(カテゴリ)のみで、送信元のユーザのカテゴリがなくなるならルール削除
        _delete_rules << create_delete_rule_object.(rule, delete_mail_category_from_actions)
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => "操作が「#{_mail_cat_action_name}」のみで、#{get_reason_of_delete_mail_category_action.(true)}"
          }
        end
        next
      end

      delete_mail_category_to_actions = create_mail_action_object.(rule.orig_no_category_mail_categories_to_actions) #送信先に設定されているカテゴリが削除されるメール送信(カテゴリ)
 
      if rule.count_actions == delete_mail_category_to_actions.size then
        #操作がメール送信(カテゴリ)のみで、送信先のカテゴリが削除されるならルール削除
        _delete_rules << create_delete_rule_object.(rule, delete_mail_category_from_actions)
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => "操作が「#{_mail_cat_action_name}」のみで、#{get_reason_of_delete_mail_category_action.(false)}"
          }
        end
        next
      end
     
      all_mail_actions = (delete_mail_users_from_actions + delete_mail_users_to_actions + delete_mail_category_from_actions + delete_mail_category_to_actions)

      _rule_delete_action_ctags = rule.delete_action_ctags.map{|action_ctag| get_action_ctag_data.(action_ctag) } #操作に設定されている削除するカスタムタグ
      _is_action_ctag_deletion = rule.action_ctag_count != 0 && (rule.action_ctag_count == _rule_delete_action_ctags.size)
      
      if _is_action_ctag_deletion && rule.count_actions == 1 then
        #操作がカスタムタグのみで、操作が削除されるならルール削除
        _delete_rules << create_delete_rule_object.(rule, _rule_delete_action_ctags)
        if opts[:messages_necessary] then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => "操作に「#{_ctag_action_name}」のみが設定されており、カテゴリの削除によって、選択されている全てのカスタムタグが削除されるため。"
          }
        end
        next
      end

      if rule.count_actions == (all_mail_actions.uniq{|action| action[:action_id]}.size + (_is_action_ctag_deletion ? 1 : 0)) then
        #操作がメール送信(ユーザ)/メール送信(カテゴリ)/カスタムタグ設定のみで、操作が削除されるならルール削除
        _delete_rules << create_delete_rule_object.(rule, (all_mail_actions + _rule_delete_action_ctags))
        if opts[:messages_necessary] then
          reasons = []
          items = []
          if (delete_mail_users_from_actions.any? and _no_category_user_ids.include?(delete_mail_users_from_actions.first[:user_id])) then
            reasons << "カテゴリの削除によって、「#{_mail_user_action_name}」の送信元に設定されているユーザのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
            items << "「#{_mail_user_action_name}」"
          elsif delete_mail_users_to_actions.any? then
            reasons << "カテゴリの削除によって、「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
            items << "「#{_mail_user_action_name}」"
          end
          if (delete_mail_category_from_actions.any? and (delete_mail_category_from_actions.map{|ac| ac[:user_id]} - _no_category_user_ids).size == 0) then
            reasons << "カテゴリの削除によって、「#{_mail_cat_action_name}」の送信元に設定されているユーザのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
            items << "「#{_mail_cat_action_name}」"
          elsif delete_mail_category_to_actions.any? then
            reasons << "「#{_mail_cat_action_name}」の送信先に設定されているカテゴリが削除されるため。"
            items << "「#{_mail_cat_action_name}」"
          end
          if _is_action_ctag_deletion then
            reasons << "カテゴリの削除によって、「#{_ctag_action_name}」で選択されている全てのカスタムタグが削除されるため。"
            items << "「#{_ctag_action_name}」"
          end
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => "操作が#{items.join('・')}のみで、#{reasons.join()}" 
          }
        end
        next
      end

      item_names = []
      content_msgs = []
      reason_msgs = []

      #条件のチェック
      if delete_match_category_conds.any? then
        if rule.count_match_sub_category_conds == delete_match_category_conds.size then
          _delete_conds += delete_match_category_conds
          if opts[:messages_necessary] then
            item_names << "条件"
            content_msgs << "「#{_match_category_cond_name}」が削除されます。"
            reason_msgs << get_reason_of_delete_category_cond.()
          end
        else
          _delete_category_cond_subs += delete_match_category_conds
          if opts[:messages_necessary] then
            category_names = get_category_names.(delete_match_category_conds.map{|c| c[:sidfmd_category_id]})
            item_names << "条件"
            content_msgs << "「#{_match_category_cond_name}」から、#{category_names.join(', ')}が削除されます。"
            reason_msgs << "カテゴリが削除されるため。"
          end
        end
      end
      
      if delete_unmatch_category_conds.any? then
        if rule.count_unmatch_sub_category_conds == delete_unmatch_category_conds.size then
          _delete_conds += delete_unmatch_category_conds
          if opts[:messages_necessary] then
            item_names << "条件"
            content_msgs << "「#{_unmatch_category_cond_name}」が削除されます。"
            reason_msgs << get_reason_of_delete_category_cond.()
          end
        else
          _delete_category_cond_subs += delete_unmatch_category_conds
          if opts[:messages_necessary] then
            category_names = get_category_names.(delete_unmatch_category_conds.map{|c| c[:sidfmd_category_id]})
            item_names << "条件"
            content_msgs << "「#{_unmatch_category_cond_name}」から、#{category_names.join(', ')}が削除されます。"
            reason_msgs << "カテゴリが削除されるため。"
          end
        end
      end

      if delete_match_host_conds.any? then
        if rule.count_match_sub_host_conds == delete_match_host_conds.size then
          _delete_conds += delete_match_host_conds
          if opts[:messages_necessary] then
            item_names << "条件"
            content_msgs << "「#{_match_host_cond_name}」が削除されます。"
            reason_msgs << get_reason_of_delete_category_related_cond.(['ホスト'])
          end
        else
          _delete_host_cond_subs += delete_match_host_conds
          if opts[:messages_necessary] then
            hostnames = get_hostnames.(delete_match_host_conds.map{|c| c[:sidfmd_host_id]})
            item_names << "条件"
            content_msgs << "「#{_match_host_cond_name}」から、#{hostnames.join(', ')}が削除されます。"
            reason_msgs << "カテゴリの削除によって、ホストのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
          end
        end
      end

      if delete_unmatch_host_conds.any? then
        if rule.count_unmatch_sub_host_conds == delete_unmatch_host_conds.size then
          _delete_conds += delete_unmatch_host_conds
          if opts[:messages_necessary] then
            item_names << "条件"
            content_msgs << "「#{_unmatch_host_cond_name}」が削除されます。"
            reason_msgs << get_reason_of_delete_category_related_cond.(['ホスト'])
          end
        else
          _delete_host_cond_subs += delete_unmatch_host_conds
          if opts[:messages_necessary] then
            hostnames = get_hostnames.(delete_unmatch_host_conds.map{|c| c[:sidfmd_host_id]})
            item_names << "条件"
            content_msgs << "「#{_unmatch_host_cond_name}」から、#{hostnames.join(', ')}が削除されます。"
            reason_msgs << "カテゴリの削除によって、ホストのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
          end
        end
      end

      if delete_history_cond_chp_users.any? then
        if rule.count_history_cond_user_ids == delete_history_cond_chp_users.size then
          _delete_conds += delete_history_cond_chp_users
          if opts[:messages_necessary] then
            item_names << "条件"
            content_msgs << "「#{_history_cond_name}」が削除されます。"
            reason_msgs << get_reason_of_delete_category_related_cond.(['ユーザ'])
          end
        else
          _delete_chp_user_cond_subs += delete_history_cond_chp_users
          if opts[:messages_necessary] then
            user_names = get_user_names.(delete_history_cond_chp_users.map{|c| c[:user_id]})
            item_names << "条件"
            content_msgs << "「#{_history_cond_name}」から、#{user_names.join(', ')}が削除されます。"
            reason_msgs << "カテゴリの削除によって、ユーザのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
          end
        end
      end

      if _rule_delete_cond_ctags.any? then
        rsn_msg = "カテゴリの削除によって、「#{_ctag_cond_name}」条件で選択されているカスタムタグが削除されるため。"
        all_ctag_cond_count = _rule_each_cond_ctags_count.size
        _delete_conds += _rule_delete_cond_ctags.select{|c| _rule_delete_ctag_cond_ids.include?(c[:cond_id])} #条件自体削除
        delete_ctags_of_conds = _rule_delete_cond_ctags.select{|c| !_rule_delete_ctag_cond_ids.include?(c[:cond_id])} #条件自体は削除されない削除対象のカスタムタグ
        _delete_cm_ctags += delete_ctags_of_conds if delete_ctags_of_conds.any?
        if opts[:messages_necessary] then
          del_ctag_names = delete_ctags_of_conds.any? ? get_ctag_names.(delete_ctags_of_conds) : []
          if _rule_delete_ctag_cond_ids.any? then
            item_names << "条件"
            if del_ctag_names.any? then
              content_msgs << "#{_rule_delete_ctag_cond_ids.size}つの「#{_ctag_cond_name}」条件が削除され、残りの「#{_ctag_cond_name}」条件から、カスタムタグ名#{del_ctag_names.join(', ')}が削除されます。"
            else
              content_msgs << (all_ctag_cond_count == _rule_delete_cond_ctags.size ?
                "「#{_ctag_cond_name}」条件が削除されます。" :
                "#{(_rule_delete_ctag_cond_ids.size == 1 || all_ctag_cond_count == _rule_delete_ctag_cond_ids.size) ? '' : "#{_rule_delete_ctag_cond_ids.size}つの"}「#{_ctag_cond_name}」条件が削除されます。"
              )
            end
            reason_msgs << rsn_msg
          else
            if del_ctag_names.any? then
              item_names << "条件"
              content_msgs << "「#{_ctag_cond_name}」条件から、カスタムタグ名#{del_ctag_names.join(', ')}が削除されます。"
              reason_msgs << rsn_msg
            end
          end
        end
      end

      #操作のチェック
      #メール送信の操作は、送信元ユーザ削除なら送信先はチェックしない
      if delete_mail_users_from_actions.any? then
        _delete_actions += delete_mail_users_from_actions
        if opts[:messages_necessary] then
          item_names << "操作"
          content_msgs << "「#{_mail_user_action_name}」が削除されます。"
          reason_msgs << get_reason_of_delete_mail_users_action.(true)
        end
      elsif delete_mail_users_to_actions.any? then
        _delete_actions += delete_mail_users_to_actions
        if opts[:messages_necessary] then
          item_names << "操作"
          if rule.count_mail_to_user_ids == delete_mail_users_to_actions.size then
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << get_reason_of_delete_mail_users_action.(false)
          else
            user_names = get_user_names.(delete_mail_users_to_actions.map{|c| c[:user_id]})
            content_msgs << "「#{_mail_user_action_name}」から、#{user_names.join(', ')}が削除されます。"
            reason_msgs << "カテゴリの削除によって、ユーザのカスタムカテゴリが、ルールのカテゴリと一致しなくなるため。"
          end
        end   
      end

      if delete_mail_category_from_actions.any? then
        _delete_actions += delete_mail_category_from_actions
        if opts[:messages_necessary] then
          item_names << "操作"
          content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
          reason_msgs << get_reason_of_delete_mail_category_action.(true)
        end
      elsif delete_mail_category_to_actions.any? then
        _delete_actions += delete_mail_category_to_actions
        if opts[:messages_necessary] then
          item_names << "操作"
          content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
          reason_msgs << get_reason_of_delete_mail_category_action.(false)
        end
      end

      if _rule_delete_action_ctags.any? then
        if _is_action_ctag_deletion then
          _delete_actions += _rule_delete_action_ctags
          if opts[:messages_necessary] then
            item_names << "操作"
            content_msgs << "「#{_ctag_action_name}」が削除されます。"
            reason_msgs << "カテゴリの削除によって、「#{_ctag_action_name}」で選択されている全てのカスタムタグが削除されるため。"
          end
        else
          _delete_am_ctags += _rule_delete_action_ctags
          if opts[:messages_necessary] then
            del_ctag_names = get_ctag_names.(_rule_delete_action_ctags)
            if del_ctag_names.any? then
              item_names << "操作"
              content_msgs << "「#{_ctag_action_name}」から、カスタムタグ名#{del_ctag_names.join(', ')}が削除されます。"
              reason_msgs << "カテゴリの削除によって、「#{_ctag_action_name}」で選択されているカスタムタグが削除されるため。"
            end
          end
        end
      end

      if opts[:messages_necessary] and item_names.any? then
        _rule_messages[rule.id] = {
          :rule_name => rule.name,
          :item_name => item_names.uniq.join("\n"),
          :content => content_msgs.join("\n"),
          :reason => reason_msgs.join("\n")
        }
      end

    end

    delete_rule_ids = _delete_rules.map{|rule| rule[:rule_id]}
    delete_cond_ids = _delete_conds.map{|cond| cond[:cond_id]}
    delete_action_ids = _delete_actions.map{|action| action[:action_id]}
    delete_category_cond_sub_ids = _delete_category_cond_subs.map{|cond_sub| cond_sub[:sub_category_cond_id ]}
    delete_host_cond_sub_ids = _delete_host_cond_subs.map{|cond_sub| cond_sub[:sub_host_cond_id ]}
    delete_chp_user_cond_sub_ids = _delete_chp_user_cond_subs.map{|cond_sub| cond_sub[:chp_user_sub_id ]}
    delete_cm_ctag_ids = _delete_am_ctags.map{|cm_ctag| cm_ctag[:cm_ctag_id]}
    delete_am_ctag_ids = _delete_am_ctags.map{|am_ctag| am_ctag[:am_ctag_id]}

    #ルールの条件・操作・ルール自体の削除の原因となったカテゴリの情報が必要になった場合は、
    #上記の各アンダースコアで始まる変数名の配列を使って取得可能。(別途、処理の実装は必要)

    ret = {
      :delete_rule_ids => delete_rule_ids,
      :delete_action_ids => delete_action_ids,
      :delete_cond_ids => delete_cond_ids,
      :delete_sub_host_cond_ids => delete_host_cond_sub_ids,
      :delete_sub_category_cond_ids => delete_category_cond_sub_ids,
      :delete_chp_user_sub_ids => delete_chp_user_cond_sub_ids,
      :delete_cm_ctag_ids => delete_cm_ctag_ids,
      :delete_am_ctag_ids => delete_am_ctag_ids,
    }
  
    ret[:rule_messages] = _rule_messages if opts[:messages_necessary]
    ret

  end

  # カスタムタグの条件・操作が設定されているルールのうち、
  # :change_category_ctag_ids(ハッシュ)にカスタムタグidとカテゴリidを指定すると、
  # (カスタムタグの)カテゴリがルールのカテゴリから外れるカスタムタグとそのルールの情報を、
  # :delete_ctag_ids(配列)にカスタムタグのidを指定すると、
  # ルールに設定されているカスタムタグとそのルールの情報を返す
  # :actionは今のところcategory_change/deleteのいずれか
  # カスタムタグのカテゴリ変更/削除時に使用することを想定
  # 今の所必要ないため、両方指定した場合の処理は未実装。
  def self.ctags_related_rule_info(**opts)
    #初めの文字がアンダースコアの変数にしてあるものは再定義しないこと
    ctag_ids = []
    unless opts[:role_admin] then
      unless opts[:user_category_ids] then
        raise ArgumentError, "must be user_category_ids in case the user's role is not admin"
      end
    end

    case opts[:action]
    when  'delete' then
      if !opts[:delete_ctag_ids] then
        raise ArgumentError, "must be delete_ctag_ids in case of the delete"
      end
    when 'category_change' then
      if !opts[:change_category_ctag_ids] then
        raise ArgumentError, "must be change_category_ctag_ids in case of the category_change"
      end
    else
      #invalid
      raise ArgumentError, "invalid action"
    end
  
    if opts[:delete_ctag_ids] then
      opts[:delete_ctag_ids] = opts[:delete_ctag_ids].map(&:to_i)
      ctag_ids |= opts[:delete_ctag_ids]
    end

    if opts[:change_category_ctag_ids] then
      ctag_ids |= opts[:change_category_ctag_ids].keys.map(&:to_i)
      #opts[:change_category_ctag_ids]のkeyはStringで統一
      key_string_change_category_ctag_ids = {}
      opts[:change_category_ctag_ids].each do|ctagid, category_id|
        key_string_change_category_ctag_ids[ctagid.to_s] = category_id.to_i
      end
      opts[:change_category_ctag_ids] = key_string_change_category_ctag_ids
    end

    _rules = if ctag_ids.any? then
      SidfmdRule.where(%|
        id IN (?) OR id IN (?)
      |,SidfmdRuleCond.joins(:sidfmd_rule_c_ctag)
                      .joins("INNER JOIN sidfmd_rule_cm_ctags ON sidfmd_rule_cm_ctags.sidfmd_rule_c_ctag_id = sidfmd_rule_c_ctags.id")
                      .where(:sidfmd_rule_cm_ctags => { :sidfmd_ctag_id => ctag_ids })
                      .select(:sidfmd_rule_id), #カスタムタグ条件
        SidfmdRuleAction.joins(:sidfmd_rule_a_ctag)
                        .joins("INNER JOIN sidfmd_rule_am_ctags ON sidfmd_rule_am_ctags.sidfmd_rule_a_ctag_id = sidfmd_rule_a_ctags.id")
                        .where(:sidfmd_rule_am_ctags => { :sidfmd_ctag_id => ctag_ids })
                        .select(:sidfmd_rule_id) #カスタムタグ操作
      )
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_conds
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_conds /*ルールのcondの数*/ 
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_actions
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_actions /*ルールのactionの数*/ 
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, COUNT (sidfmd_rule_cm_ctags.id)], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_ctags ON sidfmd_rule_c_ctags.sidfmd_rule_cond_id = sidfmd_rule_conds.id
            INNER JOIN sidfmd_rule_cm_ctags ON sidfmd_rule_cm_ctags.sidfmd_rule_c_ctag_id = sidfmd_rule_c_ctags.id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id
            GROUP BY sidfmd_rule_conds.id
          )
        ) AS cond_ctag_counts /*ルールのカスタムタグ条件毎の、設定されているカスタムタグの数*/
      ")
      .select("
        (
          SELECT COUNT(sidfmd_rule_am_ctags.id)
          FROM sidfmd_rule_actions
          INNER JOIN sidfmd_rule_a_ctags ON sidfmd_rule_a_ctags.sidfmd_rule_action_id = sidfmd_rule_actions.id
          INNER JOIN sidfmd_rule_am_ctags ON sidfmd_rule_am_ctags.sidfmd_rule_a_ctag_id = sidfmd_rule_a_ctags.id
          WHERE sidfmd_rules.id = sidfmd_rule_actions.sidfmd_rule_id
          GROUP BY sidfmd_rule_actions.id
        ) AS action_ctag_count /*ルールのカスタムタグの操作に設定されているカスタムタグの数*/
      ")
      .select(:id, :name, :sidfmd_category_id)
    else
      SidfmdRule.where(:id => 0)
    end

    if opts[:delete_ctag_ids] then
      opts[:delete_ctag_ids] = [ 0 ] if opts[:delete_ctag_ids].size == 0
      _rules = _rules.select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_ctags.id, sidfmd_rule_cm_ctags.sidfmd_ctag_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_ctags ON sidfmd_rule_c_ctags.sidfmd_rule_cond_id = sidfmd_rule_conds.id
            INNER JOIN sidfmd_rule_cm_ctags ON sidfmd_rule_cm_ctags.sidfmd_rule_c_ctag_id = sidfmd_rule_c_ctags.id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND
              sidfmd_rule_cm_ctags.sidfmd_ctag_id IN (SELECT unnest (ARRAY[#{opts[:delete_ctag_ids].join(', ')}]))
          )
        ) AS delete_cond_ctags /*削除されるカスタムタグとそのカスタムタグ条件*/
      ")
      .select("
        ARRAY(
          (
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_am_ctags.id, sidfmd_rule_am_ctags.sidfmd_ctag_id], ':')
            FROM sidfmd_rule_actions
            INNER JOIN sidfmd_rule_a_ctags ON sidfmd_rule_a_ctags.sidfmd_rule_action_id = sidfmd_rule_actions.id
            INNER JOIN sidfmd_rule_am_ctags ON sidfmd_rule_am_ctags.sidfmd_rule_a_ctag_id = sidfmd_rule_a_ctags.id
            WHERE sidfmd_rules.id = sidfmd_rule_actions.sidfmd_rule_id AND
              sidfmd_rule_am_ctags.sidfmd_ctag_id IN (SELECT unnest (ARRAY[#{opts[:delete_ctag_ids].join(', ')}]))
          )
        ) AS delete_action_ctags /*削除されるカスタムタグとそのカスタムタグ操作*/
      ")
    end

    if opts[:change_category_ctag_ids] then
      target_ctag_ids = []
      opts[:change_category_ctag_ids].each do |ctagid, category_id|
        target_ctag_ids << ctagid if category_id != 0 #カテゴリが全体(指定なし)に変更される場合は、そのタグはルールに影響なし
      end
      target_ctag_ids =  [ 0 ] if target_ctag_ids.size == 0
      _rules = _rules.select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_ctags.id, sidfmd_rule_cm_ctags.sidfmd_ctag_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_ctags ON sidfmd_rule_c_ctags.sidfmd_rule_cond_id = sidfmd_rule_conds.id
            INNER JOIN sidfmd_rule_cm_ctags ON sidfmd_rule_cm_ctags.sidfmd_rule_c_ctag_id = sidfmd_rule_c_ctags.id
            WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND
              sidfmd_rule_cm_ctags.sidfmd_ctag_id IN (SELECT unnest (ARRAY[#{target_ctag_ids.join(', ')}]))
          )
        ) AS category_change_cond_ctags /*カスタムタグ条件のうち、カテゴリが「全体(指定なし)」以外に変更されるカスタムタグ*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_am_ctags.id, sidfmd_rule_am_ctags.sidfmd_ctag_id], ':')
            FROM sidfmd_rule_actions
            INNER JOIN sidfmd_rule_a_ctags ON sidfmd_rule_a_ctags.sidfmd_rule_action_id = sidfmd_rule_actions.id
            INNER JOIN sidfmd_rule_am_ctags ON sidfmd_rule_am_ctags.sidfmd_rule_a_ctag_id = sidfmd_rule_a_ctags.id
            WHERE sidfmd_rules.id = sidfmd_rule_actions.sidfmd_rule_id AND
              sidfmd_rule_am_ctags.sidfmd_ctag_id IN (SELECT unnest (ARRAY[#{target_ctag_ids.join(', ')}]))
          )
        ) AS category_change_action_ctags /*カスタムタグ操作のうち、カテゴリが「全体(指定なし)」以外に変更されるカスタムタグ*/
      ")
    end

    _ctag_cond_name = RULE_COND_NAMES['ctag']
    _ctag_action_name = RULE_ACTION_NAMES['set-ctag']
    _rule_messages = {}
    _delete_rules = []
    _delete_conds = []
    _delete_actions = []
    _delete_cm_ctags = []
    _delete_am_ctags = []

    get_cond_ctag_data = ->(cond_ctag){
      cond_ctag = cond_ctag.split(':')
      return {
        :cond_id => cond_ctag[0].to_i,
        :cm_ctag_id => cond_ctag[1].to_i,
        :ctag_id => cond_ctag[2].to_i,
        #:is_enable => cond_ctag[3].to_s == 'true',
      }
    }

    get_action_ctag_data = ->(action_ctag){
      action_ctag = action_ctag.split(':')
      return {
        :action_id => action_ctag[0].to_i,
        :am_ctag_id => action_ctag[1].to_i,
        :ctag_id => action_ctag[2].to_i,
        #:is_enable => action_ctag[3].to_s == 'true',
      }
    }

    get_delete_each_cond_ctags = ->(rule_delete_cond_ctags){
      rule_delete_each_cond_ctags = {}
      rule_delete_cond_ctags.each do |ctag|
        rule_delete_each_cond_ctags[ctag[:cond_id]] = [] unless rule_delete_each_cond_ctags[ctag[:cond_id]]
        rule_delete_each_cond_ctags[ctag[:cond_id]] << ctag
      end
      return rule_delete_each_cond_ctags
    }

    get_each_cond_ctags_count = ->(rule_cond_ctag_counts){
      rule_each_cond_ctags_counts = {}
      rule_cond_ctag_counts.each do |cond_ctag_counts|
        cond_ctag_counts = cond_ctag_counts.split(':')
        cond_id = cond_ctag_counts[0].to_i
        ctag_count = cond_ctag_counts[1].to_i
        rule_each_cond_ctags_counts[cond_id] = ctag_count 
      end
      return rule_each_cond_ctags_counts
    }

    create_delete_rule_object = -> (rule, opts){
      { :rule_id => rule.id, :opts => Array.wrap(opts) }
    }

    get_ctag_names = ->(ctags){
      t_ctags =
        SidfmdCtag.where(:id => ctags.map{|ctag| ctag[:ctag_id]})
                  .order(:priority => :desc)
                  .order('LOWER(keyword COLLATE "C") ASC')
      unless opts[:role_admin] then
        t_ctags = t_ctags.where("sidfmd_category_id IS NULL OR sidfmd_category_id IN (?)", opts[:user_category_ids])
      end
      t_ctag_names = t_ctags.map{|c| "「#{c.keyword}」" }
    }

    _action_name = opts[:action]
    create_ctag_cond_msg = -> (delete_ctag_cond_ids, delete_ctags_of_conds, all_ctag_cond_ids){
      del_ctag_names = delete_ctags_of_conds.any? ? get_ctag_names.(delete_ctags_of_conds) : []
      msg = ''
      ret = {}
      if delete_ctag_cond_ids.any? then
        if del_ctag_names.any? then
          msg = "#{delete_ctag_cond_ids.size}つの「#{_ctag_cond_name}」条件が削除され、残りの「#{_ctag_cond_name}」条件から#{del_ctag_names.join(', ')}が削除されます。"
        else
          msg = all_ctag_cond_ids.size == delete_ctag_cond_ids ?
            "「#{_ctag_cond_name}」条件が削除されます。" :
            "#{(delete_ctag_cond_ids.size == 1 || all_ctag_cond_ids.size == delete_ctag_cond_ids.size) ? '' : "#{delete_ctag_cond_ids.size}つの"}「#{_ctag_cond_name}」条件が削除されます。"
        end
      elsif delete_ctags_of_conds.any? then
        if del_ctag_names.any? then
          msg = "「#{_ctag_cond_name}」条件から#{del_ctag_names.join(', ')}が削除されます。"
        end
      end
      if msg.size > 0 then
        reason = if _action_name == 'delete' then
          "条件に設定されているカスタムタグが削除されるため。"
        else
          "条件に設定されているカスタムタグのカテゴリが、ルールのカテゴリと一致しなくなるため。"
        end
        ret = {
          :item => "条件",
          :content => msg,
          :reason => reason
        }
      end
      ret
    }

    create_ctag_action_msg = -> (is_action_ctag_deletion, delete_action_ctags){
      msg = ''
      ret = {}
      if is_action_ctag_deletion then
        msg = "「#{_ctag_action_name}」が削除されます。"
      elsif delete_action_ctags.any? then
        del_ctag_names = get_ctag_names.(delete_action_ctags)
        if del_ctag_names.any? then
          msg = "「#{_ctag_action_name}」から#{del_ctag_names.join(', ')}が削除されます。"
        end
      end
      if msg.size > 0 then
        reason = if _action_name == 'delete' then
          "操作に設定されているカスタムタグが削除されるため。"
        else
          "操作に設定されているカスタムタグのカテゴリが、ルールのカテゴリと一致しなくなるため。"
        end
        ret = {
          :item => "操作",
          :content => msg,
          :reason => reason
        }
      end
      ret
    }

    case _action_name
    when 'delete' then #カスタムタグ削除時
      _rules.each do |rule|
        next if rule.delete_cond_ctags.size == 0 && rule.delete_action_ctags.size == 0
        _rule_delete_cond_ctags = rule.delete_cond_ctags.map{|cond_ctag| get_cond_ctag_data.(cond_ctag) } #条件に設定されている削除するカスタムタグ
        _rule_delete_each_cond_ctags = get_delete_each_cond_ctags.(_rule_delete_cond_ctags) #カスタムタグ条件毎の削除するカスタムタグ
        _rule_each_cond_ctags_count = get_each_cond_ctags_count.(rule.cond_ctag_counts) #カスタムタグ条件毎のカスタムタグ数の数
        _rule_delete_ctag_cond_ids = [] #削除するカスタムタグ条件のid

        _rule_delete_action_ctags = rule.delete_action_ctags.map{|action_ctag| get_action_ctag_data.(action_ctag) } #操作に設定されている削除するカスタムタグ
        _is_action_ctag_deletion = rule.action_ctag_count != 0 && (rule.action_ctag_count == _rule_delete_action_ctags.size)

        _rule_delete_each_cond_ctags.each do |cond_id, ctags|
          if _rule_each_cond_ctags_count[cond_id] == ctags.size then
            _rule_delete_ctag_cond_ids << cond_id
          end
        end
        if _rule_delete_ctag_cond_ids.size == rule.count_conds then
          _delete_rules << create_delete_rule_object.(rule, _rule_delete_cond_ctags)
          if opts[:messages_necessary] then
            msg = _is_action_ctag_deletion && rule.count_actions == 1 ?
              "条件に「#{_ctag_cond_name}」のみが、操作に「#{_ctag_action_name}」のみが設定されており、選択されている全てのカスタムタグが削除されるため。" :
             "条件に「#{_ctag_cond_name}」のみが設定されており、選択されている全てのカスタムタグが削除されるため。"
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => msg
            }
          end
          next
        end

        if _is_action_ctag_deletion && rule.count_actions == 1 then
          _delete_rules << create_delete_rule_object.(rule, _rule_delete_action_ctags)
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "操作に「#{_ctag_action_name}」のみが設定されており、選択されている全てのカスタムタグが削除されるため。"
            }
          end
          next
        end

        item_names = []
        content_msgs = []
        reason_msgs = []

        #条件のチェック
        delete_ctags_of_conds = _rule_delete_cond_ctags.select{|c| !_rule_delete_ctag_cond_ids.include?(c[:cond_id])} #条件自体は削除されない削除対象のカスタムタグ
        _delete_cm_ctags += delete_ctags_of_conds if delete_ctags_of_conds.any?
        if _rule_delete_ctag_cond_ids.any? then
          _delete_conds += _rule_delete_cond_ctags.select{|c| _rule_delete_ctag_cond_ids.include?(c[:cond_id])} #条件自体が削除
          if opts[:messages_necessary] then
            msg = create_ctag_cond_msg.(_rule_delete_ctag_cond_ids, delete_ctags_of_conds, _rule_each_cond_ctags_count.keys)
            if msg.size > 0 then 
              item_names << msg[:item]
              content_msgs << msg[:content]
              reason_msgs << msg[:reason]
            end
          end
        elsif delete_ctags_of_conds.any? then
          if opts[:messages_necessary] then
            msg = create_ctag_cond_msg.(_rule_delete_ctag_cond_ids, delete_ctags_of_conds, _rule_each_cond_ctags_count.keys)
            if msg.size > 0 then
              item_names << msg[:item]
              content_msgs << msg[:content]
              reason_msgs << msg[:reason]
            end
          end
        end

        #操作のチェック
        if _is_action_ctag_deletion then
          _delete_actions += _rule_delete_action_ctags
          if opts[:messages_necessary] then
            msg = create_ctag_action_msg.(_is_action_ctag_deletion, _rule_delete_action_ctags)
            if msg.size > 0 then
              item_names << msg[:item]
              content_msgs << msg[:content]
              reason_msgs << msg[:reason]
            end
          end
        elsif _rule_delete_action_ctags.any? then
          _delete_am_ctags += _rule_delete_action_ctags
          if opts[:messages_necessary] then
            msg = create_ctag_action_msg.(_is_action_ctag_deletion, _rule_delete_action_ctags)
            if msg.size > 0 then
              item_names << msg[:item]
              content_msgs << msg[:content]
              reason_msgs << msg[:reason]
            end
          end
        end

        if opts[:messages_necessary] and item_names.any? then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => item_names.uniq.join("\n"),
            :content => content_msgs.join("\n"),
            :reason => reason_msgs.join("\n")
          }
        end
      end
    when 'category_change' then #カスタムタグのカテゴリ変更時
      _rules.each do |rule|
        if !rule.sidfmd_category_id || (rule.category_change_cond_ctags.size == 0 && rule.category_change_action_ctags.size == 0)
          next #ルールのカテゴリが「全体」ならカスタムタグのカテゴリは影響なし
        end
        _rule_delete_cond_ctags = [] #ルールのカテゴリから外れてルールから削除される、条件のカスタムタグ
        _rule_delete_action_ctags = [] #ルールのカテゴリから外れてルールから削除される、操作のカスタムタグ
        rule.category_change_cond_ctags.each do |cond_ctag|
          ctag_data = get_cond_ctag_data.(cond_ctag)
          if rule.sidfmd_category_id != opts[:change_category_ctag_ids][ctag_data[:ctag_id].to_s].to_i then
            _rule_delete_cond_ctags << ctag_data
          end
        end
        rule.category_change_action_ctags.each do |action_ctag|
          ctag_data = get_action_ctag_data.(action_ctag)
          if rule.sidfmd_category_id != opts[:change_category_ctag_ids][ctag_data[:ctag_id].to_s].to_i then
            _rule_delete_action_ctags << ctag_data
          end
        end
    
        _rule_each_cond_ctags_count = get_each_cond_ctags_count.(rule.cond_ctag_counts) #カスタムタグ条件毎のカスタムタグ数の数  
        _rule_delete_each_cond_ctags = get_delete_each_cond_ctags.(_rule_delete_cond_ctags) #カスタムタグ条件毎のルールから削除するカスタムタグ
        _rule_delete_ctag_cond_ids = [] #ルールから削除するカスタムタグ条件のid
        _is_action_ctag_deletion = rule.action_ctag_count != 0 && (rule.action_ctag_count == _rule_delete_action_ctags.size)

        _rule_delete_each_cond_ctags.each do |cond_id, ctags|
          if _rule_each_cond_ctags_count[cond_id] == ctags.size then
            _rule_delete_ctag_cond_ids << cond_id
          end
        end

        if _rule_delete_ctag_cond_ids.size == rule.count_conds then
          _delete_rules << create_delete_rule_object.(rule, _rule_delete_cond_ctags)
          if opts[:messages_necessary] then
            msg = _is_action_ctag_deletion && rule.count_actions == 1 ?
              "条件に「#{_ctag_cond_name}」のみが、操作に「#{_ctag_action_name}」のみが設定されており、選択されている全てのカスタムタグのカテゴリが、ルールのカテゴリと一致しなくなるため。" :
              "条件に「#{_ctag_cond_name}」のみが設定されており、選択されている全てのカスタムタグのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => msg
            }
          end
          next
        end

        if _is_action_ctag_deletion && rule.count_actions == 1 then
          _delete_rules << create_delete_rule_object.(rule, _rule_delete_action_ctags)
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "操作に「#{_ctag_action_name}」のみが設定されており、選択されている全てのカスタムタグのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        item_names = []
        content_msgs = []
        reason_msgs = []

        #条件のチェック
        delete_ctags_of_conds = _rule_delete_cond_ctags.select{|c| !_rule_delete_ctag_cond_ids.include?(c[:cond_id])} #条件自体は削除されない削除対象のカスタムタグ
        _delete_cm_ctags += delete_ctags_of_conds if delete_ctags_of_conds.any?
        if _rule_delete_ctag_cond_ids.any? then
          _delete_conds += _rule_delete_cond_ctags.select{|c| _rule_delete_ctag_cond_ids.include?(c[:cond_id])} #条件自体が削除
          if opts[:messages_necessary] then
            msg = create_ctag_cond_msg.(_rule_delete_ctag_cond_ids, delete_ctags_of_conds, _rule_each_cond_ctags_count.keys)
            if msg.size > 0 then
              item_names << msg[:item]
              content_msgs << msg[:content]
              reason_msgs << msg[:reason]
            end
          end
        elsif delete_ctags_of_conds.any? then
          if opts[:messages_necessary] then
            msg = create_ctag_cond_msg.(_rule_delete_ctag_cond_ids, delete_ctags_of_conds, _rule_each_cond_ctags_count.keys)
            if msg.size > 0 then
              item_names << msg[:item]
              content_msgs << msg[:content]
              reason_msgs << msg[:reason]
            end
          end
        end

        #操作のチェック
        if _is_action_ctag_deletion then
          _delete_actions += _rule_delete_action_ctags
          if opts[:messages_necessary] then
            msg = create_ctag_action_msg.(_is_action_ctag_deletion, _rule_delete_action_ctags)
            if msg.size > 0 then
              item_names << msg[:item]
              content_msgs << msg[:content]
              reason_msgs << msg[:reason]
            end
          end
        elsif _rule_delete_action_ctags.any? then
          _delete_am_ctags += _rule_delete_action_ctags
          if opts[:messages_necessary] then
            msg = create_ctag_action_msg.(_is_action_ctag_deletion, _rule_delete_action_ctags)
            if msg.size > 0 then
              item_names << msg[:item]
              content_msgs << msg[:content]
              reason_msgs << msg[:reason]
            end
          end
        end

        if opts[:messages_necessary] and item_names.any? then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => item_names.uniq.join("\n"),
            :content => content_msgs.join("\n"),
            :reason => reason_msgs.join("\n")
          }
        end
      end
    end

    is_unchangeable_ctag = false
    unchangeable_ctags = []

    if opts[:role_admin] then
      delete_rule_ids = _delete_rules.map{|rule| rule[:rule_id]}.uniq
      delete_cond_ids = _delete_conds.map{|cond| cond[:cond_id]}.uniq
      delete_action_ids = _delete_actions.map{|action| action[:action_id]}.uniq
    else
      #管理者でなければ、カテゴリ範囲外のルールのルール自体が削除される、または操作・条件自体が削除される場合は変更できないカスタムタグを取得する
      unchangeable_ctag_ids = []
      t_rules = {}
      _delete_rules.each do |rule|
        if t_rules[rule[:rule_id]] then
          t_rules[rule[:rule_id]] += rule[:opts]
        else
          t_rules[rule[:rule_id]] = rule[:opts]
        end
      end
      t_conds = {}
      _delete_conds.each do |cond|
        t_conds[cond[:cond_id]] = [] unless t_conds[cond[:cond_id]]
        t_conds[cond[:cond_id]] << cond[:ctag_id]
      end
      t_actions = {}
      _delete_actions.each do |action|
        t_actions[action[:action_id]] = [] unless t_actions[action[:action_id]]
        t_actions[action[:action_id]] << action[:ctag_id]
      end
      unchangeable_rules = (opts[:user_category_ids].size > 0 ? SidfmdRule.where(%|sidfmd_category_id NOT IN (?) OR sidfmd_category_id IS NULL|, opts[:user_category_ids]) : SidfmdRule.all)

      unchangeable_rules.where(:id => t_rules.keys).each do |rule|
        unchangeable_ctag_ids |= t_rules[rule.id].map{|opt| opt[:ctag_id]}
        t_rules.delete(rule.id)
        _rule_messages.delete(rule.id) if opts[:messages_necessary]
      end

      SidfmdRuleCond.where(%|
        id IN (?) AND sidfmd_rule_id IN (?)
      |, t_conds.keys, unchangeable_rules.select(:id)
      )
      .select('id AS cond_id', 'sidfmd_rule_id AS rule_id')
      .each do |cond|
        unchangeable_ctag_ids |= t_conds[cond.cond_id]
        t_conds.delete(cond.cond_id)
        _rule_messages.delete(cond.rule_id) if opts[:messages_necessary]
      end

      SidfmdRuleAction.where(%|
        id IN (?) AND sidfmd_rule_id IN (?)
      |, t_actions.keys, unchangeable_rules.select(:id)
      )
      .select('id AS action_id', 'sidfmd_rule_id AS rule_id')
      .each do |action|
        unchangeable_ctag_ids |= t_actions[action.action_id]
        t_actions.delete(action.action_id)
        _rule_messages.delete(action.rule_id) if opts[:messages_necessary]
      end

      delete_rule_ids = t_rules.keys
      delete_cond_ids = t_conds.keys
      delete_action_ids = t_actions.keys
      if unchangeable_ctag_ids.any? then
        is_unchangeable_ctag = true
        if opts[:messages_necessary] then
          unchangeable_ctags =
            SidfmdCtag.where(:id => unchangeable_ctag_ids)
                      .where("sidfmd_category_id IS NULL OR sidfmd_category_id IN (?)", opts[:user_category_ids])
                      .order(:priority)
                      .order('LOWER(keyword COLLATE "C") ASC')
                      .pluck(:keyword)
        end
      end
      if opts[:messages_necessary] then
        unchangeable_rules.where(:id => _rule_messages.keys).each do |rule|
          _rule_messages.delete(rule.id)
        end
      end 
    end

    ret = {
      :delete_rule_ids => delete_rule_ids,
      :delete_cond_ids => delete_cond_ids,
      :delete_action_ids => delete_action_ids,
      :delete_cm_ctag_ids => _delete_cm_ctags.map{|cm_ctag| cm_ctag[:cm_ctag_id]},
      :delete_am_ctag_ids => _delete_am_ctags.map{|am_ctag| am_ctag[:am_ctag_id]},
      :is_unchangeable_ctag => is_unchangeable_ctag,
      :unchangeable_ctags => unchangeable_ctags
    }

    ret[:rule_messages] = _rule_messages if opts[:messages_necessary]
    ret
  end

  # 指定したローカルプロダクトidに関連する条件が設定されているルールと処理するデータを返す
  # 今の所ローカルプロダクトの削除時の処理のみ
  # :actionは今の所deleteのみ
  # ローカルプロダクトの削除時に使用することを想定
  def self.lproducts_related_rule_info(**opts)
    #初めの文字がアンダースコアの変数にしてあるものは再定義しないこと
    unless opts[:role_admin] then
      raise ArgumentError, "only admin users permited to execute"
      #unless opts[:user_category_ids] then
      #  raise ArgumentError, "must be user_category_ids in case the user's role is not admin"
      #end
    end

    case opts[:action]
    when  'delete' then
      if !opts[:delete_lproduct_ids] || !opts[:delete_lversion_ids] then
        raise ArgumentError, "must be delete_lproduct_ids and delete_lversion_ids in case of the delete"
      end
    else
      #invalid
      raise ArgumentError, "invalid action"
    end

    _rules = if opts[:delete_lproduct_ids].any? or opts[:delete_lversion_ids].any? then
      SidfmdRule.where(%|
        id IN (?)
      |, SidfmdRuleCond.joins(:sidfmd_rule_c_match_app)
                       .joins("INNER JOIN sidfmd_rule_cm_lapp_subs ON sidfmd_rule_cm_lapp_subs.sidfmd_rule_c_match_app_id = sidfmd_rule_c_match_apps.id")
                       .where("
                         sidfmd_rule_cm_lapp_subs.sidfmd_lproduct_id IN (?) OR
                         sidfmd_rule_cm_lapp_subs.sidfmd_lversion_id IN (?)
                       ", opts[:delete_lproduct_ids], opts[:delete_lversion_ids])
                       .select(:sidfmd_rule_id), #プロダクト限定・除外条件
      )
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_conds
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_conds /*ルールのcondの数*/ 
      ")
      .select("
        (
          SELECT COUNT(sidfmd_rule_cm_app_subs.id)
          FROM sidfmd_rule_conds
          INNER JOIN sidfmd_rule_c_match_apps ON sidfmd_rule_c_match_apps.sidfmd_rule_cond_id = sidfmd_rule_conds.id
          INNER JOIN sidfmd_rule_cm_app_subs ON sidfmd_rule_cm_app_subs.sidfmd_rule_c_match_app_id = sidfmd_rule_c_match_apps.id
          WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_apps.inv IS FALSE
        ) AS cond_product_match_count /* プロダクト限定に設定されているSIDfmプロダクト・バージョンの数*/
      ")
      .select("
        (
          SELECT COUNT(sidfmd_rule_cm_app_subs.id)
          FROM sidfmd_rule_conds
          INNER JOIN sidfmd_rule_c_match_apps ON sidfmd_rule_c_match_apps.sidfmd_rule_cond_id = sidfmd_rule_conds.id
          INNER JOIN sidfmd_rule_cm_app_subs ON sidfmd_rule_cm_app_subs.sidfmd_rule_c_match_app_id = sidfmd_rule_c_match_apps.id
          WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_apps.inv IS TRUE
        ) AS cond_product_unmatch_count /* プロダクト除外に設定されているSIDfmプロダクト・バージョンの数*/
      ")
      .select("
        (
          SELECT COUNT(sidfmd_rule_cm_lapp_subs.id)
          FROM sidfmd_rule_conds
          INNER JOIN sidfmd_rule_c_match_apps ON sidfmd_rule_c_match_apps.sidfmd_rule_cond_id = sidfmd_rule_conds.id
          INNER JOIN sidfmd_rule_cm_lapp_subs ON sidfmd_rule_cm_lapp_subs.sidfmd_rule_c_match_app_id = sidfmd_rule_c_match_apps.id
          WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_apps.inv IS FALSE
        ) AS cond_lproduct_match_count /* プロダクト限定に設定されているローカルプロダクト・バージョンの数*/
      ")
      .select("
        (
          SELECT COUNT(sidfmd_rule_cm_lapp_subs.id)
          FROM sidfmd_rule_conds
          INNER JOIN sidfmd_rule_c_match_apps ON sidfmd_rule_c_match_apps.sidfmd_rule_cond_id = sidfmd_rule_conds.id
          INNER JOIN sidfmd_rule_cm_lapp_subs ON sidfmd_rule_cm_lapp_subs.sidfmd_rule_c_match_app_id = sidfmd_rule_c_match_apps.id
          WHERE sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND sidfmd_rule_c_match_apps.inv IS TRUE
        ) AS cond_lproduct_unmatch_count /* プロダクト除外に設定されているローカルプロダクト・バージョンの数*/
      ")
      .select(:id, :name, :sidfmd_category_id)
    else
      SidfmdRule.where(:id => 0)
    end

    if opts[:delete_lproduct_ids].any? or opts[:delete_lversion_ids].any? then 
      delete_lproduct_ids = opts[:delete_lproduct_ids].any? ? opts[:delete_lproduct_ids] : [-1]
      delete_lversion_ids = opts[:delete_lversion_ids].any? ? opts[:delete_lversion_ids] : [-1]
      _rules = _rules.select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_lapp_subs.id, sidfmd_lproduct_id, sidfmd_lversion_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_apps ON sidfmd_rule_c_match_apps.sidfmd_rule_cond_id = sidfmd_rule_conds.id
            INNER JOIN sidfmd_rule_cm_lapp_subs ON sidfmd_rule_cm_lapp_subs.sidfmd_rule_c_match_app_id = sidfmd_rule_c_match_apps.id
            WHERE
              sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND
              sidfmd_rule_c_match_apps.inv IS FALSE AND
              ( 
                sidfmd_rule_cm_lapp_subs.sidfmd_lproduct_id IN (SELECT unnest(ARRAY[#{delete_lproduct_ids.join(', ')}])) OR
                sidfmd_rule_cm_lapp_subs.sidfmd_lversion_id IN (SELECT unnest(ARRAY[#{delete_lversion_ids.join(', ')}]))
              )
          )
        ) AS delete_macth_lproducts /*プロダクト限定に設定されている削除されるローカルプロダクト・バージョン*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_conds.id, sidfmd_rule_cm_lapp_subs.id, sidfmd_lproduct_id, sidfmd_lversion_id], ':')
            FROM sidfmd_rule_conds
            INNER JOIN sidfmd_rule_c_match_apps ON sidfmd_rule_c_match_apps.sidfmd_rule_cond_id = sidfmd_rule_conds.id
            INNER JOIN sidfmd_rule_cm_lapp_subs ON sidfmd_rule_cm_lapp_subs.sidfmd_rule_c_match_app_id = sidfmd_rule_c_match_apps.id
            WHERE
              sidfmd_rules.id = sidfmd_rule_conds.sidfmd_rule_id AND
              sidfmd_rule_c_match_apps.inv IS TRUE AND
              ( 
                sidfmd_rule_cm_lapp_subs.sidfmd_lproduct_id IN (SELECT unnest(ARRAY[#{delete_lproduct_ids.join(', ')}])) OR
                sidfmd_rule_cm_lapp_subs.sidfmd_lversion_id IN (SELECT unnest(ARRAY[#{delete_lversion_ids.join(', ')}]))
              )
          )
        ) AS delete_unmacth_lproducts /*プロダクト除外に設定されている削除されるローカルプロダクト・バージョン */
      ")
    end

    _match_product_cond_name = RULE_COND_NAMES['product-match']
    _unmatch_product_cond_name = RULE_COND_NAMES['product-unmatch']
    _rule_messages = {}
    _delete_rules = []
    _delete_conds = []
    _delete_actions = []
    _delete_cm_lapps = []

    create_delete_rule_object = -> (rule, opts){
      { :rule_id => rule.id, :opts => Array.wrap(opts) }
    }

    get_cond_lproduct_data = -> (cond_lproduct){
      cond_lproduct = cond_lproduct.split(':')
      {
        :cond_id => cond_lproduct[0].to_i,
        :cm_id => cond_lproduct[1].to_i,
        :lproduct_id => cond_lproduct[2].to_i,
        :lversion_id => cond_lproduct[3].to_i,
      }
    }

    #get_lprod_ids = -> (cond_lproducts){
    #  cond_lproducts.map{|lp| lp[:lversion_id].to_i == 0 ? lp[:lproduct_id] : nil}.compact
    #}

    #get_lver_ids = -> (cond_lproducts){
    #  cond_lproducts.select{|lv| lv[:lversion_id].to_i != 0 ? lv[:lversion_id] : nil}.compact
    #}

    get_lprod_names = -> (cond_lproducts){
      lp_ids = []
      lv_ids = []
      cond_lproducts.each do |cond_lp|
        if cond_lp[:lversion_id].to_i == 0 then
          lp_ids << cond_lp[:lproduct_id]
        else
          lv_ids << cond_lp[:lversion_id]
        end
      end
      lp_names = lp_ids.any? ? SidfmdLproduct.where(:id => lp_ids).pluck(:name) : []
      lp_names += SidfmdLversion.where(:id => lv_ids).pluck(:fullname) if lv_ids.any?
      lp_names = lp_names.sort.map{|name| "「#{name}」"}
    }
 
    case opts[:action]
    when 'delete'
      _rules.each do |rule|
        _match_prod_count = rule.cond_product_match_count + rule.cond_lproduct_match_count # プロダクト限定に設定されているプロダクト数
        _unmatch_prod_count = rule.cond_product_unmatch_count + rule.cond_lproduct_unmatch_count # プロダクト除外に設定されているプロダクト数
        _is_match_prod_delete = rule.delete_macth_lproducts.size > 0 && rule.delete_macth_lproducts.size == _match_prod_count # プロダクト限定を削除するか否か
        _is_unmatch_prod_delete = rule.delete_unmacth_lproducts.size > 0 && rule.delete_unmacth_lproducts.size == _unmatch_prod_count #プロダクト除外を削除するか否か
        
        _rule_delete_cond_match_lproducts = rule.delete_macth_lproducts.map{|lp| get_cond_lproduct_data.(lp)} # プロダクト限定に設定されている削除するローカルプロダクトのデータ
        _rule_delete_cond_unmatch_lproducts = rule.delete_unmacth_lproducts.map{|lp| get_cond_lproduct_data.(lp)} # プロダクト除外に設定されている削除するローカルプロダクトのデータ

        if _is_match_prod_delete && _is_unmatch_prod_delete && rule.count_conds == 2 then
          _delete_rules << create_delete_rule_object.(rule, (_rule_delete_cond_match_lproducts + _rule_delete_cond_unmatch_lproducts))
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "条件に「#{_match_product_cond_name}」と「#{_unmatch_product_cond_name}」のみが設定されており、選択されている全てのプロダクトが削除されるため。" 
            }
          end
          next
        end

        if rule.count_conds == 1 && (_is_match_prod_delete || _is_unmatch_prod_delete) then
          _delete_rules << create_delete_rule_object.(rule, (_rule_delete_cond_match_lproducts + _rule_delete_cond_unmatch_lproducts))
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "条件に「#{_is_match_prod_delete ? _match_product_cond_name : _unmatch_product_cond_name}」のみが設定されており、選択されている全てのプロダクトが削除されるため。"
            }
          end
          next
        end

        item_names = []
        content_msgs = []
        reason_msgs = []

        if _rule_delete_cond_match_lproducts.any? or _rule_delete_cond_unmatch_lproducts.any? then
          if _is_match_prod_delete && _is_unmatch_prod_delete then # プロダクト限定・除外の両方の条件削除
            _delete_conds += (_rule_delete_cond_match_lproducts + _rule_delete_cond_unmatch_lproducts)
            if opts[:messages_necessary] then
              item_names << "条件"
              content_msgs << "「#{_match_product_cond_name}」と「#{_unmatch_product_cond_name}」が削除されます。"
              reason_msgs << "条件に設定されている全てのプロダクトが削除されるため。"
            end
          elsif _is_match_prod_delete then # プロダクト限定削除
            _delete_conds += _rule_delete_cond_match_lproducts
            _delete_cm_lapps += _rule_delete_cond_unmatch_lproducts
            if opts[:messages_necessary] then
              del_unmatch_lprod_names = get_lprod_names.(_rule_delete_cond_unmatch_lproducts)
              item_names << "条件"
              if del_unmatch_lprod_names.any? then # プロダクト除外からプロダクトを削除
                content_msgs << "「#{_match_product_cond_name}」が削除され、「#{_unmatch_product_cond_name}」から#{del_unmatch_lprod_names.join(', ')}が削除されます。"
                reason_msgs << "「#{_match_product_cond_name}」の条件に設定されている全てのプロダクトが削除され、「#{_unmatch_product_cond_name}」に設定されているプロダクトが削除されるため。"
              else
                content_msgs << "「#{_match_product_cond_name}」が削除されます。"
                reason_msgs << "条件に設定されている全てのプロダクトが削除されるため。"
              end
            end
          elsif _is_unmatch_prod_delete then # プロダクト除外削除
            _delete_conds += _rule_delete_cond_unmatch_lproducts
            _delete_cm_lapps += _rule_delete_cond_match_lproducts
            if opts[:messages_necessary] then
              del_match_lprod_names = get_lprod_names.(_rule_delete_cond_match_lproducts)
              item_names << "条件"
              if del_match_lprod_names.any? then # プロダクト限定からプロダクト削除
                content_msgs << "「#{_unmatch_product_cond_name}」が削除され、「#{_match_product_cond_name}」から#{del_match_lprod_names.join(', ')}が削除されます。"
                reason_msgs << "「#{_unmatch_product_cond_name}」の条件に設定されている全てのプロダクトが削除され、「#{_match_product_cond_name}」に設定されているプロダクトが削除されるため。"
              else
                content_msgs << "「#{_unmatch_product_cond_name}」が削除されます。"
                reason_msgs << "条件に設定されている全てのプロダクトが削除されるため。"
              end
            end
          else
            if _rule_delete_cond_match_lproducts.any? then # プロダクト限定からプロダクト削除
              _delete_cm_lapps += _rule_delete_cond_match_lproducts
              if opts[:messages_necessary] then
                del_match_lprod_names = get_lprod_names.(_rule_delete_cond_match_lproducts)
                if del_match_lprod_names.any? then
                  item_names << "条件"
                  content_msgs << "「#{_match_product_cond_name}」から#{del_match_lprod_names.join(', ')}が削除されます。"
                  reason_msgs << "条件に設定されているプロダクトが削除されるため。"
                end 
              end
            end
            if _rule_delete_cond_unmatch_lproducts.any? then # プロダクト除外からプロダクトを削除
              _delete_cm_lapps += _rule_delete_cond_unmatch_lproducts
              if opts[:messages_necessary] then
                del_unmatch_lprod_names = get_lprod_names.(_rule_delete_cond_unmatch_lproducts)
                if del_unmatch_lprod_names.any? then
                  item_names << "条件"
                  content_msgs << "「#{_unmatch_product_cond_name}」から#{del_unmatch_lprod_names.join(', ')}が削除されます。"
                  reason_msgs << "条件に設定されているプロダクトが削除されるため。"
                end 
              end
            end
          end
        end

        if opts[:messages_necessary] and item_names.any? then
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => item_names.uniq.join("\n"),
            :content => content_msgs.join("\n"),
            :reason => reason_msgs.join("\n")
          }
        end
        
      end
    end

    delete_rule_ids = _delete_rules.map{|rule| rule[:rule_id]}.uniq
    delete_cond_ids = _delete_conds.map{|cond| cond[:cond_id]}.uniq
    delete_action_ids = _delete_actions.map{|action| action[:action_id]}.uniq #今のところactionは無し(常に空)
    delete_cm_lapp_ids = _delete_cm_lapps.map{|cm_cond| cm_cond[:cm_id]}.uniq

    ret = {
      :delete_rule_ids => delete_rule_ids,
      :delete_cond_ids => delete_cond_ids,
      :delete_action_ids => delete_action_ids,
      :delete_cm_lapp_ids => delete_cm_lapp_ids
    }

    ret[:rule_messages] = _rule_messages if opts[:messages_necessary]
    ret

  end

end
